// Edit date at the bottom of the file

// (c) Kevin G. Dunn, McMaster University, June 2006 : dunnkg@mcmaster.ca

// To add:
// Add function to extract sub-components of an image (for block wise processing)
// Residual image calculation: change so that loadings and nPC's to be input, instead of Residual operator.  Update instructions so that figure, imshow... shows the image correctly
// Function: Im2Scores: image space to score space

#include "mex.h"
#include "math.h"   // because we use the "sqrt" function
//1
void ComputeXTX(mxArray *plhs[],const mxArray *prhs[]);
void unsignedcharXTXcalc(unsigned char *X, double *XTXmatrix, int nLayers, int nPixels);
void doubleXTXcalc(double *X, double *XTXmatrix, int nLayers, int nPixels);
//2
void ImageStats(mxArray *plhs[],const mxArray *prhs[],int nlhs);
//3
void ImagePCA(mxArray *plhs[],const mxArray *prhs[],int nlhs);
//4
void ScorePlot(mxArray *plhs[], const mxArray *prhs[]);
//5
void Mask2Im(mxArray *plhs[], const mxArray *prhs[],int nrhs);
//6
void Residual(mxArray *plhs[], const mxArray *prhs[],int nrhs);

/*--------------------------------------------------------------------------------------------------------------------------------------------
[1] ComputeXTX:          INPUTS:   Mandatory:  1. imageX        (nRows x nCols x nLayers)   [uint8]
                         OUTPUTS:  Mandatory:  1. XTXmatrix     (nLayers x nLayers)         [double]

    AIM: If this function were to be computed in Matlab, we would unfold the incoming image into a 2-D matrix with nRows*nCols rows
         and nLayers columns.  We would have to convert it to double precision, increasing the memory requirements 8 times and then
         compute X'*X.  This MEX function avoids the unnecessary memory usage.
*/
void ComputeXTX(mxArray *plhs[],const mxArray *prhs[])
{
    unsigned char *imageX;                      // Input  1 : array (of type unsigned char) containing the uint8 input image    (nRows x nCols x nLayers)
    double        *imageXdouble;                // Input  1 : as above: alternative when input imageX is of type double         (nRows x nCols x nLayers)
    double        *XTXmatrix;                   // Output 1 : X transpose X matrix of the unfolded image                        (nLayers x nLayers)

    int            nLayers;                     // Working  : number of layers in the image (eg: 3 for RGB images)
    int            nPixels;                     // Working  : number of pixels per layer in the image
    int            nDims;                       // Working  : number of dimensions of the input
    int           *ImageDimensions;             // Working  : pointer to the image size: cannot be changed

    nDims = mxGetNumberOfDimensions(prhs[1]);                               // Get the dimensions of the input image
    if (nDims != 3 )
      mexErrMsgTxt("libmia error in function ComputeXTX: The input image must be stored in a three dimensional variable.\n");
    else
    { ImageDimensions = mxGetDimensions(prhs[1]);                           // Get the dimensions of the input image: RETURNS A "const int" - because we cannot change its value
      nLayers         = ImageDimensions[2];                                 // Get number of layers in the image
      nPixels         = ImageDimensions[0]*ImageDimensions[1];              // Number of pixels in this image
    }

    plhs[0]   = mxCreateDoubleMatrix(nLayers,nLayers,mxREAL);               // Set output pointer 1 to the XTX matrix
    XTXmatrix = mxGetPr(plhs[0]);                                           // Create C pointers to copies of the output matrices

    if (mxGetClassID((mxArray *)prhs[1]) == mxDOUBLE_CLASS){
      imageXdouble = mxGetPr(prhs[1]);
      doubleXTXcalc(imageXdouble, XTXmatrix, nLayers, nPixels);
    }

    if (mxGetClassID((mxArray *)prhs[1]) == mxUINT8_CLASS){
      imageX = (unsigned char *)mxGetData(prhs[1]);
      unsignedcharXTXcalc(imageX, XTXmatrix, nLayers, nPixels);
    }
return;
}

void unsignedcharXTXcalc(unsigned char *X, double *XTXmatrix, int nLayers, int nPixels)
{
// XTXcalc: for unsigned char input
//  unsigned char *imageX;                      // Input  1 : array (of type unsigned char) containing the uint8 input image    (nRows x nCols x nLayers)
//  double        *XTXmatrix;                   // Output 1 : X transpose X matrix of the unfolded image                        (nLayers x nLayers)
    double         temp;                        // Working  : holds intermediate matrix multiplication results
    int            i,j,k;                       // Working  : index variables for looping

  for (i=0;i<nLayers;i++) {                                                 // Compute X'*X directly from the original image
    for (j=0;j<nLayers;j++) {
      temp = 0;
      for (k=0;k<nPixels;k++){
        temp += X[k+i*nPixels] * X[k+j*nPixels];
      }
      XTXmatrix[j+i*nLayers] = temp;
    }
  }
  return;
}

void doubleXTXcalc(double *X, double *XTXmatrix, int nLayers, int nPixels)
{
// XTXcalc: for double unsigned char input
//  unsigned char *imageX;                      // Input  1 : array (of type unsigned char) containing the uint8 input image    (nRows x nCols x nLayers)
//  double        *XTXmatrix;                   // Output 1 : X transpose X matrix of the unfolded image                        (nLayers x nLayers)
    double         temp;                        // Working  : holds intermediate matrix multiplication results
    int            i,j,k;                       // Working  : index variables for looping

  for (i=0;i<nLayers;i++) {                                                 // Compute X'*X directly from the original image
    for (j=0;j<nLayers;j++) {
      temp = 0;
      for (k=0;k<nPixels;k++){
        temp += X[k+i*nPixels] * X[k+j*nPixels];
      }
      XTXmatrix[j+i*nLayers] = temp;
    }
  }
  return;
}


/*--------------------------------------------------------------------------------------------------------------------------------------------
[2] ImageStats:          INPUTS:   Mandatory:  1. imageX    (nRows x nCols x nLayers)   [uint8]
                         OUTPUTS:  Mandatory:  1. sumX      (1 x nLayers)               [double]
                                               2. meanX     (1 x nLayers)               [double]
                                               3. minX      (1 x nLayers)               [double]
                                               4. maxX      (1 x nLayers)               [double]
                                               5. varX      (1 x nLayers)               [double]


    AIM: This function computes basic statistics about the image, such as the sum of pixel values in each layer, the mean pixel value in
         each layer, the variance of pixel values in each layer.  It can be readily extended to compute other information
*/
void ImageStats(mxArray *plhs[],const mxArray *prhs[],int nlhs)
{
    unsigned char *imageX;                      // Input  1 : array (of type unsigned char) containing the uint8 input image    (nRows x nCols x nLayers)
    double        *imageXdouble;                // Input  1 : as above, but alternative when the input is of type double        (nRows x nCols x nLayers)
    double        *sumX;                        // Output 1 : the sum of each column in the unfolded X matrix                   (1 x nLayers)
    double        *meanX;                       // Output 2 : the mean value of the pixel intensities in each layer             (1 x nLayers)
    double        *minX;                        // Output 3 : the minimum pixel intensity value in each layer                   (1 x nLayers)
    double        *maxX;                        // Output 4 : the maximum pixel intensity value in each layer                   (1 x nLayers)
    double        *varX;                        // Output 5 : the variance of the pixel intensities in each layer               (1 x nLayers)

    double         doubleflag;                  // Working: A 0-1 variable indicating if the input image is of type double(1) or uint8(0)
    int            nLayers;                     // Working: The number of layers in the image (eg: 3 for RGB images)
    int            nPixels;                     // Working: The number of pixels per layer in the image
    int            nDims;                       // Working: The number of dimensions of the input
    int            j,k;                         // Working: Index variables for looping (mainly matrix multiplication)
    int           *ImageDimensions;             // Working: A pointer to the image size: cannot be changed

// ------ INPUT CHECKING ------
    nDims             = mxGetNumberOfDimensions(prhs[1]);                   // Get the dimensions of the input image: result is always 2 or greater
    if (nDims > 3 )
       mexErrMsgTxt("libmia error in function ImageStats: The input image must be stored in a three dimensional variable.\n");
    else
    {
      if (nDims==2)
      { ImageDimensions = mxGetDimensions(prhs[1]);                         // Get the dimensions of the input image: RETURNS A "const int" - we cannot change its value
        nLayers         = 1;                                                // Layers = 1 for a "flat" image
        nPixels         = ImageDimensions[0]*ImageDimensions[1];            // Compute the number of pixels in this image
      }
      if (nDims==3)
      { ImageDimensions = mxGetDimensions(prhs[1]);
        nLayers         = ImageDimensions[2];
        nPixels         = ImageDimensions[0]*ImageDimensions[1];
      }
    }

// ------ GET INPUTS AND CREATE THE OUTPUTS ------

    if (mxGetClassID((mxArray *)prhs[1]) == mxDOUBLE_CLASS){
      imageXdouble = mxGetPr(prhs[1]);
      doubleflag = 1;
    }
    if (mxGetClassID((mxArray *)prhs[1]) == mxUINT8_CLASS){
      imageX  = (unsigned char *)mxGetData(prhs[1]);                        // pointer to the first element of the first input: must use mxGetData for unsigned char
      doubleflag = 0;
    }

    plhs[0] = mxCreateDoubleMatrix(1,nLayers,mxREAL);                       // output pointer 1 to the sum of X
    sumX    = mxGetPr(plhs[0]);
    plhs[1] = mxCreateDoubleMatrix(1,nLayers,mxREAL);                       // output pointer 2 to the mean of X
    meanX   = mxGetPr(plhs[1]);

    if (nlhs >= 3){
      plhs[2] = mxCreateDoubleMatrix(1,nLayers,mxREAL);                     // output pointer 3 to the minimum value in each layer of X
      minX    = mxGetPr(plhs[2]);
      plhs[3] = mxCreateDoubleMatrix(1,nLayers,mxREAL);                     // output pointer 4 to the maximum value in each layer of X
      maxX    = mxGetPr(plhs[3]);
      plhs[4] = mxCreateDoubleMatrix(1,nLayers,mxREAL);                     // output pointer 5 to the variance of X
      varX    = mxGetPr(plhs[4]);
    }


// ------ DO THE CALCULATIONS ------

    if (doubleflag==1){
      for (j=0;j<nLayers;j++){
        sumX[j] = 0;                                                        // Initialize (even if Matlab does it)
        for (k=0;k<nPixels;k++){
          sumX[j] += imageXdouble[k+j*nPixels];
        }
        meanX[j] = sumX[j] / nPixels;
      }//end: for j

      if (nlhs >= 3){
        for (j=0;j<nLayers;j++) {
            minX[j] = (double)imageXdouble[j*nPixels];                      // Set the first value as being both the minimum and maximum value
            maxX[j] = (double)imageXdouble[j*nPixels];
            varX[j] = 0;
            for (k=0;k<nPixels;k++) {                                       // Scan through all pixels to find the minimum and maximum
              if (minX[j] > imageXdouble[j*nPixels+k])
                  minX[j] = (double)imageXdouble[j*nPixels+k];
              if (maxX[j] < imageXdouble[j*nPixels+k])
                  maxX[j] = (double)imageXdouble[j*nPixels+k];
              varX[j] += (imageXdouble[k+j*nPixels] - meanX[j]) * (imageXdouble[k+j*nPixels] - meanX[j]);
            }
            varX[j] /= (nPixels - 1);
        }//end: for j
      }//end: if nlhs
    }//end: if (doubleflag==1)

    if (doubleflag==0){
      for (j=0;j<nLayers;j++){
        sumX[j] = 0;                                                        // Initialize (even if Matlab does it)
        for (k=0;k<nPixels;k++){
          sumX[j] += imageX[k+j*nPixels];
        }
        meanX[j] = sumX[j] / nPixels;
      }//end: for j
      if (nlhs >= 3){
        for (j=0;j<nLayers;j++) {
            minX[j] = (double)imageX[j*nPixels];                            // Set the first value as being both the minimum and maximum value
            maxX[j] = (double)imageX[j*nPixels];
            varX[j] = 0;
            for (k=0;k<nPixels;k++) {                                       // Scan through all pixels to find the minimum and maximum
              if (minX[j] > imageX[j*nPixels+k])
                  minX[j] = (double)imageX[j*nPixels+k];
              if (maxX[j] < imageX[j*nPixels+k])
                  maxX[j] = (double)imageX[j*nPixels+k];
              varX[j] += (imageX[k+j*nPixels] - meanX[j]) * (imageX[k+j*nPixels] - meanX[j]);
            }
            varX[j] /= (nPixels - 1);
        }//end: for j
      }//end: if nlhs
    }//end: if (doubleflag==1)

return;
}


/*--------------------------------------------------------------------------------------------------------------------------------------------
[3] ImagePCA:               [2 inputs]   imageX                                              (nRows x nCols x nLayers)
                                         Number of principal components to extract           (nPCs)
                            [11 outputs] The XTX matrix                                      (nLayers x nLayers)
                                         The loadings matrix                                 (nLayers x nLayers)
                                         The eigenvalue matrix                               (nLayers x nLayers)
                                         The scores as images                                (nRows x nCols x nPCs)
                                         The minimum score value before scaling              (1 x nPCs)
                                         The maximum score value before scaling              (1 x nPCs)
                                         The score scaling vector                            (1 x nPCs)

    AIM: Will extract the number of principal components as desired by the user (see matrix dimensions above).
         One cannot extract nPCs > nLayers and a check is done for this condition.

         The code is self explanatory and consists mostly of matrix multiplication in a way as to avoid a large memory overhead
         This function is very much faster than Matlab for large images (say 1024 x 1024 x 4) because it avoids the creation of very
         large matrices in memory.  It is approximately of the same time order (still less memory though) for smaller images with
         say a total of 1E6 pixels.  This function calls Matlab's SVD.M function to compute the singular value decomposition, to
         guarentee that results are always consistent between this function and by doing the same thing in Matlab.

         Bug fixed in this function on 5 March 2004.
*/
void ImagePCA(mxArray *plhs[],const mxArray *prhs[],int nlhs)
{
    double        *XTXmatrix,                   // Output  1: The X transpose X matrix of the unfolded image
                  *Loadings,                    // Output  2: A matrix (nLayers x nLayers) of loading vectors [ p_1 ; p_2 ; ... ; p_N]'
                  *Eigenvalues,                 // Output  3: A matrix (nLayers x nLayers) containing eigenvalues on the diagonal
                  *TScores,                     // Output  4: An array (ImageDimension) containing the TScores as images
                  *TScoreMin,                   // Output  5: A matrix (1 x nLayers) containing the smallest TScore value from every layer
                  *TScoreMax,                   // Output  6: A matrix (1 x nLayers) containing the largest TScore value from every layer
                  *TScoreScale;                 // Output  7: A matrix (1 x nLayers) containing the scaling factor used to generate TScore images

    unsigned char *imageX;                      // Input 1  : An array (of type unsigned char) containing the uint8 input image
    double        *imageXdouble;                // Input 1  : Same as above for when the input image is of type "double"
    int            nPCs;                        // Input 2  : A variable indicating the number of principal components to extract

    double         *InputPCs,                   // Working: The number of principal components the user wishes to extract
                   doubleflag;                  // Working: A 0-1 variable indicating if the input image is of type double(1) or uint8(0)
    int            ImageRows,                   // Working: The number of rows in the image (number of horizontal pixel lines)
                   ImageCols,                   // Working: The number of columns in the image (number of vertical pixel lines)
                   nLayers,                     // Working: The number of layers in the image (eg: 3 for RGB images)
                   nPixels,                     // Working: The number of pixels per layer in the image
                   i,j,k,                       // Working: Index variables for looping (mainly matrix multiplication)
                   TScoreDimensions[3],         // Working: A matrix to specify the dimensions of the scores as images
                  *ImageDimensions;             // Working: A pointer to the image size
    mxArray       *SVDCallLHS[3],               // Working: The output array of pointers for the Matlab SVD call
                  *SVDCallRHS[1];               // Working: The input array of pointers for the Matlab SVD call


    ImageDimensions  = mxGetDimensions(prhs[1]);                            // Get the dimensions of the input image
    ImageRows        = ImageDimensions[0];
    ImageCols        = ImageDimensions[1];                                  // Get descriptive names for the image dimensions
    nLayers          = ImageDimensions[2];
    nPixels          = ImageDimensions[0]*ImageDimensions[1];               // Compute the number of pixels in this image

    if (mxGetClassID((mxArray *)prhs[1]) == mxDOUBLE_CLASS){
      imageXdouble   = mxGetPr(prhs[1]);
      doubleflag     = 1;
    }
    if (mxGetClassID((mxArray *)prhs[1]) == mxUINT8_CLASS){
      imageX         = (unsigned char *)mxGetData(prhs[1]);                 // pointer to the first element of the first input
      doubleflag     = 0;
    }

    InputPCs         = mxGetPr(prhs[2]);
    if (InputPCs[0] <1) {
      mexErrMsgTxt("libmia error in function ImagePCA: The number of principal components to extract may not be less than one.\n"); }
    if (InputPCs[0] > nLayers) {
      mexErrMsgTxt("libmia error in function ImagePCA: The number of principal components to extract may not exceed the number of layers in the image.\n");  }
    nPCs             = (int)InputPCs[0];

    TScoreDimensions[0] = ImageDimensions[0];
    TScoreDimensions[1] = ImageDimensions[1];
    TScoreDimensions[2] = nPCs;                                             // Change the 3rd image dimension = nPCs to extract

    plhs[0] = mxCreateDoubleMatrix(nLayers,nLayers,mxREAL);                 // Set output pointer 1  to the XTX matrix
    plhs[1] = mxCreateDoubleMatrix(nLayers,nLayers,mxREAL);                 // Set output pointer 2  to the loadings matrix
    plhs[2] = mxCreateDoubleMatrix(nLayers,nLayers,mxREAL);                 // Set output pointer 3  to the eigenvalue matrix
    plhs[3] = mxCreateNumericArray(3,TScoreDimensions,mxDOUBLE_CLASS,mxREAL);// Set output pointer 4  to the scores as images
    plhs[4] = mxCreateDoubleMatrix(1,nPCs,mxREAL);                          // Set output pointer 5  to the minimum score value before scaling
    plhs[5] = mxCreateDoubleMatrix(1,nPCs,mxREAL);                          // Set output pointer 6  to the maximum score value before scaling
    plhs[6] = mxCreateDoubleMatrix(1,nPCs,mxREAL);                          // Set output pointer 7  to the score scaling vector

    XTXmatrix    = mxGetPr(plhs[0]);                                        // Create C pointers to copies of the output matrices
    TScores      = mxGetPr(plhs[3]);
    TScoreMin    = mxGetPr(plhs[4]);
    TScoreMax    = mxGetPr(plhs[5]);
    TScoreScale  = mxGetPr(plhs[6]);

    if (doubleflag==1)
      doubleXTXcalc(imageXdouble, XTXmatrix, nLayers, nPixels);
    if (doubleflag==0)
      unsignedcharXTXcalc(imageX, XTXmatrix, nLayers, nPixels);

    SVDCallRHS[0] = plhs[0];                                                // A pointer to the XTXmatrix
    mexCallMATLAB(3, SVDCallLHS, 1, SVDCallRHS, "svd");                     // Call the SVD routine
    plhs[1]       = SVDCallLHS[0];                                          // Assign the loadings to the first SVD output
    plhs[2]       = SVDCallLHS[1];                                          // and the eigenvalues to the second SVD output
    Loadings      = mxGetPr(plhs[1]);
    Eigenvalues   = mxGetPr(plhs[2]);                                       // and then give them nice names to work with later


    if (doubleflag==1){
      for (i=0;i<nPCs;i++){                                                 // Now calculate the T-scores from T = XP
        for (j=0;j<nLayers;j++){
          for (k=0;k<nPixels;k++){
            // This next line had the bug: the last counter, nLayers, was originally nPCs.  The result of this bug was
            // that if the user extracted fewer principal components than layers in the image (if nPCs < nLayer) then,
            // the T-scores 2, 3, ... nPCs were incorrectly computed.  This caused score plots to have a very correlated
            // and distorted shape. This bug was picked up a while ago, but fixed on 5 March 2004.
            TScores[k+i*nPixels] +=  imageXdouble[k+j*nPixels] * Loadings[j+i*nLayers];
          }
        }
      }//end: for i
    }
    if (doubleflag==0){                                                     // As above, except using type "uint8"
      for (i=0;i<nPCs;i++){
        for (j=0;j<nLayers;j++){
          for (k=0;k<nPixels;k++){
            TScores[k+i*nPixels] +=  imageX[k+j*nPixels] * Loadings[j+i*nLayers];
          }
        }
      }//end: for i
    }

    for (i=0;i<nPCs;i++)                                                    // Compute the minimum and maximum T-score values
    {
      TScoreMin[i] = TScores[i*nPixels];
      TScoreMax[i] = TScores[i*nPixels];
      for (k=0;k<nPixels;k++)
      {
        if (TScoreMin[i] > TScores[i*nPixels+k])
        {   TScoreMin[i] = TScores[i*nPixels+k]; }
        if (TScoreMax[i] < TScores[i*nPixels+k])
        {   TScoreMax[i] = TScores[i*nPixels+k]; }
      }
    }

    for (i=0;i<nPCs;i++)                                                    // Compute the T-score scaling factors
    {
      if (TScoreMin[i] == TScoreMax[i])                                     // Avoids division by zero and automatically
        TScoreScale[i] = 0;                                                 // handles the case when nPC < nLayers
      else
        TScoreScale[i] = 255/(TScoreMax[i]-TScoreMin[i]);

      for (k=0;k<nPixels;k++)                                               // So now scale the T-score matrix
      { TScores[i*nPixels+k] = (unsigned char)(( TScores[i*nPixels+k]-TScoreMin[i] ) * (TScoreScale[i]));      }
    }

  return;                                                                   // End of sub-function
}



/*--------------------------------------------------------------------------------------------------------------------------------------------
[4] ScorePlot:              [2 inputs]   T-score image that is on the horizontal axis   (TScoreHoriz);
                                         T-score image that is on the vertical axis     (TScoreVrtcl)
                            [1 output]   An image of the score space (TScorePlot)

    AIM: This function file will compute the score plot image given two columns of the T_Score matrix that have already
         been scaled to be images.  In other words, the two columns must be of type uint8 and have the same number of
         elements.  The first T_Score matrix is assumed to be on the horizontal axis and the second T_Score matrix on
         the vertical axis.  Since this operation is really constucting a histogram with two axes, the axis labelling
         is as appears in the following diagram:

         0.----------------------------.
          |                            |
          |                            |
          |                            |        What the function does is to go down the columns of input 1 and 2 at
          |                            |        the same rate and look at the values which lie between 0 and 255.
          |                            |        It then puts a "1" in the plot at the horizontal and vertical
 Input 2  |                            |        intersection point.  If there is already a number there, it increments
          |                            |        it by one.
          |                            |
          |                            |        At the end of going through both the input vectors we are left with a
          |                            |        "terrain map" of the two vectors.  Vectors with the same features
          |                            |        (ie numerical number) appear in the same space.
          |                            |
       255.----------------------------.
           0                          255
                     Input 1
*/
void ScorePlot(mxArray *plhs[], const mxArray *prhs[])
{
    unsigned char *TScoreHoriz;                 // Input 1  : the scaled T-score values that are on the horizontal axis         (nRows x nCols) [uint8]
    unsigned char *TScoreVrtcl;                 // Input 2  : the scaled T-score values that are on the vertical axis           (nRows x nCols) [uint8]
    double        *TScoreImage;                 // Output 1 : the T-score plot                                                  (256 x 256)     [double]

    int            a, b, k, nPixels;            // Working  : integers holding temporary results and values
    int            nHRows, nHCols;              // Working  : integers describing the size of the first input (horizontal)      (1 x 1) [int] each
    int            nVRows, nVCols;              // Working  : integers describing the size of the second input (vertical)       (1 x 1) [int] each

    nHRows = mxGetM(prhs[1]);                                               // Get the dimensions of the first input
    nHCols = mxGetN(prhs[1]);
    nVRows = mxGetM(prhs[2]);                                               // Get the dimensions of the second input
    nVCols = mxGetN(prhs[2]);

    if (nHRows != nVRows)                                                   // Checks that the two inputs are of the same size
       {mexErrMsgTxt("libmia error in function ScorePlot: the two score image inputs must have the same number of rows.\n");}
    if (nHCols != nVCols)
       {mexErrMsgTxt("libmia error in function ScorePlot: the two score image inputs must have the same number of columns.\n");}

    nPixels     = nHRows*nHCols;
    TScoreHoriz = (unsigned char *)mxGetData(prhs[1]);                      // Get the actual input data: horizontal axis T_Score
    TScoreVrtcl = (unsigned char *)mxGetData(prhs[2]);                      // Get the actual input data: vertical axis T_Score
    plhs[0]     = mxCreateDoubleMatrix(256,256,mxREAL);                     // Set the output pointer to the output matrix and
    TScoreImage = mxGetPr(plhs[0]);                                         // get a pointer to the first element

// ------ DO THE CALCULATIONS ------
    for(k=0;k<nPixels;k++){                                                 // NOW DO THE ACTUAL WORK
       a = TScoreHoriz[k];                                                  // Go down the vectors at the same rate and
       b = TScoreVrtcl[k];                                                  // put "one" (as in the number) at the intersection
       TScoreImage[a*256+b]++;                                              // or increment if there is already a number there.
    }
  return;
}







/*--------------------------------------------------------------------------------------------------------------------------------------------
[5] Mask2Im:                [5 inputs]   T-score image that is on the horizontal axis                                   (TScoreHoriz)
                                         T-score image that is on the vertical axis                                     (TScoreVrtcl)
                                         Mask over the T_horiz - T_vrtcl score space                                    (Mask)
                                         The fill colour (or fill data) applied to image pixels falling under the mask  (Fill)
                                         The input image can optionally be supplied                                     (imageX)
                            [2 outputs]  The image space with the pixels falling under mask replaced by the fill        (MaskedImage)
                                         The number of pixel replacements made.                                         (nPixelsMasked)

    USAGE:   1.   [MaskedImage,nPixelsMasked] = mask2im(tHoriz,tVert,Mask)
             2.   [MaskedImage,nPixelsMasked] = mask2im(tHoriz,tVert,Mask,Fill)
             3.   [MaskedImage,nPixelsMasked] = mask2im(tHoriz,tVert,Mask,Fill,OriginalImage)

    AIM: Conceptually this function would first form the score plot (a 256x256 plot) created by the first two inputs.  The third
         input, Mask, would then be overlaid on this score plot.  The variable "Mask" contains zero and non-zero entries.  Non-zero
         entries are considered to form the mask, while zero entries are outside the mask boundary.  The mask region does not need
         to be connected in any way.

         Since the first two inputs have the same row and column dimensions as the original image space, we know which image-space
         pixels fall inside the mask area.  When this function is used as shown in usage form 1, then the output, "MaskedImage" is
         a variable the same size as the first two inputs containing 0 and 1 elements: 0=pixel did not fall under mask, 1=it did.

         The second usage form will create an output "MaskedImage" whose third dimension size is equal to length(Fill).  This allows,
         for example, the output to be an RGB image.  So if fill = [1 0 1], then pixels in the image space falling under the mask
         will have a magenta/purple colour, while pixels outside the mask will have zero entry.

         The last usage form allows the pixels outside the mask to be replaced by pixels from the original image.  Pixels inside the
         mask are replaced by the elements in "Fill".  If the input image is of type "double", then "Fill" will be used as provided.
         If the input image is of type "uint8" then "Fill" should be entered as a fracional value, since it will be multiplied by 256
         and a conversion to uint8 will be made.  This last usage form allows one to view the masked pixels in the context of
         the original image.  This form requires that length(Fill) = size(OriginalImage,3).

         The number of replacements pixels falling under the mask can optionally be returned at the end of the algorithm.

         [The algorithm's implementation does not follow the above description, rather if proceeds very similar to the algorithm
          to compute the ScorePlot (see the MEX functions in this file).  Take the two T-scores and go down them at equal speed
          looking at the values in these vectors (they are actually matrices, but we take advantage of Matlab's columnwise
          reading of matrices and see them then as vectors).  At each element down T_horiz and T_vert we can compute where that
          combination would have fallen in the score plot (see the previous MEX function).  If that position is currently under
          the mask, then since the point down the T-scores that we are currently at corresponds exactly to its point in the image
          space (T-scores have the same dimension as the image space) we colour that point using the "Fill" colour in the image
          space.  If the point is not under the mask, then just just duplicate that pixel of the image space to the output array.]



*/
void Mask2Im(mxArray *plhs[], const mxArray *prhs[],int nrhs)            // Mask to image
{
    unsigned char *TScoreHoriz;                 // Input 1  : the scaled T-score values that are on the horizontal axis         (nRows x nCols) [uint8]
    unsigned char *TScoreVrtcl;                 // Input 2  : the scaled T-score values that are on the vertical axis           (nRows x nCols) [uint8]
    double        *Mask;                        // Input 3  : a binary mask over the score space: 1=in mask, 0=outside mask     (256 x 256)     [double]
    double        *Fill;                        // Input 4  : the desired "fill"                                                (1 x nLayers)   [double]
    unsigned char *imageX;                      // Input 5  : the original image (optional)                                     (nRows x nCols x nLayers) [uint8]
    double        *imageXdouble;                // Input 5  : alternative to above: the original image (optional)               (nRows x nCols x nLayers) [uint8]

    double        *MaskedImage;                 // Output 1 : the resulting image from masking in the score space               (nRows x nCols x nLayers)
    unsigned char *MaskedImage8;                // Output 1 : alternative (if supplied input image (input 5) was uint8)         (nRows x nCols x nLayers)
    double        *nPixelsMasked;               // Output 2 : the number of pixels from input image found inside mask area      (1 x 1)         [double]

    double         doubleflag;                  // Working  : a 0-1 variable indicating if the input image is of type double(1) or uint8(0)
    int            a, b, j, k;                  // Working  : integers loop counters                                            (1 x 1) [int] each
    int            nPixels, nLayers;            // Working  : integers that represent matrix sizes                              (1 x 1) [int] each
    int            nHRows, nHCols;              // Working  : integers describing the size of the first input (horizontal)      (1 x 1) [int] each
    int            nVRows, nVCols;              // Working  : integers describing the size of the second input (vertical)       (1 x 1) [int] each
    int            nFRows, nFCols;              // Working  : integers describing the size of the optional fourth input         (1 x 1) [int] each
    int            nXRows, nXCols;              // Working  : integers describing the size of the optional fifth input          (1 x 1) [int] each
    const int     *DimImage;                    // Working  : numbers describing the input image dimensions
    int            DimOutput[3];                // Working  : used to form the output image size when only "Fill" is given      (1 x 3) [int]


// ------ GET INPUTS and ERROR CHECKING ------

    nrhs = nrhs - 1;                                                        // The "first input" is a text string, ignore this

    if (mxGetNumberOfDimensions(prhs[1])>2)
      mexErrMsgTxt("libmia error in function Mask2Im: the first score image input should be a 2-dimensional matrix.\n");
    if (mxGetNumberOfDimensions(prhs[2])>2)
      mexErrMsgTxt("libmia error in function Mask2Im: the second score image input should be a 2-dimensional matrix.\n");

    nHRows = mxGetM(prhs[1]);                                               // Get the dimensions of the first input
    nHCols = mxGetN(prhs[1]);
    nVRows = mxGetM(prhs[2]);                                               // Get the dimensions of the second input
    nVCols = mxGetN(prhs[2]);

    TScoreHoriz   = (unsigned char *)mxGetData(prhs[1]);                    // Get the horizontal axis's T-scores
    TScoreVrtcl   = (unsigned char *)mxGetData(prhs[2]);                    // Get the vertical axis's T-scores
    Mask          =                  mxGetPr(prhs[3]);                      // Get the pointer to the binary mask

    if (nrhs>=4){
      if (mxGetNumberOfDimensions(prhs[4])>2)
        mexErrMsgTxt("libmia error in function Mask2Im: the variable 'Fill' should be a 2-dimensional row or column vector.\n");
      nFRows = mxGetM(prhs[4]);
      nFCols = mxGetN(prhs[4]);
      if (  (nFRows!=1)&&(nFCols!=1)  )
        mexErrMsgTxt("libmia error in function Mask2Im: the variable 'Fill' should be a row or column vector.\n");
      nLayers = nFRows;
      if (nFCols > nLayers)
        nLayers = nFCols;
      Fill = mxGetPr(prhs[4]);                                              // Get the pointer to the desired fill - type "double"
    }

    if (nrhs>=5){
      if (mxGetClassID((mxArray *)prhs[5]) == mxDOUBLE_CLASS){
        imageXdouble   = mxGetPr(prhs[5]);
        doubleflag     = 1;
      }
      if (mxGetClassID((mxArray *)prhs[5]) == mxUINT8_CLASS){
        imageX         = (unsigned char *)mxGetData(prhs[5]);               // Get the image space
        doubleflag     = 0;
      }
      DimImage = mxGetDimensions(prhs[5]);                                  // Get the dimensions of the input image: RETURNS A "const int" - we cannot change its value
      nXRows = DimImage[0];
      nXCols = DimImage[1];

      if (mxGetNumberOfDimensions(prhs[5])>3)
         mexErrMsgTxt("libmia error in function Mask2Im: the image input must be a 2-D or 3-D image.\n");
      if ( (mxGetNumberOfDimensions(prhs[5])>2)&&(nLayers!=DimImage[2]) )   // nLayers was computed above [see "if (nrhs>=4)" section]
         mexErrMsgTxt("libmia error in function Mask2Im: the image input must have the same number of layers as the 'Fill' vector input.\n");
      if ( (mxGetNumberOfDimensions(prhs[5])==2)&&(nLayers!=1) )            // Input image is also "flat"
         mexErrMsgTxt("libmia error in function Mask2Im: the image input must have the same number of layers as the 'Fill' vector input.\n");
      if (nHRows != nXRows)                                                 // Check the optional fifth input
         mexErrMsgTxt("libmia error in function Mask2Im: the image input must have the same number of rows as the first two score image inputs.\n");
      if (nHCols != nXCols)
         mexErrMsgTxt("libmia error in function Mask2Im: the image input must have the same number of columns as the first two score image inputs.\n");
    }

    if (nHRows != nVRows)                                                   // Checks that the two inputs are of the same size
       mexErrMsgTxt("libmia error in function Mask2Im: the two score image inputs must have the same number of rows.\n");
    if (nHCols != nVCols)
       mexErrMsgTxt("libmia error in function Mask2Im: the two score image inputs must have the same number of columns.\n");


// ------ CREATE EMPTY OUTPUTS ------

    nPixels = nXRows*nXCols;
    if (nrhs==3)
      plhs[0] = mxCreateDoubleMatrix(nHRows,nHCols,mxREAL);

    if (nrhs==4){                                                           // User has specified a fill: scalar or vector ?
      if (nLayers==1)
        plhs[0] = mxCreateDoubleMatrix(nHRows,nHCols,mxREAL);               // Scalar case ===> output is 2-D matrix
      if (nLayers>1){
        DimOutput[0] = nHRows;
        DimOutput[1] = nHCols;
        DimOutput[2] = nLayers;
        plhs[0] = mxCreateNumericArray(3, DimOutput, mxDOUBLE_CLASS, mxREAL);
      }
    }

    if (nrhs==5){                                                           // Create output depending on 5th input's type
      if (mxGetClassID((mxArray *)prhs[5]) == mxDOUBLE_CLASS)
        plhs[0] = mxCreateNumericArray(mxGetNumberOfDimensions(prhs[5]), mxGetDimensions(prhs[5]), mxDOUBLE_CLASS, mxREAL);
      if (mxGetClassID((mxArray *)prhs[5]) == mxUINT8_CLASS)
        plhs[0] = mxCreateNumericArray(mxGetNumberOfDimensions(prhs[5]), mxGetDimensions(prhs[5]), mxUINT8_CLASS,  mxREAL);
    }

    if (doubleflag==1)
      MaskedImage = mxGetPr(plhs[0]);
    if (doubleflag==0)
      MaskedImage8 = (unsigned char *)mxGetData(plhs[0]);                   // create pointer to first element in output

    plhs[1] = mxCreateDoubleMatrix(1,1,mxREAL);                             // Get pointer to second output
    nPixelsMasked = mxGetPr(plhs[1]);


// ------ DO THE CALCULATIONS ------

    if (doubleflag==1){
      if (nrhs==5){
        for(k=0;k<nPixels;k++){
          a = (int)TScoreHoriz[k];                                          // Go down the vectors at the same rate
          b = (int)TScoreVrtcl[k];
          if (Mask[a*256+b] == 1){
            nPixelsMasked[0]++;                                             // If the current pixel lies under the mask
            for(j=0;j<nLayers;j++)
              MaskedImage[k+j*nPixels] = Fill[j];                           // then replace it with the desired fill data
          }
          else{
            for(j=0;j<nLayers;j++)
              MaskedImage[k+j*nPixels] = imageXdouble[k+j*nPixels];         // otherwise, just put the original image pixel there
          }
        }//end: for k
      }//end: if nrhs==5
      if (nrhs==4){
        for(k=0;k<nPixels;k++){
          a = (int)TScoreHoriz[k];                                          // Go down the vectors at the same rate
          b = (int)TScoreVrtcl[k];
          if (Mask[a*256+b] == 1){
            nPixelsMasked[0]++;                                             // If the current pixel lies under the mask
            for(j=0;j<nLayers;j++)
              MaskedImage[k+j*nPixels] = Fill[j];                           // then replace it with the desired fill data
          }
          else{
            for(j=0;j<nLayers;j++)
              MaskedImage[k+j*nPixels] = 0;                                 // otherwise, just put a zero there
          }
        }//end: for k
      }//end: if nrhs==4
    }//end: if doubleflag==1


    if (doubleflag==0){
      if (nrhs==5){
        for(k=0;k<nPixels;k++){
          a = (int)TScoreHoriz[k];                                          // Go down the vectors at the same rate
          b = (int)TScoreVrtcl[k];
          if (Mask[a*256+b] == 1){
            nPixelsMasked[0]++;                                             // If the current pixel lies under the mask
            for(j=0;j<nLayers;j++)
              MaskedImage8[k+j*nPixels] = (unsigned char)(255*Fill[j]);     // then replace it with the desired fill data
          }
          else{
            for(j=0;j<nLayers;j++)
              MaskedImage8[k+j*nPixels] = imageX[k+j*nPixels];              // otherwise, just put the original image pixel there
          }
        }//end: for k
      }//end: if nrhs==5
      if (nrhs==4){
        for(k=0;k<nPixels;k++){
          a = (int)TScoreHoriz[k];                                          // Go down the vectors at the same rate
          b = (int)TScoreVrtcl[k];
          if (Mask[a*256+b] == 1){
            nPixelsMasked[0]++;                                             // If the current pixel lies under the mask
            for(j=0;j<nLayers;j++)
              MaskedImage8[k+j*nPixels] = Fill[j];                          // then replace it with the desired fill data
          }
          else{
            for(j=0;j<nLayers;j++)
              MaskedImage8[k+j*nPixels] = 0;                                // otherwise, just put a zero there
          }
        }//end: for k
      }//end: if nrhs==4
    }//end: if doubleflag==0
return;
}



/*--------------------------------------------------------------------------------------------------------------------------------------------
[6] Residual:               [2 inputs]   Original image                 (imageX)            [uint8 or double]
                                         Residual operator              (ResidOperator)     [double]

                            [2 outputs]  The residual image             (ResidImage)        [uint8]
                                         Largest residual found         (maxValue)          [double]

    USAGE:        [ResidImage,maxValue] = residual(imageX,ResidOperator)

    AIM:  Uses the matrix form of the algorithm supplied in Geladi and Grahn, p 183.  To compute the residual image after "n" principal
          components have been extracted (n <= number of layers in image=nLayers).  To do this we use the original image, reshaped into
          a 2-D called XReshaped, and the loading matrix P.  We first construct the residual operator, ResidOperator, in Matlab:
                  ResidOperator = eye(nLayers) - P(:,1:n) * P(:,1:n)';              % Note: n <= N and also note the transpose on the last matrix
                  ResidImage    = sqrt(sum( (XReshaped*ResidOperator) .* (XReshaped*ResidOperator) , 2 ));
                  ResidImage    = uint8(reshape(ResidImage,nRows,nCols,nLayers));

          The last two lines of Matlab code are performed in this MEX library, because we have to first reshape and convert and the
          3-way image to a 2-way matrix in Matlab's "double" type, which uses a lot of memory for large (> 1E6 pixels) images.

          The output image is of type "uint8" always so that it can be displayed with a "jet" colourmap.  We have not encountered cases
          where the "DModX" or residual values are so high that they require a larger output form, such as "uint16" or "double"
*/

void Residual(mxArray *plhs[], const mxArray *prhs[],int nrhs)
{
    unsigned char *imageX;                      // Input 1  : An array (of type unsigned char) containing the uint8 input image
    double        *imageXdouble;                // Input 1  : Same as above for when the input image is of type "double"
    double        *ResidOperator;               // Input 2  : the residual operator

    unsigned char *ResidImage;                  // Output 1 : the residual image (DModX image if you will)
    double        *maxValue;                    // Output 2 : the maximum residual value found in the image

    int            i, j, k;
    int            nPixels,nLayers,ResidDims[2];// Working:  integers holding temporary results and values
    int           *ImageDimensions;             // Working:  a pointer to the image size
    int           *ResidOperatorDims;           // Working:  a pointer to the residual operator's size
    int            doubleflag;                  // Working:  indicates if input is double(1) or uint8(0)
    double         temp1, temp2, MaxValueSoFar; // Working:  temporary variables of double precision type


// ------ GET INPUTS and ERROR CHECKING ------
    ImageDimensions   = mxGetDimensions(prhs[1]);                             // Get the dimensions of the input image
    ResidOperatorDims = mxGetDimensions(prhs[2]);
    nPixels         = ImageDimensions[0]*ImageDimensions[1];
    nLayers         = ImageDimensions[2];
    ResidDims[0]    = ImageDimensions[0];
    ResidDims[1]    = ImageDimensions[1];

    if (ResidOperatorDims[0] != ResidOperatorDims[1]) {
      mexErrMsgTxt("libmia error in function Residual: The residual operator matrix must be a square matrixnumber of principal components to extract may not be less than one.\n"); }
    if (ResidOperatorDims[0] != nLayers) {
      mexErrMsgTxt("libmia error in function Residual: The residual operator matrox must have as many rows as the number of layers in the image.\n");  }

    if (mxGetClassID((mxArray *)prhs[1]) == mxDOUBLE_CLASS){
      imageXdouble   = mxGetPr(prhs[1]);
      doubleflag     = 1;
    }
    if (mxGetClassID((mxArray *)prhs[1]) == mxUINT8_CLASS){
      imageX         = (unsigned char *)mxGetData(prhs[1]);                 // Get the image data
      doubleflag     = 0;
    }
    ResidOperator   = mxGetPr(prhs[2]);                                     // Get the residual operator

    plhs[0]         = mxCreateNumericArray(2,ResidDims,mxUINT8_CLASS,mxREAL);
    plhs[1]         = mxCreateDoubleMatrix(1,1,mxREAL);                     // Set the output pointers to the two outputs
    ResidImage      = (unsigned char *)mxGetData(plhs[0]);
    maxValue        = mxGetPr(plhs[1]);

    MaxValueSoFar = 0.0;
    if (doubleflag==1){
      for(i=0;i<nPixels;i++){
         temp2 = 0.0;
         for(j=0;j<nLayers;j++){
            temp1 = 0.0;
            for (k=0;k<nLayers;k++){
              temp1 += imageXdouble[i+k*nPixels]*ResidOperator[k+j*nLayers];
            }
            temp2 += temp1 * temp1;                                         // Sum of squares value
         }
         ResidImage[i] = (int)(sqrt(temp2));                                // Convert to an integer and store it in the Residual Image
          if (ResidImage[i] > MaxValueSoFar)                                // Find the largest residual value so far
             MaxValueSoFar = ResidImage[i];
      }
      maxValue[0] = MaxValueSoFar;
    }

    if (doubleflag==0){
      for(i=0;i<nPixels;i++){
         temp2 = 0.0;
         for(j=0;j<nLayers;j++){
            temp1 = 0.0;
            for (k=0;k<nLayers;k++){
              temp1 += imageX[i+k*nPixels]*ResidOperator[k+j*nLayers];
            }
            temp2 += temp1 * temp1;                                         // Sum of squares value
         }
         ResidImage[i] = (int)(sqrt(temp2));                                // Convert to an integer and store it in the Residual Image
          if (ResidImage[i] > MaxValueSoFar)                                // Find the largest residual value so far
             MaxValueSoFar = ResidImage[i];
      }
      maxValue[0] = MaxValueSoFar;
    }


  return;
}




//------------------ MAIN FUNCTION ----------------------------------------------------------------------------------------------------------------
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
    char *FunctionToExecute;            // Determine which function to execute
    int   FunctionNumber;               // Determine the function number based on the first input's string
    int   k;                            // An index
    char  DefaultFunctionName[5]="help";

    if (nrhs > 0){
        if(!mxIsChar(prhs[0]))
	        mexErrMsgTxt("libmia is called internally from maccmia.\n");
        if (mxGetM(prhs[0])!=1)
            mexErrMsgTxt("libmia is called internally from maccmia.\n");

        FunctionToExecute = mxArrayToString(prhs[0]);                       // what about using mxGetString ?????
        for (k=0;k<mxGetNumberOfElements(prhs[0]);k++)
        {  if(isalnum(FunctionToExecute[k]))                                // Check if the input string is alphanumeric
            FunctionToExecute[k] = tolower(FunctionToExecute[k]);
        else
            mexErrMsgTxt("libmia is called internally from maccmia.\n");
        }
    }
    else{
        FunctionToExecute = &DefaultFunctionName[0];
    }

    FunctionNumber = 0;
    if (!strcmp(FunctionToExecute,"help")                   && (nrhs==1) )  // Note the nrhs counts both the text input and then the remaining inputs
        FunctionNumber = 0;
    if (!strcmp(FunctionToExecute,"xtx")                    && (nrhs==2) )
        FunctionNumber = 1;
    if (!strcmp(FunctionToExecute,"stats")                  && (nrhs==2) )
        FunctionNumber = 2;
    if (!strcmp(FunctionToExecute,"imagepca")               && (nrhs==3) )
        FunctionNumber = 3;
    if (!strcmp(FunctionToExecute,"scoreplot")              && (nrhs==3) )
        FunctionNumber = 4;
    if (!strcmp(FunctionToExecute,"mask2im")                && (nrhs>=4) )
        FunctionNumber = 5;
    if (!strcmp(FunctionToExecute,"residual")               && (nrhs==3) )
        FunctionNumber = 6;
    if (!strcmp(FunctionToExecute,"ver")                    && (nrhs==1) )
        FunctionNumber = 100;

    switch(FunctionNumber) {
        case 0 :    mexErrMsgTxt("libmia is called internally from maccmia.\n");

        case 1 : //ComputeXTX function is being called
                if (   (mxGetClassID((mxArray *)prhs[1]) != mxUINT8_CLASS)&&(mxGetClassID((mxArray *)prhs[1]) != mxDOUBLE_CLASS)   )
                    mexErrMsgTxt("libmia error in function ComputeXTX: image must be of type 'uint8' or of type 'double'.\n");
                ComputeXTX(plhs,prhs);
                break;

        case 2 : //ImageStats function is being called
                if (   (mxGetClassID((mxArray *)prhs[1]) != mxUINT8_CLASS)&&(mxGetClassID((mxArray *)prhs[1]) != mxDOUBLE_CLASS)   )
                    mexErrMsgTxt("libmia error in function ImageStats: image must be of type 'uint8' or of type 'double'.\n");
                ImageStats(plhs,prhs,nlhs);
                break;

        case 3 : //ImagePCA function is being called
                if (   (mxGetClassID((mxArray *)prhs[1]) != mxUINT8_CLASS)&&(mxGetClassID((mxArray *)prhs[1]) != mxDOUBLE_CLASS)   )
                    mexErrMsgTxt("libmia error in function ImagePCA: image must be of type 'uint8' or of type 'double'.\n");
                ImagePCA(plhs,prhs,nlhs);
                break;

        case 4 : //ScorePlot function is being called
                if (   (mxGetClassID((mxArray *)prhs[1]) != mxUINT8_CLASS)  ||  (mxGetClassID((mxArray *)prhs[2]) != mxUINT8_CLASS)   )
                    mexErrMsgTxt("libmia error in function ScorePlot: both inputs must be of type 'uint8'.\n");
                ScorePlot(plhs,prhs);
                break;

        case 5 : //Mask2Im function is being called
                if (   (mxGetClassID((mxArray *)prhs[1]) != mxUINT8_CLASS)  ||  (mxGetClassID((mxArray *)prhs[2]) != mxUINT8_CLASS)   )
                    mexErrMsgTxt("libmia error in function Mask2Im: the first two inputs must be of type 'uint8'.\n");
                Mask2Im(plhs,prhs,nrhs);
                break;
        case 6 : //Residual function is being called
                if (   (mxGetClassID((mxArray *)prhs[1]) != mxUINT8_CLASS)&&(mxGetClassID((mxArray *)prhs[1]) != mxDOUBLE_CLASS)   )
                    mexErrMsgTxt("libmia error in function Residual: the image input must be a 'uint8' or 'double' image.\n");
                Residual(plhs,prhs,nrhs);
                break;
        case 100: // Version information
                mexPrintf("libmia information: This is version 1.00 of the libmia library.\n");
                mexPrintf("Author contact:     Kevin G. Dunn: dunnkg@mcmaster.ca          \n");
                mexPrintf("Last updated:       06 June 2010                               \n");
                break;
        }

    if (nrhs > 0){
        mxFree(FunctionToExecute);                                          // Free up the string memory (if used)
    }

}   /* end function */
