function maccmia(action,varargin)

% TO DO LIST: [as at November 2006] 

% Make the GUI re-sizable
% Try a uint16 image

% The McMaster Advanced Control Consortium's Multivariate Image Analysis Package (MACCMIA)
%
% This file consists of a series of callbacks (functions) that are used by
% the Image Analysis Interface.  Executing any of the functions contained
% in this file outside of this interface may have unpredictable results.
%
% It is recommended that you only call this function by only typing the
% command "maccmia" at the MATLAB prompt.  The user interface will take care
% of calling the correct callbacks.
%
% Further, this software is not Freeware and is distributed to the member
% companies of the MACC.  If you are a member of this consortium, feel free
% to use the software within the terms and conditions agreed to by you or your
% company's representative and the MACC.  You may also contact the author at any
% time regarding technical support of the software and to report any bugs.
% Finally suggestions for improvement of the software are gratefully accepted.
%
% Author      : Kevin Dunn (dunnkg@mcmaster.ca)
% Last update : June 2010
% Version     : 1.90
% (c) 2009      McMaster Univeristy 
%               Department of Chemical Engineering,
%               McMaster Advanced Control Consortium
% Webpage     : http://macc.mcmaster.ca/maccmia/

global SmlUD;
if nargin == 0                                                                  % Initialize and launch the GUI interface
    SmlUD = [];
    Initialize_MACCMIA_GUI_Interface;
end;


% --------------------------------------------------------------------
function fig = getGUIHandle(hCallback,tagString)
% Goes up the handle heirarchy until it finds the GUI parent.
  fig = hCallback;
  while ~isempty(fig) && ~isequal(get(fig,'parent'),0) && ~strcmp(get(fig,'tag'),tagString)
    fig = get(fig,'parent');
  end;
return;

% --------------------------------------------------------------------
function MyCLA(varargin)
% This performs a similar job to Matlab's "cla" function, except it performs the action on each input handle
% if the input argument is a vector of handles.  Additionally, it sets the axis to be invisible, because
% this is what the next line of code would invariably be anyway.

  if nargin>0 && all(strcmp(get(varargin{1},'Type'),'axes'))                    % If first argument is a set of axes handles the apply CLA to all these axes
      temp = varargin{1};
      for k = 1:length(temp)
          hHandle = temp(k);
          if ishandle(hHandle)
              hDelete = findobj(allchild(hHandle),'flat','serializable','on');
              delete(hDelete);
              reset(hHandle(ishandle(hHandle)));                                % Do a re-check at the end (in case one of the children deleted the parent)
              set(hHandle,'Visible','off');
          end;
      end;
  end
return;

% --------------------------------------------------------------------
function hImage = ShowImage(ImageData,hAxis,Xcenter,Ycenter,CLim,CDataMapping,varargin)
% Plots the image data, ImageData, in the named axis, hAxis, with the given center points on the
% the X- and Y-axes, Xcenter and Ycenter respectively.   If input CLim is a string, then the image
% is displayed so that the smallest value in the image is CLim(1) and the largest is CLim(2).  If CLim
% is left empty, then it is not adjusted.  The function returns the handle to the image.
%
% If ImageData = [], then the function just adjusts the properties given in varargin.
% In this case, the axis must already contain an image.
%
% Optional inputs may be given in the following order:
%   XLim:           the extent of the axis that is actually displayed: if less than the true image size, then it will appear that the image is zoomed
%   YLim:           as above, except for the Y-axis
%   XTick:          the location of the tick marks on the x-axis
%   YTick:          the location of the tick marks on the y-axis
%   XTickLabel:     the labels displayed at the above tick marks
%   YTickLabel:     the labels displayed at the above tick marks
%   XGrid:          'on' or 'off'
%   YGrid:          'on' or 'off'

  nRows = size(ImageData,1);
  nCols = size(ImageData,2);
  if ishandle(hAxis)
      nargs = max(size(varargin));                                              % The number of extra input arguments
      if isempty(ImageData)                                                     % If no image then return
          return;
      else
          if isempty(Xcenter)
              Xcenter = [0 nCols] + 0.5;                                        % Note that X is the second dimension
          end;
          if isempty(Ycenter)
              Ycenter = [0 nRows] + 0.5;                                        % and Y is the first dimension
          end;
          if ischar(CLim)
              CLim = double([min(ImageData(:)) max(ImageData(:))]);
              if CLim(2)==CLim(1)
                  CLim(2) = CLim(2)+eps;
              end;
          end;
          hImage = image(Xcenter,Ycenter,ImageData,'Parent',hAxis,'CDataMapping',CDataMapping);
          XLim       = [0 nCols] + 0.5;                                         % Default values for XLim
          YLim       = [0 nRows] + 0.5;
      end;
      XTick      = [];                                                          % The rest of the defaults
      YTick      = [];
      XTickLabel = [];
      YTickLabel = [];
      XGrid      = 'off';
      YGrid      = 'off';
      if nargs >= 8
          XLim       = varargin{1};
          YLim       = varargin{2};
          XTick      = varargin{3};
          YTick      = varargin{4};
          XTickLabel = varargin{5};
          YTickLabel = varargin{6};
          XGrid      = varargin{7};
          YGrid      = varargin{8};
      elseif nargs >= 7
          XLim       = varargin{1};
          YLim       = varargin{2};
          XTick      = varargin{3};
          YTick      = varargin{4};
          XTickLabel = varargin{5};
          YTickLabel = varargin{6};
          XGrid      = varargin{7};
      elseif nargs >= 6
          XLim       = varargin{1};
          YLim       = varargin{2};
          XTick      = varargin{3};
          YTick      = varargin{4};
          XTickLabel = varargin{5};
          YTickLabel = varargin{6};
      elseif nargs >= 5
          XLim       = varargin{1};
          YLim       = varargin{2};
          XTick      = varargin{3};
          YTick      = varargin{4};
          XTickLabel = varargin{5};
      elseif nargs >= 4
          XLim       = varargin{1};
          YLim       = varargin{2};
          XTick      = varargin{3};
          YTick      = varargin{4};
      elseif nargs >= 3
          XLim       = varargin{1};
          YLim       = varargin{2};
          XTick      = varargin{3};
      elseif nargs >= 2
          XLim       = varargin{1};
          YLim       = varargin{2};
      elseif nargs >= 1
          XLim       = varargin{1};
      elseif nargs >= 0
          % use the defaults
      end;
      set(hAxis,'XLim',XLim,'YLim',YLim,'XTick',XTick,'YTick',YTick,'XTickLabel',XTickLabel,'YTickLabel',YTickLabel, ...
                'XGrid',XGrid,'YGrid',YGrid,'DataAspectRatio', [1 1 1],'Visible','off');
      if ~isempty(CLim)
          set(hAxis,'CLim',CLim);
      end;
  end; %end: if ishandle
return;

% --------------------------------------------------------------------
function BW = ComputeScoreMask(ROIX,ROIY,nRows,nCols)
% This function is called because we have the raw data points where the user click the mouse and we need to round them off to
% integer values and compute a B/W (black/white) image indicating which points are inside and which points are outside.
% This BW image must lie on the pixels, so we cannot just draw straight lines connecting the points.
%    Imagine getting from shop A to another shop B in a city thathas streets in perfectly square grids
%    We must go the shortest distance along the streets that connect the two shops.
%
% Returns a DOUBLE type matrix output

% Step 1. Round off the pixel values to the lowest integer points
% Step 2. Proceed to get the straightest line to join the dots and create the BW image

% STEP 1
  ROIX    = floor(ROIX+1);                                                      % Round off the points
  ROIY    = floor(ROIY+1);
  BW      = zeros(nRows+1,nCols);                                               % Create a blank BW matrix for score space
                                                                                % (note that an extra row is required initially)
% STEP 2
  nSegments = length(ROIX)-1;
  for counter = 1:nSegments
      x1 = ROIX(counter);  x2 = ROIX(counter+1);
      y1 = ROIY(counter);  y2 = ROIY(counter+1);

      if (x1 ~= x2)
          dx = abs(x2 - x1);  dy = abs(y2 - y1);   flip = 0;
          if (dx >= dy)
              if (x1 > x2)                                                      % Always "draw" from left to right.
                  t = x1; x1 = x2; x2 = t;   t = y1; y1 = y2; y2 = t;   flip = 1;
              end
              m = (y2 - y1)/(x2 - x1);
              x = (x1:x2).';    y = round(y1 + m*(x - x1));
          else
              if (y1 > y2)                                                      % Always "draw" from bottom to top.
                  t = x1; x1 = x2; x2 = t;   t = y1; y1 = y2; y2 = t;   flip = 1;
              end
              m = (x2 - x1)/(y2 - y1);
              y = (y1:y2).';    x = round(x1 + m*(y - y1));
          end

          if (flip)                                                             % If they were flipped, then flip them
              x = flipud(x);       y = flipud(y);                               % back again and
              x1 = ROIX(counter);  x2 = ROIX(counter+1);                        % also set x1, x2, y1 and y2
              y1 = ROIY(counter);  y2 = ROIY(counter+1);                        % back to their original values
          end;

          if ( (x1 < 1) || (x1 > nCols) || (x2 < 1) || (x2 > nCols) || (y1 < 1) || (y1 > nRows) || (y2 < 1) || (y2 > nRows) )
              % some of the points are out of the range, so force them to the border of the image
              xLow  = find(x<1);
              yLow  = find(y<1);
              xHigh = find(x>nCols);
              yHigh = find(y>nRows);

              if (length(xLow)),  x(xLow) = ones(size(xLow)); end;              % Pixels that are to the left of the left border
              if (length(yLow)),  y(yLow) = ones(size(yLow)); end;              % Pixels that are above the top border
              if (length(xHigh)), x(xHigh) = (nCols+1).*ones(size(xHigh)); end; % Pixels that are to the right of the right border
              if (length(yHigh)), y(yHigh) = (nRows+1).*ones(size(yHigh)); end; % Pixels that are below the bottom border
          end;

          diffx      = diff(x);
          dx_indices = find(diffx);
          dx_indices = dx_indices(:);

          if x2 > x1
              mark_val   = 1;
          else
              mark_val   = -1;
              dx_indices = dx_indices + 1;
          end;
          BW_indices     = [y(dx_indices) (x(dx_indices)-1)] * [1; (nRows+1) ];
          BW(BW_indices) = BW(BW_indices) + mark_val(ones(size(BW_indices)),1);
      end;                                                                      % End: x1 ~= x2
  end;                                                                          % End: for 1:NumSegments

  BW        = double(cumsum(BW) ~= 0);
  BW(end,:) = [];                                                               % Remove the last row
return;

% --------------------------------------------------------------------
function logo = McMasterLogo
% This function returns the McMaster Logo
logo=uint8([
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,0,0,0,0,1,1,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,0,1,1,1,1,1,0,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,0,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,0,0,0,0,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,1,1,0,0,0,0,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,1,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,1,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,1,1,1,1,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,0,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,1,1,0,1,1,0,1,1,1,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,1,1,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,1,1,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,1,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,0,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,0,1,0,0,0,0,0,1,1,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,0,0,0,0,0,0,1,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,0,0,0,0,1,1,0,0,0,0,1,1,0,0,0,0,1,1,1,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,0,0,0,0,1,1,0,0,0,0,0,1,1,0,0,0,0,0,1,1,0,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,0,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,0,0,1,1,1,1,0,0,1,1,1,1,0,1,1,1,1,1,0,0,0,0,0,1,1,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1,0,1,1,0,1,1,0,0,1,1,1,1,1,0,1,1,1,1,1,0,0,1,1,1,0,0,0,0,1,1,0,0,0,0,0,1,1,0,0,0,0,1,1,1,0,0,0,0,1,1,1,0,0,1,1,1,1,0,0,0,1,1,1,1,0,0,1,1,0,1,1,0,1,1,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,0,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,0,0,1,1,1,1,1,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,0,0,1,0,0,0,0,1,1,1,1,1,1,1,0,0,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,0,1,0,0,1,1,1,1,1,1,1,0,0,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,0,0,0,0,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0;
0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,0,0,1,0,1,1,1,1,0,0,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1;
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0;
1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,0,0,1,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1;
1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,0,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1;
1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1;
1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1;
1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1;
1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1;
1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1;
1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1;
1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1;
1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1;
1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1;
1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1;
1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1;
1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,1,1,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1;
1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1;
1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1;
1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,1,1,0,0,0,0,0,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,1,1,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,1,1,1,1,1,1,1,0,0,0,0,0,1,1,0,0,0,0,0,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,0,0,0,0,1,1,1,1,1,1,0,0,1,1,0,0,1,1,1,1,1,0,0,0,0,0,1,0,0,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,0,0,0,1,1,0,1,1,0,0,0,0,1,1,0,0,0,0,1,0,1,1,0,0,0,0,1,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,1,0,0,0,0,0,0,1,1,1,0,1,0,1,1,1,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,1,0,0,0,1,1,1,0,0,0,0,0,0,1,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,1,1,1,0,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,1,1,0,0,0,1,1,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,0,0,0,0,1,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,0,0,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,1,0,1,1,0,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,1,1,1,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,0,1,1,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,0,0,1,0,1,1,1,1,1,0,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,0,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,1,1,1,1,1,1,1,0,0,0,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,0,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,1,1,1,0,0,1,1,1,0,0,0,1,1,1,1,1,1,0,0,0,0,1,0,0,1,1,1,1,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,0,0,0,1,1,1,1,0,0,0,0,1,1,0,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,0,0,0,1,1,0,0,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,1,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1;
1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1]);
return


% --------------------------------------------------------------------
function Initialize_MACCMIA_GUI_Interface
% This function is called to layout the GUI and set up the environment for image analysis later on.
% Step 0. Check if the GUI is already running; quit this function if this is the case
% Step 1. If not, set up a structure of the information that will be used later on
% Step 2. Draw the figure window and place the required "uicontrols" in their correct position
% Step 3. Store the user information, make the GUI visible and ready to use

% Step 0:
  GUI_Handle = findobj(allchild(0), 'tag', 'MACC_MIA_GUI');
  if ~isempty(GUI_Handle)                                                       % The window exists,
      figure(GUI_Handle);                                                       % make it visible above other figures
      return;
  else                                                                          % otherwise display the user agreement
%      AgreementMessage = {'This multivariate image analysis package, MACCMIA, is a copyrighted', ...
%                          'package produced by the McMaster Advanced Control Consortium (MACC)', ...
%                          'for the sole use of its member companies and should{\bf not} be', ...
%                          'distributed outside of it.  You or your company''s representative', ...
%                          'must have signed the license agreement to use this software, a copy', ...
%                          'of which appears in the user''s tutorial manual.',' ', ...
%                          'Do you agree to these terms ?'};
      AgreementMessage = {'This multivariate image analysis package, MACCMIA, is a copyrighted', ...
                          'package produced by the McMaster Advanced Control Consortium (MACC).', ...
                          'This particular version is made available to illustrate the concepts of', ...
                          'MIA, multivariate image analysis.  No warranty or guarantee is provided as ', ...
                          'to the suitability to your application.',' ' ...
                          'Support is generally not available for this software, however reasonable',  ...
                          'effort will be made to answer any questions (see Help/About for details).', ...
                          'Please, however, read the user''s tutorial manual first, a copy of which', ...
                          'is distributed with the software.',' ', ...
                          'Do you agree to these terms ?'};

      AgreementOptions.Default = 'No';
      AgreementOptions.Interpreter = 'tex';
      AgreementAnswer  = questdlg(AgreementMessage,'MACCMIA User Agreement','Yes','No',AgreementOptions);
      switch AgreementAnswer,
      case 'Yes'
         tempFolder = tempdir;
%          default{1} = tempdir;
%          tempString = {'Score tool requires a temporary directory to store intermediate data files.  These data files will be deleted when the program exists.  Click OK to accept the following default directory that was found on your computer, or change the directory to another name.'};
%          for k = 1:3
%              hInput = inputdlg(tempString,'Temporary data storage location',1,default);
%              if exist(char(hInput))==7
%                  tempFolder = char(hInput);
%                  break;   % <-- this causes a warning in Matlab 6.5
%              else
%                  tempFolder = [];
%              end;
%          end;
%          if isempty(tempFolder)
%              uiwait(msgbox({'A valid temporary directory was not specified.  Exiting ...'},'Invalid temporary directory','modal'));
%              return;
%          end;

      case 'No',
          return;
      end;
  end;

% Step 2:
  ScreenDepth = get(0,'ScreenDepth');                                           % Get the level of grey resolution
  if ScreenDepth>8
      grayres=256;
  else
      grayres=128;
  end;
  handles.GUI_Handle = ...
    figure('Name','MACC Multivariate Image Analysis Package (MACCMIA)', ...     % The name appearing in the window
           'NumberTitle','off', ...                                             % Prevents display of Figure No. x in title
           'Tag','MACC_MIA_GUI',...                                             % Figure tag
           'HandleVisibility','on', ...                                         % The handles are visible to the workspace
           'Visible','off', ...                                                 % Make it invisible while we set it up
           'Resize','off', ...                                                  % Do not allow the user to resize the window
           'BusyAction','queue', ...                                            % Put other conflicting commands in a queue
           'Interruptible','off', ...                                           % Don't interrupt while setting it up
           'BackingStore','off', ...                                            % Leads to faster animation when figures are updated very frequently
           'Color',get(0,'defaultUicontrolBackgroundColor'), ...                % Background colour
           'Pointer','watch', ...                                               % Make the pointer look like an hour-glass
           'DoubleBuffer','on', ...                                             % Allows fast redrawing of images
           'IntegerHandle','off', ...                                           % Prevent handles from being integers
           'Colormap', gray(grayres), ...                                       % Set the default colourmap
           'CloseRequestFcn',@FileExit_Callback, ...                            % The function that will close the window
           'Units','pixels', ...                                                % Use pixel units to define the layout
           'Menubar','none' ...                                                 % Do not allow the menu bar to appear
          );


  SmlUD = CreateNewDataStructures(handles.GUI_Handle,'main');                   % Get a clean data storage structure
  addpath(SmlUD.InitialDir);                                                    % Add the MACCMIA directory to the Matlab path
  SmlUD.tempFolder = tempFolder;                                                % Also store the temporary folder selected above: IT MUST END WITH A TRAILING SLASH
  handles = CreateGUI(handles.GUI_Handle,handles);                              % Send the current handles, and get them back again after creating the GUI

  SmlUD.tHorizSize = get(handles.TScoreHorizAxis,'Position');
  set(handles.FileSaveButton, 'Enable','off');                                  % Disable these functions until an image is loaded successfully
  set(handles.FilePrintButton,'Enable','off');
  set(handles.HistoButton,    'Enable','off');
  set(handles.FileSave,       'Enable','off');
  set(handles.FilePrint,      'Enable','off');
  set(handles.Tools,          'Enable','off');

% Step 3
  setappdata(handles.GUI_Handle,'Small_Data',SmlUD);
  setappdata(handles.GUI_Handle,'Handles_Data', handles);
  RemoveInterfaceWidgets;                                                       % Set all the interface widgets to be invisible

  pos = get(handles.GUI_Handle,'Position');

  set(handles.GUI_Handle,'Pointer','arrow', ...                                 % Set the mouse to an arrow head
                         'HandleVisibility','callback', ...                     % only callbacks can see the handles
                         'Visible','on');                                       % Make the figure window visible

  drawnow;
  set(handles.GUI_Handle,'Position',pos);
return;

% --------------------------------------------------------------------
function output = CreateNewDataStructures(GUI_Handle,StructureType)
% This function creates empty data storage structures depending on the type requested.

  if strcmp(StructureType,'Image')
      Image.FullFileName                    = [];                               % File name where the original file can be found        (string)
      Image.StructureFile                   = [];                               % The file where this storage structure will be saved   (string)
      Image.DisplayName                     = [];                               % The file name that can be displayed, eg: "sample.bmp" (string)
      Image.BaseFileName                    = [];                               % The base file name, eg: "sample" in "sample.bmp"      (string)
      Image.ImageExtension                  = [];                               % The extension of the image type, eg: ".bmp"           (string)
      Image.SourceVariableName              = [];                               % The source image variable name in the MAT file        (string)

      Image.nRows                           = [];                               % Number of rows    (VERTICALLY)            (scalar) [double]
      Image.nCols                           = [];                               % Number of columns (HORIZONTALLY)          (scalar) [double]
      Image.nLayers                         = [];                               % Number of layers or channels              (scalar) [double]

      Image.X                               = [];                               % The actual image data                     (nRows x nCols x nLayers x nFrames) [double or uint8]
      Image.Class                           = [];                               % Type of data stored in X: "double" or "uint8" (string)
      Image.minX                            = [];                               % Minimum pixel intensity in every layer    (nLayers x nFrames) [double]
      Image.maxX                            = [];                               % Maximum pixel intensity in every layer    (nLayers x nFrames) [double]
      Image.avgX                            = [];                               % Average pixel intensity in every layer    (nLayers x nFrames) [double]
      Image.varX                            = [];                               % Pixel intensity variance in each layer    (nLayers x nFrames) [double]
      Image.XTX                             = [];                               % (X_unfolded)' * (X_unfolded)              (nLayers x nLayers x nFrames) [double]
      Image.nPC                             = [];                               % The number of principal components to use (scalar) [double]
      Image.P                               = [];                               % The loadings matrix                       (nLayers x nPC)     [double]

      Image.Eigenvalues                     = [];                               % The eigenvalues from the SVD decomposition(1 x nPC) [double]
      Image.minT                            = [];                               % Minimum values used to scale the T-scores (1 x nPC) [double]
      Image.maxT                            = [];                               % Maximum values used to scale the T-scores (1 x nPC) [double]
      Image.scaleT                          = [];                               % The scaling vector = 255./(maxT-minT)     (1 x nPC) [double]
      Image.T                               = [];                               % The T-scores, refolded as images          (nRows x nCols x nLayers x nFrames) [uint8]

      Image.ScoreImage                      = uint8(zeros(256));                % The default score image - a 256x256 matrix of uint8 zeros
      Image.Mask                            = uint8(zeros(256));                % The default score space mask - a 256x256 matrix of uint8 zeros
      Image.MaskXdata                       = [];                               % The x-coordinate vertices of the mask
      Image.MaskYdata                       = [];                               % The y-coordinate vertices of the mask
      Image.MaskXindex                      = [];                               % The T-score index used to get these coordinates: horizontal axis
      Image.MaskYindex                      = [];                               % The T-score index used to get these coordinates: vertical axis

      Image.ErrMsg                          = [];                               % The error messages that accumulate when processing an image object
      output = Image;
      return;
  end;

  ScreenSize = get(0,'ScreenSize');                                             % Get the size of the user's screen
  if (ScreenSize(1) ~= ScreenSize(2))
    % This step will happen when the user has changed the screen resolution while Matlab has been running
    % Not too critical to operation, but it is better if we have a clean start
    error('Screen resolution has changed.  Please quit and restart Matlab to get the screen settings correct');
  end;

  SmlUD.MACCMIA_Version                     = 1.80;                             % The data storage structure version
  SmlUD.InitialDir                          = pwd;                              % The current directory
  SmlUD.FileNamePath                        = {' ',pwd};                        % Set the file name and path to the current working dir.
  SmlUD.CurrentDateTime                     = now;                              % Get the current date and time to store in the handles
  SmlUD.tempFolder                          = [];                               % The temporary directory to store data
  SmlUD.ScreenSize                          = ScreenSize;                       % Information regarding the user's screen size
  SmlUD.DataHasBeenSaved                    = 1;                                % 1 if true, otherwise 0

  SmlUD.RGBmode                             = 1;                                % 1 if GUI is in RGB mode, otherwise 0
  SmlUD.ZoomMode                            = 0;                                % 1 if in zoom mode, otherwise 0
  SmlUD.expMode                             = 0;                                % 1 if in explore mode, otherwise 0
  SmlUD.adjMode                             = 0;                                % 1 if in adjustment mode, otherwise 0
  SmlUD.adjModeButton                       = 0;                                % 1 if in adjustment mode and mouse button is pushed down, otherwise 0

  SmlUD.hAxis                               = 1;                                % The current score to be put on the horizontal axis
  SmlUD.vAxis                               = 2;                                % The current score to be put on the vertical axis

%  SmlUD.ScoreImage                          = [];                               % The current score space image
%  SmlUD.XMaskPoints                         = [];                               % Mask information from the score space
%  SmlUD.YMaskPoints                         = [];                               % Mask information from the score space
%  SmlUD.ScoreMaskState                      = [];                               % Structure containing the state of the figure before masking
  SmlUD.ImageMaskColour                     = [ 1 1 1 ];                        % RGB colour selection for the mask in the image space
  SmlUD.ScoreMaskColour                     = [ 0 1 1 ];                        % RGB colour selection for the mask in the image space
%  SmlUD.SelectedPixels                      = [];                               % Pixels selected in the image space to map in score space
%  SmlUD.MouseClickR                         = [];                               % The image row where the mouse was clicked in image space
%  SmlUD.MouseClickC                         = [];                               % The image col where the mouse was clicked in image space
  SmlUD.CountMessage                        = 0;                                % Count the number of times that the message about pushing "space" bar has been shown

  SmlUD.Points                              = [];                               % A mini-structure that contains the clicked mouse points from the image space
  SmlUD.Points.List                         = [];                               % The list of pixels selected in the image space to map in score space
  SmlUD.Points.R                            = [];                               % The image-space row    location where the mouse was clicked
  SmlUD.Points.C                            = [];                               % The image-space column location where the mouse was clicked

  SmlUD.RLayer                              = 1;                                % Default, Red   = T1
  SmlUD.GLayer                              = 2;                                % Default, Green = T2
  SmlUD.BLayer                              = 3;                                % Default, Blue  = T3
  SmlUD.MaxNumPCBars                        = 5;                                % Maximum number of bars in PC bar plot
  SmlUD.ScoreMaskSize                       = 16;                               % The initial size of the score space explorer mask
  SmlUD.ImageMaskSize                       = 24;                               % The initial size of the image space explorer mask

  SmlUD.ScoreSpaceColourMap = [                                                 % The colour map used in the score space
    0.01041666666667 0 0; 0.79166666666667 0 0; 1 0.05208333333333 0; 1 0.31250000000000 0; 1 0.31250000000000 0; ...
    1 0.31250000000000 0; 1 0.31250000000000 0; 1 0.57291666666667 0; 1 0.57291666666667 0; 1 0.57291666666667 0; ...
    1 0.57291666666667 0; 1 0.57291666666667 0; 1 0.57291666666667 0; 1 0.57291666666667 0; 1 0.57291666666667 0; ...
    1 0.57291666666667 0; 1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; ...
    1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; ...
    1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; ...
    1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; 1 0.83333333333333 0; 1 1 0.14062500000000; ...
    1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; ...
    1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; ...
    1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; ...
    1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; ...
    1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; 1 1 0.140625; ...
    1 1 0.140625; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 0.531250; 1 1 0.531250; 1 1 0.531250; 1 1 0.531250; ...
    1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; ...
    1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; ...
    1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; ...
    1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; ...
    1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; ...
    1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; ...
    1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 1 1 1; 0 1 1];


  Circle = ones(16,16).*NaN;
  Value = 2;
  Layer = 1;
  Circle(1,8:9) = Value;
  Circle(2,[6:7 10:11]) = Value;
  Circle(3,[5 12]) = Value;
  Circle(4,[4 13]) = Value;
  Circle(5,[3 14]) = Value;
  Circle(6,[3 14]) = Value;
  Circle(7,[2 15]) = Value;
  Circle(8,[1 16]) = Value;
  Circle(9,[1 16]) = Value;
  Circle(10,[2 15]) = Value;
  Circle(11,[3 14]) = Value;
  Circle(12,[3 14]) = Value;
  Circle(13,[4 13]) = Value;
  Circle(14,[5 12]) = Value;
  Circle(15,[6:7 10:11]) = Value;
  Circle(16,8:9) = Value;

  Circle(2,8:9) = Layer;
  Circle(3,[6:7 10:11]) = Layer;
  Circle(4,[5 12]) = Layer;
  Circle(5,[4 13]) = Layer;
  Circle(6,[4 13]) = Layer;
  Circle(7,[3 14]) = Layer;
  Circle(8,[2 15]) = Layer;
  Circle(9,[2 15]) = Layer;
  Circle(10,[3 14]) = Layer;
  Circle(11,[4 13]) = Layer;
  Circle(12,[4 13]) = Layer;
  Circle(13,[5 12]) = Layer;
  Circle(14,[6:7 10:11]) = Layer;
  Circle(15,8:9) = Layer;


  SmlUD.CurrentPointerShape = get(GUI_Handle,'Pointer');                        % Get current mouse settings
  SmlUD.CirclePointerShape  = Circle;

  temp = setptr('hand');        SmlUD.HandOpenPointerShape = temp{4};
  temp = setptr('closedhand');  SmlUD.HandClosedPointerShape = temp{4};

  output = SmlUD;
% setappdata(GUI_Handle,'Mouse_Data',MseUD);
return;                                                                         % return - end of this function


% --------------------------------------------------------------------
function handles = CreateGUI(GUI_Handle,handles)
% This function creates the GUI elements.  The handle to the GUI figure window is supplied (GUI_Handle) as well as a structure (empty or populated)
% that contains the handles to the elements.  This structure is returned in the output.

  ScreenSize  = get(0,'ScreenSize');                                            % Get the size of the user's screen
  if (ScreenSize(3) < 800) || (ScreenSize(4) < 600)
      if ishandle(GUI_Handle), delete(GUI_Handle); end;                         % Screen size is too small
      error('A screen resolution greater than 800 x 600 is required to run this software.');
  end
  if (ScreenSize(3) < 1024) || (ScreenSize(4) < 768)
      uiwait(msgbox({'This software is able to operate on screen resolutions below 1024 x 768, but certain elements may not appear as intended.'},'Warning: low screen resolution'))
  end;

  FigurePosition = get(GUI_Handle, 'position');                                 % Get the position of the current figure
  if all(ScreenSize(3:4)==[800 600])
      FigurePosition(1) = 0;
      FigurePosition(2) = 0;

      FigurePosition(3) = 800;
      FigurePosition(4) = 600;
  else
      FigurePosition(1) = round(0.03*ScreenSize(3));
      FigurePosition(2) = round(0.08*ScreenSize(4));
      FigurePosition(3) = ScreenSize(3)-2*FigurePosition(1);
      FigurePosition(4) = ScreenSize(4)-2*FigurePosition(2);

  end;

  ratio  = FigurePosition(4)/FigurePosition(3);                                 % Get the ratio so that it is 0.75 and the the largest side fits the screen
  if ratio < 0.75
      FigurePosition(3) = FigurePosition(4)/0.75;
  else
      FigurePosition(4) = FigurePosition(3)*0.75;
  end;
  if all(ScreenSize(3:4)==[800 600])
      %FigurePosition(1) = max(0,round((ScreenSize(3) - FigurePosition(3))/2));
      %FigurePosition(2) = max(0,round((ScreenSize(4) - FigurePosition(4))/2));
      FigurePosition = [10 10 700 525];
      set(GUI_Handle,'Position',FigurePosition);
      movegui(GUI_Handle,'center');   % This function shows the interesting property of the window called: get(GUI_Handle,'OuterPosition')
  else
      FigurePosition(1) = max(5,round((ScreenSize(3) - FigurePosition(3))/2));
      FigurePosition(2) = max(5,round((ScreenSize(4) - FigurePosition(4))/2));
      set(GUI_Handle,'Position',FigurePosition);
  end;


  %movegui(GUI_Handle,'center');   % This function shows the interesting property of the window called: get(GUI_Handle,'OuterPosition')

  Cols = FigurePosition(3);
  Rows = FigurePosition(4);

  % Basic parameters that affect the layout
  ImageScoreAxes    = 0.39*Cols;                                                % The image and the scores square axes
  LittleZoomAxes    = 0.12*Cols;                                                % The small zoomed window
  SideSpace         = 0.08*Cols;                                                % Spacing from the edge of the GUI to the axes
  InterSpacing      = 0.06*Cols;                                                % The space between the two square axes
  TopSpace          = 0.07*Rows;                                                % The space from top of GUI to top of axes
  PopupWidth        = 0.057*Cols;                                               % The width of the standard popup's
  WinColour         = get(0,'defaultUicontrolBackgroundColor');                 % Windows background colour

  junkHandle = uicontrol('Style','Pushbutton','FontName','Tahoma','FontSize',10,'Position',[450 300 80 30],'String','Redo mask','Visible','on');
  ExtentText = get(junkHandle,'Extent');
  if ExtentText(3) > 80
      TextSize = 9;                                                             % Use 8 point sized text: User has "Large Fonts"
  else
      TextSize = 10;                                                            % Else use 9 point text
  end;
  set(junkHandle,'FontSize',TextSize);
  ExtentText = get(junkHandle,'Extent');
  TextHeight = ExtentText(4)+2;
  delete(junkHandle);


  %DropdownSpace     = floor((ImageScoreAxes - 4*TextHeight - 5*TextHeight)/3);  % 4*DropdownSpace + 9*TextHeight = ImageScoreAxes:  The space between each popup
  DropdownSpace     = floor((ImageScoreAxes - 3*TextHeight - 3*TextHeight)/4);  % 4*DropdownSpace + 9*TextHeight = ImageScoreAxes:  The space between each popup

  % Now comes the step where we create all the widgets, specify their locations and callbacks
  %
  % Summary of widget types and a basic summary of what they will be used for:
  %
  % AXES:               ImagePlotAxes      : The general purpose axes on the left hand side of the GUI
  %                     ScorePlotAxes      : The general purpose axes on the right hand side of the GUI
  %                     ZoomImageAxes      : The small square axes below the ImagePlotAxes that contains zoomed in local views
  %                     ZoomScoreAxes      : The small square axes below the ScorePlotAxes that contains zoomed in local views
  %                     ColourBarAxes      : The thin rectangular axes to display the scores and residuals colour bar
  %                     LogoAxes           : The axes that display the McMaster logo
  %
  % LINES:              LineH1             : Line 1 used during score mask creation
  %                     LineH2             : Line 2 used during score mask creation
  %
  % POPUPS:             ImageRed           : The red drop-down popup
  %                     ImageGreen         : The green drop-down popup
  %                     ImageBlue          : The blue drop-down popup
  %                     ImageMaskColour    : Changes the colour of the current mask in the image space
  %                     ScoreMaskColour    : Changes the colour of the current mask in the score space
  %                     TScoreHorizAxis    : Dropdown for the horizontal axis on the scores axis
  %                     TScoreVrtclAxis    : Dropdown for the vertical axis on the scores axis
  %
  % LIST BOXES:         ImageStatsList     : Statistics regarding the image
  %
  % MENU ITEMS:         FILE               : Main tree: File
  %                     FileOpen           :    sub tree - Open
  %                     FileSave           :    sub tree - Save
  %                     FilePrint          :    sub tree - Print
  %                     FilePrintsetup     :    sub tree - Print setup
  %                     FileExit           :    sub tree - Exit
  %
  %                     TOOLS              : Main tree: Help%
  %                     ToolsHistoAnalysis :    sub tree - Histogram Analysis
  %                     ToolsScoreAnalysis :    sub tree - PCA Scores Analysis
  %                     ToolsLoadgAnalysis :    sub tree - PCA Loadings Analysis
  %                     ToolsModelBuilding :    sub tree - Interactive Model Building
  %
  %                     HELP               : Main tree: Help
  %
  % CHECK BOXES:        GUI_ExploreScores  : The check box to enter interactive mode of exploring the score space
  %
  % TEXT FIELDS:        ImageTextRed       : The word "Red"
  %                     ImageTextGreen     : The word "Green"
  %                     ImageTextBlue      : The word "Blue"
  %                     ImageMaskColourText: The word "Mask | Colour" - to change the mask colour in the image space
  %                     ScoreMaskColourText: The word "Mask | Colour" - to change the mask colour in the score space
  %                     expScoreSideMaskSizeText: The words "Score space mask size" - to change the mask size in the score space when "Exploring"
  %                     expImageSideMaskSizeText: The words "Image space mask size" - to change the mask size in the image space when "Exploring"
  %                     LHSPlotLabel       : The label above the left hand side plot
  %                     RHSPlotLabel       : The label above the right hand side plot
  %                     ZoomScoreInfo      : To display the number of pixels under the mask
  %
  % PUSH BUTTONS:       ResetModel         : Resets the model building by clearing all masks, clicked points and binary masks
  %                     ResetImageClicks   : Resets the image click storage to an empty matrix
  %
  % CONTEXT MENUS:      ImageContextMenu   : Image space context menu
  %                     ScoreContextMenu   : Score space context menu
  %                     LstBxContextMenu   : List box context menu
  %
  % TEXT EDIT FIELDS:   expScoreMaskSize   : The user enters the size of the mask when in the "Explore" mode of the score space
  %                     expImageMaskSize   : The user enters the size of the mask when in the "Explore" mode of the image space


  %----------------------------------------
  % AXES:               ImagePlotAxes      : The general purpose axes on the left hand side of the GUI
  %                     ScorePlotAxes      : The general purpose axes on the right hand side of the GUI
  %                     ZoomImageAxes      : The small square axes below the ImagePlotAxes that contains zoomed in local views
  %                     ZoomScoreAxes      : The small square axes below the ScorePlotAxes that contains zoomed in local views
  %                     ColourBarAxes      : The thin rectangular axes to display the scores and residuals colour bar
  %                     LogoAxes           : The axes that display the McMaster logo

  Axs.Interruptible  = 'off';                                                   % Axes fields that are different to the default values
  Axs.BusyAction     = 'queue';
  Axs.Units          = 'Pixels';
  Axs.Parent         = GUI_Handle;
  handles.ImagePlotAxes = axes(Axs,'Position',[SideSpace                                 Rows-TopSpace-ImageScoreAxes ImageScoreAxes ImageScoreAxes]);
  handles.ScorePlotAxes = axes(Axs,'Position',[SideSpace+ImageScoreAxes+InterSpacing     Rows-TopSpace-ImageScoreAxes ImageScoreAxes ImageScoreAxes]);
  handles.ZoomImageAxes = axes(Axs,'Position',[SideSpace+ImageScoreAxes-LittleZoomAxes   LittleZoomAxes               LittleZoomAxes LittleZoomAxes]);
  handles.ZoomScoreAxes = axes(Axs,'Position',[Cols-SideSpace-LittleZoomAxes             LittleZoomAxes               LittleZoomAxes LittleZoomAxes]);
  handles.ColourBarAxes = axes(Axs,'Position',[Cols-SideSpace+0.03*Cols                  Rows-TopSpace-ImageScoreAxes 0.02*Cols      ImageScoreAxes]);
  handles.LogoAxes      = axes(Axs,'Position',[1                                         1                            SideSpace   SideSpace*163/124]); % that ratio is due to image's aspect ratio
  logo = double(McMasterLogo);
  logo(:,:,2) = double(McMasterLogo);
  logo(:,:,3) = double(McMasterLogo);
  logo(logo==1) = WinColour(1);
  logo = uint8(logo.*255);
  ShowImage(logo,handles.LogoAxes,[],[],[],'direct');
  set(handles.LogoAxes,'Visible','off');


  %----------------------------------------
  % LINES:              LineH1             : Line 1 used during score mask creation
  %                     LineH2             : Line 2 used during score mask creation
  handles.LineH1 = line('Parent',handles.ScorePlotAxes,'XData',[],'YData',[],'Visible','off','Clipping','off','Color','k','LineStyle','-');
  handles.LineH2 = line('Parent',handles.ScorePlotAxes,'XData',[],'YData',[],'Visible','off','Clipping','off','Color','w','LineStyle',':');

  %----------------------------------------
  % POPUPS:             ImageRed           : The red drop-down popup
  %                     ImageGreen         : The green drop-down popup
  %                     ImageBlue          : The blue drop-down popup
  %                     ImageMaskColour    : Changes the colour of the current mask in the image space
  %                     ScoreMaskColour    : Changes the colour of the current mask in the score space
  %                     TScoreHorizAxis    : Dropdown for the horizontal axis on the scores axis
  %                     TScoreVrtclAxis    : Dropdown for the vertical axis on the scores axis
  Ctl.Interruptible = 'off';                                                    % Control fields different to default values
  Ctl.BusyAction    = 'queue';
  Ctl.Enable        = 'on';
  Ctl.String        = ' ';
  Ctl.Units         = 'Pixels';
  Ctl.Style         = 'Popupmenu';
  Ctl.FontSize      = TextSize;
  Ctl.Parent        = GUI_Handle;

  handles.ImageRed   = uicontrol(Ctl, ...
                             'Position',[3 Rows-TopSpace-1*TextHeight-1*TextHeight-1*DropdownSpace PopupWidth TextHeight], ...
                             'Tag','ImageRed',...
                             'Callback',@UpdateLHS);
  handles.ImageGreen = uicontrol(Ctl, ...
                             'Position',[3 Rows-TopSpace-2*TextHeight-2*TextHeight-2*DropdownSpace PopupWidth TextHeight], ...
                             'Tag','ImageGreen',...
                             'Callback',@UpdateLHS);
  handles.ImageBlue  = uicontrol(Ctl, ...
                             'Position',[3 Rows-TopSpace-3*TextHeight-3*TextHeight-3*DropdownSpace PopupWidth TextHeight], ...
                             'Tag','ImageBlue',...
                             'Callback',@UpdateLHS);
  handles.ImageMaskColour = uicontrol(Ctl, ...
                             'Position',[3 Rows-TopSpace-ImageScoreAxes-3*TextHeight               PopupWidth+15 TextHeight], ...
                             'Callback',@MaskColour_Callback);
                             set(handles.ImageMaskColour,'String',{'White','Red','Green','Blue','Yellow','Cyan','Black'}');

  handles.ScoreMaskColour = uicontrol(Ctl, ...
                             'Position',[Cols-3-PopupWidth-15 Rows-TopSpace-ImageScoreAxes-3*TextHeight PopupWidth+15 TextHeight], ...
                             'Tag','ScoreMaskColour',...
                             'Callback',@MaskColour_Callback);
                             set(handles.ScoreMaskColour,'String',{'White','Red','Green','Blue','Yellow','Cyan','Black'}');

  handles.TScoreHorizAxis = uicontrol(Ctl, ...
                             'Position',[Cols-SideSpace-(ImageScoreAxes+PopupWidth)/2   Rows-TopSpace-ImageScoreAxes-2*TextHeight     PopupWidth TextHeight], ...
                             'Tag','TScoreHorizAxis',...
                             'Callback',@UpdateRHS);
  handles.TScoreVrtclAxis = uicontrol(Ctl, ...
                             'Position',[Cols-SideSpace+(SideSpace-PopupWidth)/2        Rows-TopSpace-0.5*(ImageScoreAxes+TextHeight) PopupWidth TextHeight], ...
                             'Tag','TScoreVrtclAxis',...
                             'Callback',@UpdateRHS);


  %----------------------------------------
  % LIST BOXES:         ImageStatsList     : Statistics regarding the image
  Ctl.Style = 'listbox';
  handles.ImageStatsList = uicontrol(Ctl, ...
                                      'Position',[SideSpace+ImageScoreAxes+InterSpacing 0.025*Rows ImageScoreAxes 2*LittleZoomAxes], ...
                                      'Tag','ImageStatsListbox','Callback','','Max',2,'HorizontalAlignment','Left','FontName','FixedWidth','FontSize',10,...
                                      'String','Channel   Min   Max   Mean   Std.Dev. ');  % Fixed-width can't go smaller than 12pt
                                      % Max > Min implies the user can select multiple lines
%  extent = get(handles.ImageStatsList,'Extent');
%  positn = get(handles.ImageStatsList,'Position');
%  if extent(3) > positn(3)                                                      % The text in the drop-down is cut off on the RHS edge
%      diff = extent(3) - positn(3);
%      positn(3) = positn(3) + 1.20*diff;                                        % Increase it to 120% of what is required
%      positn(1) = positn(1) - 0.60*diff;                                        % But shift the GUI element to the left by half the additional adjustment
%      set(handles.ImageStatsList,'Position',positn);
%  end;


  %----------------------------------------
  % MENU ITEMS:         File               : Main tree: File
  %                     FileOpen           :    sub tree - Open
  %                     FileSave           :    sub tree - Save
  %                     FilePrint          :    sub tree - Print
  %                     FilePrintsetup     :    sub tree - Print setup
  %                     FileExit           :    sub tree - Exit
  %                     TOOLS              : Main tree: Tools
  %                     ToolsHistoAnalysis :    sub tree - Histogram Analysis
  %                     ToolsScoreAnalysis :    sub tree - Analysis of PCA Scores
  %                     ToolsLoadgAnalysis :    sub tree - Analysis of PCA Loadings
  %                     ToolsModelBuilding :    sub tree - Interactive Model Building
  %                     HELP               : Main tree: Help
  handles.File               = uimenu(GUI_Handle,'Label','File');        % Create the file menu, with the main interface as parent
  handles.FileOpen           = uimenu(handles.File,'Label','Open' ,       'Callback',@FileOpen_Callback, 'Accelerator','O');
  handles.FileSave           = uimenu(handles.File,'Label','Save' ,       'Callback',@FileSave_Callback, 'Accelerator','S');
  handles.FilePrint          = uimenu(handles.File,'Label','Print',       'Callback',@FilePrint_Callback,'Accelerator','P','Separator','on');
  handles.FileExit           = uimenu(handles.File,'Label','Exit',        'Callback',@FileExit_Callback, 'Accelerator','Q','Separator','on');

  handles.Tools              = uimenu(GUI_Handle,'Label','Tools');       % Create the tools menu, with the main interface as parent
  handles.ToolsHistoAnalysis = uimenu(handles.Tools,'Label','Histogram Analysis','Callback',@HistogramAnalysis_Callback);
  handles.ToolsScoreAnalysis = uimenu(handles.Tools,'Label','Analysis of PCA Scores','Callback',@ScoresAnalysisMode_Callback,'Separat','on');
  handles.ToolsLoadgAnalysis = uimenu(handles.Tools,'Label','Analysis of PCA Loadings','Callback',@LoadingsAnalysisMode_Callback);
  handles.ToolsModelBuilding = uimenu(handles.Tools,'Label','Interactive Model Interrogation','Callback',@ModelBuildingMode_Callback);

  handles.Help               = uimenu(GUI_Handle,'Label','Help');        % Create the Help menu, with the main interface as parent
  handles.HelpAbout          = uimenu(handles.Help,'Label','About' ,      'Callback',@HelpAbout_Callback);

  %----------------------------------------
  % CHECK BOXES:        GUI_ExploreScores  : The check box to enter interactive mode of exploring the score space
  Ctl.Style = 'Checkbox';
  handles.GUI_ExploreScores  = uicontrol(Ctl, ...
                                         'Position',[0.70*Cols 0.015*Rows 0.15*Cols TextHeight], ...
                                         'String','Explore mode', ...
                                         'TooltipString','Dynamically explore the score and image space (Ctrl-E)', ...
                                         'Callback',@ExploreScores_Callback);

  %----------------------------------------
  % TEXT FIELDS:        ImageTextRed       : The word "Red"
  %                     ImageTextGreen     : The word "Green"
  %                     ImageTextBlue      : The word "Blue"
  %                     ImageMaskColourText: The word "Mask | Colour" - to change the mask colour in the image space
  %                     ScoreMaskColourText: The word "Mask | Colour" - to change the mask colour in the score space
  %                     expScoreSideMaskSizeText: The words "Score space mask size" - to change the mask size in the score space when "Exploring"
  %                     expImageSideMaskSizeText: The words "Image space mask size" - to change the mask size in the image space when "Exploring"
  %                     LHSPlotLabel       : The label above the left hand side plot
  %                     RHSPlotLabel       : The label above the right hand side plot
  %                     ZoomScoreInfo      : To display the number of pixels under the mask

  Txt.Interruptible = 'off';                                                    % Text fields different to default values
  Txt.BusyAction    = 'queue';
  Txt.Style         = 'text';
  Txt.Units         = 'pixels';
  Txt.HorizontalAli = 'center';
  Txt.Background    = WinColour;
  Txt.FontSize      = TextSize;
  Txt.Parent        = GUI_Handle;

  handles.ImageTextRed = uicontrol(Txt, ...
                             'Position',[3 Rows-TopSpace-1*TextHeight-0*TextHeight-1*DropdownSpace PopupWidth TextHeight], ...
                             'ForegroundColor',[1 0 0], 'String','Red');
  handles.ImageTextGreen = uicontrol(Txt, ...
                             'Position',[3 Rows-TopSpace-2*TextHeight-1*TextHeight-2*DropdownSpace PopupWidth TextHeight], ...
                             'ForegroundColor',[0 1 0], 'String','Green');
  handles.ImageTextBlue = uicontrol(Txt, ...
                             'Position',[3 Rows-TopSpace-3*TextHeight-2*TextHeight-3*DropdownSpace PopupWidth TextHeight], ...
                             'ForegroundColor',[0 0 1], 'String','Blue');
  handles.ImageMaskColourText = uicontrol(Txt, ...
                             'Position',[3 Rows-TopSpace-ImageScoreAxes-2*TextHeight               PopupWidth TextHeight*2], ...
                             'ForegroundColor',[0 0 0], 'String',{'Mask','Colour'}');
  handles.ScoreMaskColourText = uicontrol(Txt, ...
                             'Position',[Cols-SideSpace Rows-TopSpace-ImageScoreAxes-2*TextHeight  SideSpace TextHeight*2], ...
                             'ForegroundColor',[0 0 0], 'String',{'Mask','Colour'});

  handles.expScoreSideMaskSizeText = uicontrol(Txt, ...
                             'Position',[SideSpace LittleZoomAxes*1.3+TextHeight  ImageScoreAxes-LittleZoomAxes TextHeight], ...
                             'ForegroundColor',[0 0 0], 'String','Score space mask size:','HorizontalAlignment','left');
  extent_S = get(handles.expScoreSideMaskSizeText,'Extent');
  handles.expImageSideMaskSizeText = uicontrol(Txt, ...
                             'Position',[SideSpace LittleZoomAxes*1.3-TextHeight  ImageScoreAxes-LittleZoomAxes TextHeight], ...
                             'ForegroundColor',[0 0 0], 'String','Image space mask size:','HorizontalAlignment','left');
  extent_I = get(handles.expImageSideMaskSizeText,'Extent');
  max_extent = max([extent_S(3) extent_I(3)])*1.2;

  handles.LHSPlotLabel  = uicontrol(Txt, ...
                             'Position',[SideSpace Rows-round(TopSpace*2/3) ImageScoreAxes TextHeight], ...
                             'ForegroundColor',[0 0 0], 'String','LHS plot');
  handles.RHSPlotLabel  = uicontrol(Txt, ...
                             'Position',[SideSpace+ImageScoreAxes+InterSpacing    Rows-round(TopSpace*2/3) ImageScoreAxes TextHeight], ...
                             'ForegroundColor',[0 0 0], 'String','RHS plot');
  handles.ZoomScoreInfo = uicontrol(Txt,'Position',[Cols-SideSpace-ImageScoreAxes 0.60*LittleZoomAxes ImageScoreAxes TextHeight], ...
                             'ForegroundColor',[0 0 0], 'String','DefaultText','HorizontalAli','left');

  handles.ZoomSpaceMessage = uicontrol(Txt,'Position',[SideSpace Rows-TopSpace-ImageScoreAxes-TextHeight*2 ImageScoreAxes TextHeight], ...
                             'ForegroundColor',[0 0 0], 'String','DefaultText','HorizontalAli','center','Visible','off');

  handles.ZoomXYLocation = uicontrol(Txt,'Position',[SideSpace 0.015*Rows ImageScoreAxes TextHeight], ...
                             'ForegroundColor',[0 0 0], 'String','DefaultText','HorizontalAli','left','Visible','off');


  %----------------------------------------
  % PUSH BUTTONS:       ResetModel         : Resets the model building by clearing all masks, clicked points and binary masks
  %                     ResetImageClicks   : Resets the image click storage to an empty matrix
  Ctl.Style = 'PushButton';
  handles.ResetModel = uicontrol(Ctl, ...
                                 'Position',[SideSpace LittleZoomAxes*1.3+TextHeight PopupWidth*2   TextHeight], ...
                                 'Tag','ResetModel', ...
                                 'String','Reset model', ...
                                 'Callback',@ResetModel_Callback);
  handles.ResetImageClicks = uicontrol(Ctl, ...
                                 'Position',[SideSpace LittleZoomAxes*1.3-TextHeight PopupWidth*2.7 TextHeight], ...
                                 'Tag','ResetImageClicks', ...
                                 'String','Reset clicked points', ...
                                 'Callback',@ResetImageClicks_Callback);

  handles.StartAdjustMask = uicontrol(Ctl, ...
                                 'Position',[SideSpace+ImageScoreAxes+InterSpacing LittleZoomAxes*1.3-TextHeight PopupWidth*2.7 TextHeight], ...
                                 'String','Adjust the mask', ...
                                 'Callback',@AdjustMask_Callback);
  AdjustPosition = get(handles.StartAdjustMask,'Position');

  r        = ones(floor(AdjustPosition(4)),floor(AdjustPosition(3))).*255;
  r(:,:,2) = ones(floor(AdjustPosition(4)),floor(AdjustPosition(3))).*0;
  r(:,:,3) = ones(floor(AdjustPosition(4)),floor(AdjustPosition(3))).*255;
  handles.EndAdjustMask = uicontrol(Ctl,'Position',AdjustPosition,'String','End adjusting','Callback','uiresume(gcbf)','CData',uint8(r));


  %----------------------------------------
  % CONTEXT MENUS:      ImageContextMenu   : Image space context menu
  %                     ScoreContextMenu   : Score space context menu
  %                     LstBxContextMenu   : List box context menu
  handles.ImageContextMenu = uicontextmenu('Parent',GUI_Handle,'Tag','UICM_ImageSpace');
  handles.ScoreContextMenu = uicontextmenu('Parent',GUI_Handle,'Tag','UICM_ScoreSpace');
  uimenu(handles.ImageContextMenu,'Label','Send this axis to a new figure window','Callback',@SendLHSImageToNewFigureWindow);
  uimenu(handles.ScoreContextMenu,'Label','Send this axis to a new figure window','Callback',@SendRHSImageToNewFigureWindow);

  %----------------------------------------
  % TEXT EDIT FIELDS:   expScoreMaskSize   : The user enters the size of the mask when in the "Explore" mode of the score space
  %                     expImageMaskSize   : The user enters the size of the mask when in the "Explore" mode of the image space
  Ctl.Style = 'Edit';
  handles.expImageMaskSize = uicontrol(Ctl, ...
                                       'Position',[SideSpace+max_extent LittleZoomAxes*1.3-TextHeight  35 TextHeight], ...
                                       'String','12','Callback',@expMaskSizeAdjust_Callback);
  handles.expScoreMaskSize = uicontrol(Ctl, ...
                                       'Position',[SideSpace+max_extent LittleZoomAxes*1.3+TextHeight  35 TextHeight], ...
                                       'String','128','Callback',@expMaskSizeAdjust_Callback);

%   set(handles.expScoreMaskSize,   'Visible','off');
%   set(handles.expImageMaskSize,   'Visible','off');


  %----------------------------------------
  % MENU BAR ICONS

  ZoomMore = [NaN   NaN   NaN   NaN     0     0     0     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN; ...
              NaN   NaN     0     0   NaN   NaN   NaN   NaN     0     0   NaN   NaN   NaN   NaN   NaN   NaN; ...
              NaN     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN   NaN; ...
              NaN     0   NaN   NaN   NaN     0     0   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN   NaN; ...
                0   NaN   NaN   NaN   NaN     0     0   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN; ...
                0   NaN   NaN     0     0     0     0     0     0   NaN   NaN     0   NaN   NaN   NaN   NaN; ...
                0   NaN   NaN     0     0     0     0     0     0   NaN   NaN     0   NaN   NaN   NaN   NaN; ...
                0   NaN   NaN   NaN   NaN     0     0   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN; ...
              NaN     0   NaN   NaN   NaN     0     0   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN   NaN; ...
              NaN     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN   NaN; ...
              NaN   NaN     0     0   NaN   NaN   NaN   NaN     0     0     0     0   NaN   NaN   NaN   NaN; ...
              NaN   NaN   NaN   NaN     0     0     0     0   NaN   NaN     0     0     0   NaN   NaN   NaN; ...
              NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0     0     0   NaN   NaN; ...
              NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0     0     0   NaN; ...
              NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN];

%   ZoomLess = [NaN   NaN   NaN   NaN     0     0     0     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN; ...
%               NaN   NaN     0     0   NaN   NaN   NaN   NaN     0     0   NaN   NaN   NaN   NaN   NaN   NaN; ...
%               NaN     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN   NaN; ...
%               NaN     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN   NaN; ...
%                 0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN; ...
%                 0   NaN   NaN     0     0     0     0     0     0   NaN   NaN     0   NaN   NaN   NaN   NaN; ...
%                 0   NaN   NaN     0     0     0     0     0     0   NaN   NaN     0   NaN   NaN   NaN   NaN; ...
%                 0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN; ...
%               NaN     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN   NaN; ...
%               NaN     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN   NaN   NaN   NaN; ...
%               NaN   NaN     0     0   NaN   NaN   NaN   NaN     0     0     0     0   NaN   NaN   NaN   NaN; ...
%               NaN   NaN   NaN   NaN     0     0     0     0   NaN   NaN     0     0     0   NaN   NaN   NaN; ...
%               NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0     0     0   NaN   NaN; ...
%               NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0     0     0   NaN; ...
%               NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN   NaN];

  ZoomMore(:,:,2) = ZoomMore(:,:,1);   ZoomMore(:,:,3) = ZoomMore(:,:,1);
  ZoomLess = ZoomMore;
  ZoomLess(4:5,6:7,:) = NaN;
  ZoomLess(8:9,6:7,:) = NaN;

  JustZoom = ZoomLess;
  JustZoom(6:7,4:9,:) = NaN;

  FileOpen = ones(16,16,3);
  FileOpen(:,1:2,:)  = NaN;    FileOpen(:,14:16,:) = NaN;       % cut off the sides
  FileOpen(:,3,:)  = 0;        FileOpen(:,13,:) = 0;            % draw side lines
  FileOpen(1,:,:)  = NaN;      FileOpen(16,:,:) = NaN;          % cut off the top and bottom
  FileOpen(2,3:13,:) = 0;      FileOpen(15,3:13,:) = 0;         % draw top and bottom
  FileOpen(3,9:10,:) = 0;      FileOpen(4,[9 11],:) = 0;        % draw triange in the upper right corner
  FileOpen(5,[9 12],:) = 0;    FileOpen(6,[9,13],:) = 0;
  FileOpen(2,10:13,:) = NaN;   FileOpen(3,11:13,:) = NaN;
  FileOpen(4,[12 13],:) = NaN; FileOpen(5,13,:) = NaN;
  FileOpen(7,9:13,:) = 0;

  FileSave(:,:,1) = ones(16,16).*0.5;
  FileSave(:,:,2) = ones(16,16).*0.5;
  FileSave(:,:,3) = ones(16,16).*0.0;
  FileSave(:,[1 16],:) = NaN;
  FileSave([1 16],:,:) = NaN;

  FileSave(2:14,2,:) = 0;
  FileSave([2 15],3:14,:) = 0;
  FileSave(15,2,:) = NaN;
  FileSave(2:15,15,:) = 0;
  FileSave(3:7,[4 13],:) = 0;
  FileSave([8 11],5:12,:) = 0;
  FileSave(4,14,:) = 0;
  FileSave(12:14,5:10,:) = 0;
  FileSave(11:14,13,:) = 0;
  FileSave(12:14,11:12,:) = 1;
  FileSave(3:7,5:12,:) = 0.8;
  FileSave(3,14,:) = 1;


%  FileExit(:,:,1) = ones(16,16).*0.3;
%  FileExit(:,:,2) = ones(16,16).*0.3;
%  FileExit(:,:,3) = ones(16,16).*0.3;
%  FileExit = ones(16,16,3).*NaN;
%  FileExit(2:15,[4 13],:) = 0;
%  FileExit([2 15],4:13,:) = 0;

  Print(:,:,1) = ...
          [NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN;
           NaN   NaN   NaN   NaN   NaN     0     0     0     0     0     0     0     0     0   NaN   NaN
           NaN   NaN   NaN   NaN     0     1     1     1     1     1     1     1     1     0   NaN   NaN
           NaN   NaN   NaN   NaN     0     1     0     0     0     0     0     1     0   NaN   NaN   NaN
           NaN   NaN   NaN     0     1     1     1     1     1     1     1     1     0   NaN   NaN   NaN
           NaN   NaN   NaN     0     1     0     0     0     0     0     1     0     0     0     0   NaN
           NaN   NaN     0     1     1     1     1     1     1     1     1     0   NaN     0   NaN     0
           NaN     0     0     0     0     0     0     0     0     0     0   NaN     0   NaN     0     0
             0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN     0   NaN     0
             0     0     0     0     0     0     0     0     0     0     0     0     0   NaN   NaN     0
             0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN     0   NaN
             0   NaN   NaN   NaN   NaN   NaN   NaN     1     1     1   NaN   NaN     0     0     0   NaN
             0     0     0     0     0     0     0     0     0     0     0     0     0   NaN     0   NaN
           NaN     0   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN     0   NaN     0   NaN   NaN
           NaN   NaN     0     0     0     0     0     0     0     0     0     0     0   NaN   NaN   NaN
           NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN   NaN];
  Print(:,:,2) = ...
  [ 0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804   0.9804    0.9804
    0.9804    0.9804    0.9804    0.9804    0.9804         0         0         0         0         0         0         0         0         0   0.9804    0.9804
    0.9804    0.9804    0.9804    0.9804         0    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000         0   0.9804    0.9804
    0.9804    0.9804    0.9804    0.9804         0    1.0000         0         0         0         0         0    1.0000         0    0.9804   0.9804    0.9804
    0.9804    0.9804    0.9804         0    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000         0    0.9804   0.9804    0.9804
    0.9804    0.9804    0.9804         0    1.0000         0         0         0         0         0    1.0000         0         0         0        0    0.9804
    0.9804    0.9804         0    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000         0    0.9804         0   0.9804         0
    0.9804         0         0         0         0         0         0         0         0         0         0    0.9804         0    0.9804        0         0
         0    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804         0    0.9804         0   0.9804         0
         0         0         0         0         0         0         0         0         0         0         0         0         0    0.9804   0.9804         0
         0    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804         0    0.9804        0    0.9804
         0    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    1.0000    1.0000    1.0000    0.9804    0.9804         0         0        0    0.9804
         0         0         0         0         0         0         0         0         0         0         0         0         0    0.9804        0    0.9804
    0.9804         0    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804         0    0.9804         0   0.9804    0.9804
    0.9804    0.9804         0         0         0         0         0         0         0         0         0         0         0    0.9804   0.9804    0.9804
    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804    0.9804   0.9804    0.9804];

  Print(:,:,3) = ...
  [ 0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824   0.8824    0.8824
    0.8824    0.8824    0.8824    0.8824    0.8824         0         0         0         0         0         0         0         0         0   0.8824    0.8824
    0.8824    0.8824    0.8824    0.8824         0    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000         0   0.8824    0.8824
    0.8824    0.8824    0.8824    0.8824         0    1.0000         0         0         0         0         0    1.0000         0    0.8824   0.8824    0.8824
    0.8824    0.8824    0.8824         0    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000         0    0.8824   0.8824    0.8824
    0.8824    0.8824    0.8824         0    1.0000         0         0         0         0         0    1.0000         0         0         0         0   0.8824
    0.8824    0.8824         0    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000    1.0000         0    0.8824         0   0.8824         0
    0.8824         0         0         0         0         0         0         0         0         0         0    0.8824         0    0.8824        0         0
         0    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824         0    0.8824         0   0.8824         0
         0         0         0         0         0         0         0         0         0         0         0         0         0    0.8824   0.8824         0
         0    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824         0    0.8824         0   0.8824
         0    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824         0         0         0    0.8824    0.8824         0         0         0   0.8824
         0         0         0         0         0         0         0         0         0         0         0         0         0    0.8824         0   0.8824
    0.8824         0    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824         0    0.8824         0   0.8824    0.8824
    0.8824    0.8824         0         0         0         0         0         0         0         0         0         0         0    0.8824   0.8824    0.8824
    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824    0.8824   0.8824    0.8824];

   Scores = ones(16,16,3).*NaN;
   Scores(1:5,[1 14],:) = 0;
   Scores(1:4,[2 13],:) = 0;
   Scores(1:3,[3 12],:) = 0;
   Scores(1:2,4:11,:) = 0;
   Scores(1:16,7:8,:) = 0;
   Scores(16,4:11,:)  = 0;
   Scores(15,5:10,:)  = 0;

   Loadings = ones(16,16,3).*NaN;
   Loadings([1 16],1,:) = 0;
   Loadings([1:2 15:16],2,:) = 0;
   Loadings(1:16,3:4,:) = 0;
   Loadings([1 2 7 8 15 16],5,:) = 0;
   Loadings([1 2 8 9 16],6,:) = 0;
   Loadings([1 2 8 9],7:9,:) = 0;
   Loadings([1 2 8 9],10:11,:) = 0;
   Loadings([1:3 7:9],12,:) = 0;
   Loadings([2:4 6:8],13,:) = 0;
   Loadings(3:7,14,:) = 0;
   Loadings(4:6,15,:) = 0;

   Model = ones(16,16,3).*NaN;
   Model(1,1:5,:) = 0;
   Model(1:7,3,:) = 0;
   Model(5:9,5,:) = 0;
   Model(6,9:13,:) = 0;
   Model(7:12,11,:) = 0;
   Model(11,[13 15],:) = 0;
   Model(10,14,:) = 0;
   Model(12,15,:) = 0;
   Model(13,14,:) = 0;
   Model(14,13,:) = 0;
   Model(15,13:15,:) = 0;

   Histo = ones(16,16,3).*NaN;
   Histo(:,1,:)  = 0;
   Histo(16,:,:) = 0;
   Histo(7:15,4:5,:)    = 0;
   Histo(3:15,7:8,:)    = 0;
   Histo(11:15,10:11,:) = 0;
   Histo(5:15,13:14,:)  = 0;

   Help = ones(16,16,3).*NaN;
   Help(1,7:9,:) = 0;
   Help(2,6:10,:) = 0;
   Help(3,5:11,:) = 0;
   Help(4,[4:6 9:11],:) = 0;
   Help(5:6,[4:5 10:12],:) = 0;
   Help(7,[5 10:11],:) = 0;
   Help(8,9:11,:) = 0;
   Help(9,9:10,:) = 0;
   Help(10,8:10,:) = 0;
   Help([11:13 15:16],8:9,:) = 0;

   handles.GUI_Toolbar    =   uitoolbar('Parent',GUI_Handle);
   handles.FileOpenButton = uipushtool('Parent',handles.GUI_Toolbar,'Separator','off' ,'Enable','on' ,'CData',FileOpen,'ClickedCallback',@FileOpen_Callback,'TooltipString','Open a new file to analyze');
   handles.FileSaveButton = uipushtool('Parent',handles.GUI_Toolbar,'Separator','on' ,'Enable','off','CData',FileSave,'ClickedCallback',@FileSave_Callback,'TooltipString','Save results to a MACCMIA file');
   handles.FilePrintButton= uipushtool('Parent',handles.GUI_Toolbar,'Separator','off' ,'Enable','off','CData',Print,'ClickedCallback',@FilePrint_Callback,'TooltipString','Instructions for printing figures');
 %  handles.FileExitButton = uipushtool('Parent',handles.GUI_Toolbar,'Separator','off' ,'Enable','off','CData',FileExit,'ClickedCallback',@FileExit_Callback,'TooltipString','Quit MACCMIA');

   handles.ScoresButton   = uitoggletool('Parent',handles.GUI_Toolbar,'Separator','on' ,'Enable','off','CData',Scores,'ClickedCallback',@ScoresAnalysisMode_Callback,'TooltipString','Analyze the T-scores');
   handles.LoadingsButton = uitoggletool('Parent',handles.GUI_Toolbar,'Separator','off','Enable','off','CData',Loadings,'ClickedCallback',@LoadingsAnalysisMode_Callback,'TooltipString','Analyze the loadings');
   handles.ModelButton    = uitoggletool('Parent',handles.GUI_Toolbar,'Separator','off','Enable','off','CData',Model,'ClickedCallback',@ModelBuildingMode_Callback,'TooltipString','Interactive image and model interrogation');

   handles.HistoButton    = uipushtool('Parent',handles.GUI_Toolbar,'Separator','on' ,'Enable','off','CData',Histo,'ClickedCallback',@HistogramAnalysis_Callback,'TooltipString','Histograms of the image''s channels');

   handles.ZoomStartButton= uitoggletool('Parent',handles.GUI_Toolbar,'Separator','on' ,'Enable','off','CData',JustZoom,'ClickedCallback',@ZoomStartButton,'TooltipString','Turn zooming on or off (Ctrl-Z)');
   handles.ZoomMoreButton = uipushtool('Parent',handles.GUI_Toolbar,'Separator','off','Enable','off','CData',ZoomMore,'ClickedCallback',@ZoomChangeButton, 'TooltipString','Increase the zoom factor (up arrow key)');
   handles.ZoomLessButton = uipushtool('Parent',handles.GUI_Toolbar,'Separator','off','Enable','off','CData',ZoomLess,'ClickedCallback',@ZoomChangeButton, 'TooltipString','Decrease the zoom factor (down arrow key)');

   handles.HelpButton     = uipushtool('Parent',handles.GUI_Toolbar,'Separator','on' ,'Enable','on','CData',Help,'ClickedCallback',@HelpAbout_Callback,'TooltipString','About MACCMIA');
  %end of widget definitions
return;                                                                         % return - end of this function


% --------------------------------------------------------------------
function RemoveInterfaceWidgets
% This function removes all the widgets from the screen such as axes, drop down menus, push buttons and text strings

  global SmlUD;
  GUI_Handle = findobj(allchild(0), 'tag', 'MACC_MIA_GUI');                     % Find the handle of the MACCMIA interface
  if isempty(GUI_Handle)                                                        % If the window does not exist, then exit
    return;
  end
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles

  MyCLA(handles.ImagePlotAxes);                                                 % Remove the contents of the four axes and make them invisible
  MyCLA(handles.ScorePlotAxes);
  MyCLA(handles.ZoomImageAxes);
  MyCLA(handles.ZoomScoreAxes);
  MyCLA(handles.ColourBarAxes);

  set(handles.ImageRed,           'Visible','off');                             % The drop-downs
  set(handles.ImageGreen,         'Visible','off');
  set(handles.ImageBlue,          'Visible','off');
  set(handles.ImageMaskColour,    'Visible','off');
  set(handles.ScoreMaskColour,    'Visible','off');
  set(handles.TScoreHorizAxis,    'Visible','off','Position',SmlUD.tHorizSize);
  set(handles.TScoreVrtclAxis,    'Visible','off');

  set(handles.ImageTextRed,       'Visible','off');                             % Text elements on the GUI
  set(handles.ImageTextGreen,     'Visible','off');
  set(handles.ImageTextBlue,      'Visible','off');
  set(handles.ImageMaskColourText,'Visible','off');
  set(handles.ScoreMaskColourText,'Visible','off');
  set(handles.expScoreSideMaskSizeText,'Visible','off');
  set(handles.expImageSideMaskSizeText,  'Visible','off');
  set(handles.LHSPlotLabel,       'Visible','off');
  set(handles.RHSPlotLabel,       'Visible','off');
  set(handles.ZoomScoreInfo,      'Visible','off');
  set(handles.ZoomSpaceMessage,   'Visible','off');
  set(handles.ZoomXYLocation,     'Visible','off');

% Don't de-actuve
%  set(handles.ScoresButton,       'Enable','off');                             % Menu buttons
%  set(handles.LoadingsButton,     'Enable','off');
%  set(handles.ModelButton,        'Enable','off');

  set(handles.ResetModel,         'Visible','off');                             % Pushbuttons
  set(handles.ResetImageClicks,   'Visible','off');
  set(handles.StartAdjustMask,    'Visible','off');
  set(handles.EndAdjustMask,      'Visible','off');

  set(handles.ImageStatsList,     'Visible','off');                             % List boxes

  set(handles.GUI_ExploreScores,  'Visible','off');                             % Check boxes

  set(handles.expScoreMaskSize,   'Visible','off');                             % Edit text boxes
  set(handles.expImageMaskSize,   'Visible','off');

%  set(handles.HistoButton,        'Enable', 'off');                             % Major menu items off
%  set(handles.Tools,              'Enable', 'off');                             % Major menu items off

  set(GUI_Handle,'WindowButtonMotionFcn','');                                   % Turn off mouse and keybaord input functions
  set(GUI_Handle,'WindowButtonDownFcn'  ,'');
  set(GUI_Handle,'KeyPressFcn'          ,'');

return;

% --------------------------------------------------------------------
function HelpAbout_Callback(varargin)
% User has clicked Help/About so open a window displaying names/authors, copyright and
% contact information.  Different information is displayed if we are in Explore mode or mask adjustment mode.

  global SmlUD;
  GUI_Handle = findobj(allchild(0), 'tag', 'MACC_MIA_GUI');
  if isempty(GUI_Handle)                                                        % The window does not exist,
    return;
  end
  %handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the widget handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

  if SmlUD.adjMode
      uiwait(msgbox({'The score space masks can be adjusted so that you can isolate, with', ...
              'greater accuracy, the image space feature(s) of interest.', ...
              ' ', ...
              'After clicking the "Adjust the mask" button, the current mask changes to', ...
              'reveal its outline, together with the corner vertices.  You can adjust', ...
              'these corner points by clicking and dragging them with the mouse.', ...
              ' ', ...
              'Depress the mouse button close to or on the point you want to move,',...
              'hold the mouse button down while moving that point to its new position.', ...
              'Release the mouse button to finally place the point at that location.',  ...
              ' ',  ...
              'You need to click "End adjusting" to complete the process.',...
              ' '},...
          'Adjusting masks in MACCMIA','none','modal'));

  elseif SmlUD.expMode
      uiwait(msgbox({'The explore mode allows one to gain a rapid understanding of the ', ...
              'relationship between the image on the left and the multivariate score space.', ...
              ' ', ...
              'Moving the mouse over the image space shows where pixels within the', ...
              'square mask lie in the score space.  The size of this mask can be adjusted', ...
              'so that image space features fit within the mask.', ...
              ' ', ...
              'Perhaps more usefully, regions in the score space can be highlighted with',...
              'the mouse by moving the square mask over it.  This shows dynamically where', ...
              'those pixels lie in the original image.  This helps to rapidly construct',  ...
              'non-square masks later on.',  ...
              ' '},...
              'Explore mode in MACCMIA','none','modal'));
  else
      uiwait(msgbox({'The MACCMIA Package was designed and developed at the', ...
              'Department of Chemical Engineering, McMaster University.', ...
              ' ', ...
              'A tutorial manual (PDF file) is provided with the software. That', ...
              'manual describes the theory used in multivariate image analysis', ...
              'and also provides a step-by-step analysis on an example image.', ...
              ' ', ...
              'More information about MACCMIA and other MACC software ',...
              'is available at:    http://macc.mcmaster.ca/maccmia', ...
              ' ',  ...
              ' ',  ...
              'MACCMIA Package, Version 1.81',...
              '(c) March 2007, McMaster University', ...
              ' ', ...
              'Software designed and written by Kevin Dunn.', ...
              'Contact details: kevin.dunn@prosensus.ca', ...
              ' ', ...
              'For other questions and concerns please contact:', ...
              'Dr. John MacGregor    macgreg@mcmaster.ca', ...
              '                   or john.macgregor@prosensus.ca', ...
              ' '},...
          'About MACCMIA','none','modal'));
  end;
return;

% --------------------------------------------------------------------
function Image = GetImageCheck(FileNameToLoad,tempFolder)
% This function processes a image that the user wants to open.  The user must provide the full file name as
% input [created most conveniently as follows: "FileNameToLoad = fullfile(ImageFilePath,ImageFileName)"]
% The second input, tempFolder, is the temporary folder the user has chosen at startup to store the temporary results in.
%
% The output string contains the full file name to the location of the temporary file containing the necessary data.

% STEP 1: Check if the image exists and if so, then get the various forms of the image name
  Image = [];
  Image.ErrMsg = 'Error';
  if exist(FileNameToLoad,'file') == 2
      Image = CreateNewDataStructures(0,'Image');
  else
      uiwait(msgbox({'The selected file does not exist.  Returning to the interface.'}));
      return;
  end;
  Image.FullFileName  = FileNameToLoad;
  Image.StructureFile = tempname;                                               % Get a temporary file name to store the structure in
  [fPathstring,fName,fExtension,fVersion] =  fileparts(FileNameToLoad);         % Get the parts of the file name
  [tPathstring,tName,tExtension,tVersion] =  fileparts(Image.StructureFile);
  if isempty(tExtension)
      tExtension = '.mat';
  end;
  Image.StructureFile = fullfile(tempFolder,[tName,tExtension]);                % Replace "tempname" directory with user specified temp directory
  Image.DisplayName = [fName,fExtension];                                       % The file name to be displayed
  Image.BaseFileName = fName;                                                   % The base file name
  Image.ImageExtension = fExtension;                                            % The extension of the image type, eg: ".bmp"

  % STEP 2: Check if the can be opened
  if strcmpi(fExtension,'.mat')                                                 % The user has chosen to load a MAT file
      try
          MatFileContents  = load(FileNameToLoad);
          VariablesInFile  = fieldnames(MatFileContents);
          nVariables       = size(VariablesInFile,1);
          if nVariables==1                                                      % If there is only one variable, then this is probably the image
              Image.X = getfield(MatFileContents,VariablesInFile{1});
              VariableSelected = VariablesInFile{1};
          else
              hDlg.Parent = figure('ButtondownFcn'    ,'', ...
                  'Color'            ,get(0, 'defaultuicontrolbackgroundcolor'), ...  % 'color',get(0,'FactoryUicontrolBackgroundColor')
                  'Colormap'         ,[], ...
                  'IntegerHandle'    ,'off', ...
                  'InvertHardcopy'   ,'off', ...
                  'HandleVisibility' ,'callback', ...
                  'Menubar'          ,'none', ...
                  'NumberTitle'      ,'off', ...
                  'PaperPositionMode','auto', ...
                  'Resize'           ,'off', ...
                  'Visible'          ,'on', ...
                  'WindowStyle'      ,'modal', ...
                  'Name'            ,'Please choose variable to load', ...
                  'Position'         ,[360 514 250 180]);
              movegui(hDlg.Parent,'center');
              hDlg.VariableList = uicontrol('Interruptible','off','BusyAction','queue','String',VariablesInFile,'Units','Pixels', ...
                  'style','listbox','Parent',hDlg.Parent,'Position',[10 10 180 160], ...
                  'Callback','if strcmp(get(gcbf,''SelectionType''),''open''), set(gcbf,''Visible'',''off''), end');
              hDlg.OKbutton     = uicontrol('Interruptible','off','BusyAction','queue','String','OK','Units','Pixels', ...
                  'style','pushbutton','Parent',hDlg.Parent,'Position',[200 10 40 20],'Callback','set(gcbf,''Visible'',''off'')');
              waitfor(hDlg.Parent,'Visible');
              if ~ishandle(hDlg.VariableList)                                         % The user closed the dialog box prematurely
                  error({'The variable could not be loaded.  Please select the desired','variable and then push "OK".'},'Error selecting variable');
              end;
              VariableSelected = get(hDlg.VariableList,'Value');
              Image.X = getfield(MatFileContents,VariablesInFile{VariableSelected});
              if ishandle(hDlg.Parent), delete(hDlg.Parent); end;
          end; %end: if-elseif-elseif-else-end block
          Image.SourceVariableName = VariableSelected;
          clear MatFileContents;                                                    % Clear unnecessary variables for memory reasons
      catch
          errordlg({'There was an error in retrieving the image data from the MAT file.  Please', ...
                  'check the MAT file to ensure that there are variables saved in the MAT', ...
                  'file and that the MAT file can be loaded correctly.  You can see the', ...
                  'Matlab error message that caused this error by typing the following',...
                  'command at the Matlab prompt:',' ', ...
                  ['load ',FileNameToLoad], ' '},...
              'Error in MAT file');
          Image.ErrMsg = 'Error';
          return;
      end; %end: try-catch-end block (in the "strcmp(lower(fExtension),'.mat')" section)
  else                                                                          % The user has chosen to load a non-MAT file (any other image file)
      try
          Image.X = imread(FileNameToLoad);
      catch
          errordlg({'There was an error reading your image file.  The following command was', ...
                  'supplied to Matlab but failed:',' ', ...
                  ['imread(''',FileNameToLoad,''')'], ' ', ...
                  'You should type this command in Matlab and try to reproduce the error.', ...
                  'Matlab''s error message may help to better diagnose the problem.', ' ', ...
                  'You should also type "help fileformats" at the Matlab prompt to', ...
                  'view the different image types that can be opened with "imread".'}, ...
              'Error reading image file');
          Image.ErrMsg = 'Error';
          return;
      end; %end: try-catch-end block
  end;   %end: if strcmp(lower(fExtension),'.mat') - else - end block

  % Step 3: Check if the file's contents is a "good image" and save the image's characteristics in the "Image" structure
  Image.Class = class(Image.X);
  if ~isa(Image.X,'uint8') && ~isa(Image.X,'double')
      errordlg({'The input image should be of type "uint8" or type "double".',...
              'If you are not sure how to obtain such an image type, please',...
              'do not hesitate to contact MACC.  Please see the PDF manual', ...
              'for contact details'}, ...
          'Input file error','modal');
      Image.ErrMsg = 'Error';
      return;
  end;
  ImageSize = size(Image.X);                                                    % Get the size and layers of the image
  if (length(ImageSize) == 3)                                                   % ie: a single image
      if ImageSize(3) == 2
          errordlg({'The input image should have at least three channels or layers.',...
                  'The current image only has two.  In order to display images in ',...
                  'this multivariate image analysis software we require three or more' ...
                  'images layers (channels).'}, ...
              'Input file error','modal');
          Image.ErrMsg = 'Error';
          return;
      end
      Image.nRows = ImageSize(1);                                                 % Store size information for future use
      Image.nCols = ImageSize(2);
      Image.nLayers = ImageSize(3);
  else
      errordlg({'The input image must have three dimensions, as follows: ',...
              'Dimension 1: The pixel information up and down the vertical direction', ...
              'Dimension 2: The pixel information across the horizontal direction',...
              'Dimension 3: The the number of layers (or channels) [must be greater than one]',...
              ' ', ...
              'For example a standard 640 by 480 digital camera image with red, ', ...
              'green and blue (RBG) channels will have dimensions 480 x 640 x 3.', ...
              'The longer dimension is the horizontal direction, the shorter one', ...
              'is the vertical direction.',' ', ...
              'Other images such as those from near infrared cameras will have',...
              'many more channels, for example, dimensions of 1024 x 1024 x 16.'}, ...
          'Input file error','modal');
      Image.ErrMsg = 'Error';
      return;
  end;
return;


% --------------------------------------------------------------------
function FileOpen_Callback(varargin)
% This function opens a new data file containing image information, and does some pre-processing
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Check if there is any active information on the screen already, open projects etc and get the user to save or discard these
% Step 2. Load the MAT-file and determine its contents (variable names).  If there is more than one variable in the MAT-file then the
%         variable names are displayed to the user to select the matrix containing the image data
% Step 3. Check the dimensions of the image data to conform with the following
%         For now we will require the image to be pre-saved in a MAT-file of the following dimensions (size command):
%         [ nRows, nCols, nLayers ]
%          where: nRows: The number of pixels up and down the VERTICAL direction
%                 nCols: The number of pixels across the HORIZONTAL direction
%                 nLayers: The number of layers or channels in the image: RGB layers or wavelenth channels, etc
% Step 4. Unfold the image and get PCA information

% STEP 0
  global SmlUD;
  GUI_Handle = findobj(allchild(0), 'tag', 'MACC_MIA_GUI');
  if isempty(GUI_Handle)                                                        % The window does not exist,
    return;
  end
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the widget handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  OLDSmlUD = SmlUD;

% STEP 1: Check if any active information on screen already, save or discard this, then open a new project
  if get(handles.GUI_ExploreScores,'Value')                                     % Quit if we are in Explore Mode
     errordlg({'Please end Explore mode first before continuing'},'End Explore mode','modal');
     return;
  end;
% SAVING: not supported in this MACCMIA version
%  if ~OLDSmlUD.DataHasBeenSaved                                                 % Unsaved data exists
%    DestroyOldModel = questdlg('Do you wish to discard the current model without saving ?','Save model information ?','Yes','No','No');
%    if strcmp(DestroyOldModel,'No');
%       return;
%    end;
%  end;
  RemoveInterfaceWidgets;                                                       % User doesn't want to save - start over
  set(GUI_Handle,'Name','MACC Multivariate Image Analysis Package (MACCMIA)');
  set(handles.ScoresButton,       'State','off');                               % Turn off all analysis modes
  set(handles.LoadingsButton,     'State','off');
  set(handles.ModelButton,        'State','off');

  set(handles.FileSaveButton,     'Enable','off');                              % Turn off the non applicable menu options
  set(handles.FilePrintButton,    'Enable','off');
  set(handles.HistoButton,        'Enable','off');
  set(handles.FileSave,           'Enable','off');
  set(handles.FilePrint,          'Enable','off');
  set(handles.Tools,              'Enable','off');

  SmlUD = CreateNewDataStructures(GUI_Handle,'main');                           % Get clean data storage structures
  SmlUD.InitialDir = OLDSmlUD.InitialDir;                                       % but transfer over the necessary information
  SmlUD.FileNamePath = OLDSmlUD.FileNamePath;
  SmlUD.tempFolder = OLDSmlUD.tempFolder;
  SmlUD.CurrentDateTime = OLDSmlUD.CurrentDateTime;
  SmlUD.tHorizSize = OLDSmlUD.tHorizSize;
  clear('OLDSmlUD');                                                            % Consolidate memory

  hHisto = findobj(0,'tag','HistogramAnalysis');
  if ishandle(hHisto)
      delete(findobj(0,'tag','HistogramAnalysis'));
  end;

% STEP 2:  Let user click and select a MAT-file or any other type image file
  prevDir = pwd;
  cd(SmlUD.FileNamePath{2});
  
  if isunix
      temp_value = getappdata(0,'UseNativeSystemDialogs');
      setappdata(0,'UseNativeSystemDialogs',0);
  end
  [ImageFileName,ImageFilePath] = uigetfile( ...
       {'*.*',   'Any supported image file format (*.*)'; ...
        '*.mat', 'MAT file: an image inside a Matlab MAT file'; ...
        '*.bmp', 'BMP: Windows bitmap files (*.bmp)'; ...
        '*.gif', 'GIF image files (*.gif)'; ...
        '*.hdf', 'HDF image files (*.hdf)'; ...
        '*.jpeg;*.jpg','JPEG image files (*.jpeg and *.jpg)'; ...
        '*.pcx', 'PCX image files (*.pcx)'; ...
        '*.png', 'PNG image files (*.png)'; ...
        '*.tiff;*.tif','TIFF image files (*.tiff and *.tif)'}, ...
        'Pick select a file containing an image ...');
    
  if isunix
      setappdata(0,'UseNativeSystemDialogs',temp_value);
  end
  cd(prevDir);
  if isequal(ImageFileName,0)||isequal(ImageFilePath,0)
   % errordlg({'File not found'},'File Error','modal');                         % This is not always true, user could have clicked "Cancel"
     return;
  end;

  FileNameToLoad = fullfile(ImageFilePath,ImageFileName);                       % NOTE: This line is operating system dependent: NOT ANY MORE
  SmlUD.Image = GetImageCheck(FileNameToLoad,SmlUD.tempFolder);                 % Open and check the image: return file name to structure
  if ~isempty(SmlUD.Image.ErrMsg)                                               % There was an error opening this image
      return;
  else
      SmlUD.FileNamePath = {ImageFileName,ImageFilePath};                       % Store the file name and path in the data storage
  end;

% STEP 3: Check further information about the image
  NumR = SmlUD.Image.nRows;
  NumC = SmlUD.Image.nCols;
  NumZ = SmlUD.Image.nLayers;
  [SmlUD.Image.nPC,SmlUD.RGBmode] = imageinfo(NumR,NumC,NumZ,SmlUD.Image.Class,'setup');
  SmlUD.XLim = [0 NumC] + 0.5;                                                  % The zoom extents used to display the image
  SmlUD.YLim = [0 NumR] + 0.5;
  if strcmp(SmlUD.Image.Class,'uint8') && (SmlUD.Image.nLayers == 3)
      SmlUD.RGBmode = 1;
  else
      SmlUD.RGBmode = 0;
  end;
  set(GUI_Handle,'Name',strcat(['MACC Multivariate Image Analysis Package (MACCMIA) : ',SmlUD.FileNamePath{1}]));
  set(GUI_Handle,'Pointer','watch');

% STEP 4: Unfold the image and get PCA information  - code based on "singlesvd.m" (Honglu Yu)
        % Also computes the T score images using    - code based on "CalcScore.m" (Honglu Yu)

  [a,b,c,d,e,f,g] = libmia('imagepca',SmlUD.Image.X,SmlUD.Image.nPC);
  [dummy,j,h,i,k] = libmia('stats',SmlUD.Image.X);

  SmlUD.Image.XTX           = a;                                                % The X^T X matrix
  SmlUD.Image.P             = b;                                                % Store the SVD information
  SmlUD.Image.Eigenvalues   = c;                                                % The eigenvalues along the main diagonal
  SmlUD.Image.T             = uint8(d);  clear('d');                            % The T-scores as images
  SmlUD.Image.minT          = e;                                                % The minimum value used to scale the T-scores
  SmlUD.Image.maxT          = f;                                                % The maximum value used to scale the T-scores
  SmlUD.Image.scaleT        = g;                                                % The scaling vector applied to the T-scores
  SmlUD.Image.minX          = h;                                                % The minimum pixel intensity in each image layer
  SmlUD.Image.maxX          = i;                                                % The maximum pixel intensity in each image layer
  SmlUD.Image.avgX          = j;                                                % The average pixel intensity in each image layer
  SmlUD.Image.varX          = k;                                                % The pixel intensity variance in each image layer

  SmlUD.DataHasBeenSaved = 0;                                                   % Storage data has been changed but not saved to file yet
  setappdata(GUI_Handle,'Small_Data', SmlUD);
  set(GUI_Handle,'Pointer','arrow');

  set(handles.FileSaveButton,      'Enable','on');                             % Set previously disabled push-buttons back on
  set(handles.FilePrintButton,     'Enable','on');
  set(handles.HistoButton,         'Enable','on');
  set(handles.ScoresButton,        'Enable','on');
  set(handles.LoadingsButton,      'Enable','on');
  set(handles.ModelButton,         'Enable','on');

  set(handles.FileSave,            'Enable','on');                             % Set previously disabled menu options back on
  set(handles.FilePrint,           'Enable','on');
  set(handles.Tools,               'Enable','on');

  set(handles.ScoresButton,'State','on');                                       % Call "ScoresAnalysis" to kick-start the user's analysis by first
  ScoresAnalysisMode_Callback(GUI_Handle);                                      % turning on the icon and then calling the function
return;

% --------------------------------------------------------------------
function FileSave_Callback(varargin)

% This function saves the data that is currently being worked on
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Get a file name that the user wants to save to
% Step 2. Generate the saved data structure
% Step 3. Save the data structure

% STEP 0
  GUI_Handle = findobj(allchild(0), 'tag', 'MACC_MIA_GUI');
  if isempty(GUI_Handle)                                                        % The window does not exist,
    return;
  end
  SmlUD   = getappdata(GUI_Handle,'Small_Data');                                % Get the smaller data storage structure
  %handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the widget handles

% STEP 1: Get a file name and path
  [DataFileName, DataPathName] = uiputfile('*.mat', 'Save data file as a MAT-file');
  if isequal(DataFileName,0)||isequal(DataPathName,0)
      uiwait(msgbox('There was a problem saving file, please try again.','Saving error','modal'));
     return;
  end;

% STEP 2: Generate the saved data structure

    MACCMIA.Version                 = 'MACCMIA - Version 1.80';

    MACCMIA.Image.Filename          = SmlUD.Image.FullFileName;
    MACCMIA.Image.nRows             = SmlUD.Image.nRows;
    MACCMIA.Image.nCols             = SmlUD.Image.nCols;
    MACCMIA.Image.nLayers           = SmlUD.Image.nLayers;
    MACCMIA.Image.cdata             = SmlUD.Image.X;

    MACCMIA.PCAmodel.numPC          = SmlUD.Image.nPC;
    MACCMIA.PCAmodel.Loadings       = SmlUD.Image.P;
    MACCMIA.PCAmodel.Eigenvalues    = SmlUD.Image.Eigenvalues;
    MACCMIA.PCAmodel.XTXmatrix      = SmlUD.Image.XTX;
    MACCMIA.PCAmodel.Tmin           = SmlUD.Image.minT;
    MACCMIA.PCAmodel.Tmax           = SmlUD.Image.maxT;
    MACCMIA.PCAmodel.TScoresAsImages= SmlUD.Image.T;

    MACCMIA.ScoreSpace.ColourMap    = SmlUD.ScoreSpaceColourMap;

  if isempty(SmlUD.Image.MaskXdata)
     MACCMIA.ScoreSpace.Mask        = 'A mask was not defined or was reset during model building';
  else
     MACCMIA.ScoreSpace.Mask        = SmlUD.Image.Mask;
     MACCMIA.ScoreSpace.Xpoints     = SmlUD.Image.MaskXdata;
     MACCMIA.ScoreSpace.Ypoints     = SmlUD.Image.MaskYdata;
  end;


% STEP 3: Save the data structure
  CurrentDir = pwd;
  cd(DataPathName);
  if isempty(findstr('.mat',DataFileName))
    DataFileName = [DataFileName,'.mat'];
  end;
  save(DataFileName,'MACCMIA','-mat');
  cd(CurrentDir);

  SmlUD.DataHasBeenSaved = 1;
  setappdata(GUI_Handle,'Small_Data',SmlUD);
return;

% --------------------------------------------------------------------
function FilePrint_Callback(varargin)
% This function just gives a message that right-clicking on any figure axes will bring up options to
% send that figure to a new window for printing.
  uiwait(msgbox({'Right click on any image or plot to send it to a new figure window.', ...
                 'One can then adjust labels and other figure settings before using the', ...
                 'standard Matlab commands to print that window.',' ', ...
                 'For example, one can then save the new figure window to a Matlab .FIG file',...
                 'or send it to a printer or copy and paste the image to another program.'},'Printing in MACCMIA','modal'));


% --------------------------------------------------------------------
function varargout = FileExit_Callback(varargin)
% The user has clicked File/Close, or the little "X" in the top right hand corner
% 1. Confirm the user's choice to close Score Tool, if not quitting then return, otherwise steps 2, 3 and 4.
% 2. Restore the image processing toolbox's default settings
% 3. Return to the starting directory, if different from the current directory
% 4. Delete the handles and thereby exit the interface

% ALL STEPS
  global SmlUD;
  varargout = [];
  GUI_Handle = findobj(allchild(0), 'tag', 'MACC_MIA_GUI');
  if isempty(GUI_Handle)                                                        % The window does not exist,
      return;
  end
  handles = getappdata(GUI_Handle,'Handles_Data');
  if get(handles.GUI_ExploreScores,'Value')                                     % Error: we are in Explore Mode
      uiwait(errordlg({'Please end Explore mode first before continuing'},'End Explore mode','modal'));
      return;
  end;

  CloseYNSelection = questdlg('Close MIA session ?','Close MIA session ?','Yes','No','Yes');
  if strcmp(CloseYNSelection,'No');
      return;
  else
      if isempty(SmlUD)
          SmlUD = getappdata(GUI_Handle,'Small_Data');                          % Get the data storage structure
      end;

      if isstruct(SmlUD)
          hHisto = findobj(0,'tag','HistogramAnalysis');
          if ishandle(hHisto)
              delete(findobj(0,'tag','HistogramAnalysis'));
          end;



%   DELETE TEMPORARY FILES
%           if ~isempty(SmlUD.tstFullImages)
%               warnstate = warning;                                              % Delete previous testing images
%               warning off;
%               for k = 1:size(SmlUD.tstFullImages,1)
%                   try
%                       if exist(SmlUD.tstFullImages(k,:))==2                     % Only delete files
%                           delete(SmlUD.tstFullImages(k,:));
%                       end;
%                   catch
%                       disp(['Warning: One of the previous testing images that was created (',SmlUD.tstFullImages(k,:),' could not be removed.']);
%                   end;
%               end;
%               warning(warnstate);
%           end;

          cd(SmlUD.InitialDir);                                                 % Change back to the starting directory
          rmpath(SmlUD.InitialDir);                                             % Remove the Score Tool directory from the Matlab path

          if ishandle(GUI_Handle), delete(GUI_Handle); end;                     % delete the figure window
      else                                                                      % Cannot find SmlUD, but
          if ishandle(GUI_Handle), delete(GUI_Handle); end;                     % just delete the figure window anyway and quit
      end;
  end;
  clear global SmlUD
return;



% --------------------------------------------------------------------
function ScoresAnalysisMode_Callback(varargin)
% The user wishes to analyse the the scores, presented as images of the scores as well as residual images
% This will be done by presenting to the user two plots:
% Left side plot:  Present to the user a composite RGB (the first three PC's correspond to R G and B respectively
% Right side plot: Present the principal components and residuals with predefined colourmaps
%
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Set the other radio buttons to have zero values and make invisible the GUI elements that are not required
% Step 2. Set up the left hand side image of the true RGB image or the the scores viewed as an RGB image
% Step 3. Set up the right hand side image containing the individual score plots and residual plots, with a colour bar
% Step 4. Call UpdateLHS and UpdateRHS to complete the update

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  if get(handles.GUI_ExploreScores,'Value')                                     % We first need to exit Explore Mode if we are in Explore Mode
     set(handles.ScoresButton,'State','off');
     errordlg({'Please end Explore mode first before continuing'},'End Explore mode','modal');
     return;
  end;

% STEP 1
  if strcmp(get(handles.ZoomStartButton,'State'),'on')                          % If the user came here with zoom mode on, turn it off, then continue
      set(handles.ZoomStartButton,'State','off');
      set(handles.ModelButton,    'State','off');                               % We need to turn off the model building mode so that the zoom function
      ZoomStartButton(GUI_Handle);                                              % cleans up correctly
  end;
  RemoveInterfaceWidgets;                                                       % Clean up the interface before we start
  if gcbo==handles.ToolsScoreAnalysis                                           % The user has access this callback from the Menu bar
     set(handles.ScoresButton,'State','on');
  end;
  set(handles.ToolsScoreAnalysis,'Checked','on');                               % Turn on the Scores check mark
  set(handles.ToolsLoadgAnalysis,'Checked','off');                              % and turn off the other check marks
  set(handles.ToolsModelBuilding,'Checked','off');

  if ~strcmp(get(handles.ScoresButton,'State'),'on')                            % The user has pressed the scores button (we are currently in scores mode though)
     set(handles.ScoresButton,'State','on');                                    % so turn it back on and return
     return;
  end;

  set(handles.LoadingsButton,'State','off');
  set(handles.ModelButton,   'State','off');
  set(GUI_Handle,'WindowButtonMotionFcn','');
  set(GUI_Handle,'WindowButtonDownFcn'  ,'');

  if ishandle(get(handles.ImageContextMenu,'Children'))                         % If the uicontextmenu has children
    delete(get(handles.ImageContextMenu,'Children'))                            % then delete them and replace them with new ones
  end;
  uimenu(handles.ImageContextMenu,'Label','Send this axis to a new figure window','Callback',@SendLHSImageToNewFigureWindow);

  if ishandle(get(handles.ScoreContextMenu,'Children'))                         % If the uicontextmenu has children
    delete(get(handles.ScoreContextMenu,'Children'))                            % then delete them and replace them with new ones
  end;
  uimenu(handles.ScoreContextMenu,'Label','Send this axis to a new figure window','Callback',@SendRHSImageToNewFigureWindow);

% STEP 2 - The LHS: RGB colour image and/or RGB score image
  if SmlUD.RGBmode                                                              % Show in true colour, if it is RGB image
     PopupString = {'R','G','B'};
  else                                                                          % else, show it as a composite RGB image
     PopupString = 'T 1';                                                       % and while we are here, put the list box values in
     for k = 2:min(SmlUD.Image.nPC,SmlUD.Image.nLayers)
        PopupString = strvcat(PopupString,['T ',num2str(k)]);
     end;
     if k==2
        PopupString = strvcat(PopupString,'Layer 3');
     end;
  end;
  set(handles.ImageRed,  'string',PopupString,'Value',SmlUD.RLayer,'Visible','on'); % set the list box values
  set(handles.ImageGreen,'string',PopupString,'Value',SmlUD.GLayer,'Visible','on'); % to their prior values
  set(handles.ImageBlue, 'string',PopupString,'Value',SmlUD.BLayer,'Visible','on'); % and make them visible
  set(handles.ImageTextRed,  'Visible','on');                                   % Also make their titles visible
  set(handles.ImageTextGreen,'Visible','on');
  set(handles.ImageTextBlue, 'Visible','on');
  set(handles.LHSPlotLabel,  'Visible','on');                                   % the title of the image space
  set(handles.ZoomStartButton,'Enable','on');                                   % Enable zooming in and out in this mode


% STEP 3 - The LHS: individual scores and residuals with colourbars
  set(handles.RHSPlotLabel,'Visible','on');
  PopupString = 'Principal Component 1';                                        % Display the correct dropdown box content
  for k = 2:min(SmlUD.Image.nPC,SmlUD.Image.nLayers)
    PopupString = strvcat(PopupString,['Principal Component ',num2str(k)]);
  end;
  for k = 1:min(SmlUD.Image.nPC,SmlUD.Image.nLayers)
    PopupString = strvcat(PopupString,['Residual after ',num2str(k),' PC extracted']);
  end;
  set(handles.TScoreHorizAxis,'Value',1);                                       % Set the value to the default
  set(handles.TScoreHorizAxis,'String',PopupString,'Visible','on');             % and make it visible
  positn = get(handles.TScoreHorizAxis,'Position');
  extent = get(handles.TScoreHorizAxis,'Extent');
  if extent(3) > positn(3)                                                      % The text in the drop-down is cut off on the RHS edge
      diff = extent(3) - positn(3);
      positn(3) = positn(3) + 1.20*diff;                                        % Increase it to 120% of what is required
      positn(1) = positn(1) - 0.60*diff;                                        % But shift the GUI element to the left by half the additional adjustment
      set(handles.TScoreHorizAxis,'Position',positn);
  end;

% STEP 4: Call UpdateLHS and UpdateRHS to complete the update
  setappdata(handles.GUI_Handle,'Small_Data',SmlUD);
  set(GUI_Handle,'KeyPressFcn',@GUIKeyPress_Callback);
  UpdateLHS(GUI_Handle);
  UpdateRHS(GUI_Handle);
return;


% --------------------------------------------------------------------
function LoadingsAnalysisMode_Callback(varargin)
% The user wishes to analyse the the loadings, presented as a bar plot of percentage explained by each principal component
% on the left and on the right we have bar plots and scatter plots of the loading vectors themselves
%
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Set the other radio buttons to have zero values and make invisible the GUI elements that are not required
% Step 2. Draw the percentage explained barplot in the left hand side axes
% Step 3. Draw plots of the loadings against each other (scatter plot) and against themselves (bar plot)
% Step 4. Give some statistics regarding the image in the list box

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  if get(handles.GUI_ExploreScores,'Value')                                     % We first need to exit Explore Mode if we are in Explore Mode
     set(handles.LoadingsButton,'State','off');
     errordlg({'Please end Explore mode first before continuing'},'End Explore mode','modal');
     return;
  end;

% STEP 1
  if strcmp(get(handles.ZoomStartButton,'State'),'on')                          % If the user came here with zoom mode on, turn it off, then continue
    set(handles.ZoomStartButton,'State','off');
    set(handles.ModelButton,   'State','off');                                  % We need to turn off the model building mode so that the zoom function
    ZoomStartButton(GUI_Handle);                                                % cleans up correctly
  end;

  RemoveInterfaceWidgets;                                                       % Clean up the interface before we start
  if gcbo==handles.ToolsLoadgAnalysis                                           % The user has access this callback from the Menu bar
     set(handles.LoadingsButton,'State','on');
  end;
  set(handles.ToolsLoadgAnalysis,'Checked','on');                               % Turn on the Loadings check mark
  set(handles.ToolsScoreAnalysis,'Checked','off');                              % and turn off the other check marks
  set(handles.ToolsModelBuilding,'Checked','off');

  if ~strcmp(get(handles.LoadingsButton,'State'),'on')                          % The user has pressed the loadings button (we are currently in loadings analysis mode though)
    set(handles.LoadingsButton,'State','on');                                   % so turn it back on and return
    return;
  end;
  set(handles.ScoresButton,      'State','off');
  set(handles.ModelButton,       'State','off');
  set(handles.ZoomStartButton,   'Enable','off');                               % Disable zooming in and out in this mode
  set(handles.ZoomMoreButton,    'Enable','off');
  set(handles.ZoomLessButton,    'Enable','off');

  if ishandle(get(handles.ImageContextMenu,'Children'))                         % If the uicontextmenu has children
    delete(get(handles.ImageContextMenu,'Children'))                            % then delete them and replace them with new ones
  end;
  uimenu(handles.ImageContextMenu,'Label','Send this axis to a new figure window','Callback',@SendLHSImageToNewFigureWindow);

  if ishandle(get(handles.ScoreContextMenu,'Children'))                         % If the uicontextmenu has children
    delete(get(handles.ScoreContextMenu,'Children'))                            % then delete them and replace them with new ones
  end;
  uimenu(handles.ScoreContextMenu,'Label','Send this axis to a new figure window','Callback',@SendRHSImageToNewFigureWindow);

% STEP 2 :
  UpdateLHS(GUI_Handle);


% STEP 3. Loading plots
  DropDownString = 'P 1';
  for k = 2:(SmlUD.Image.nLayers)
     DropDownString = strvcat(DropDownString,['P ',num2str(k)]);
  end;
  set(handles.TScoreHorizAxis','Visible','on','String',DropDownString);         % Make the drop down menus contain P1, P2, etc
  set(handles.TScoreVrtclAxis','Visible','on','String',DropDownString);         % and make them visible
  set(handles.TScoreHorizAxis','Value',1);                                      % Show the loadings for principal component 1
  set(handles.TScoreVrtclAxis','Value',1);
  set(handles.RHSPlotLabel,    'Visible','on');
  set(handles.ScorePlotAxes,'Visible','on');                                    % Make the right hand side axes visible, and then call
  UpdateRHS(GUI_Handle);
%  ScoreOrderChange_Callback('Horiz');                                           % ScoreOrderChange_Callback to draw the plots

% STEP 4: Image statistics
  ListboxString = {'           IMAGE STATISTICS','-------------------------------------','Channel   Min   Max   Mean   Std.Dev. '};
  for k = 1:SmlUD.Image.nLayers
    NewLine = num2str(k);
    for j = 1:10-length(NewLine)
      NewLine = [NewLine,' '];
    end;
    NewLine = [NewLine,num2str(SmlUD.Image.minX(k))];
    for j = 1:16-length(NewLine)
      NewLine = [NewLine,' '];
    end;
    NewLine = [NewLine,num2str(SmlUD.Image.maxX(k))];
    for j = 1:22-length(NewLine)
      NewLine = [NewLine,' '];
    end;
    NewLine = [NewLine,num2str(SmlUD.Image.avgX(k),4)];
    for j = 1:29-length(NewLine)
      NewLine = [NewLine,' '];
    end;
    NewLine = [NewLine,num2str(sqrt(SmlUD.Image.varX(k)),6)];
    ListboxString = [ListboxString,cell({NewLine})];
  end;

  ListboxString  = [ListboxString,cell({'','             LOADING VALUES','-------------------------------------'})];
  LinesSoFar     = length(ListboxString);
  NewLoadingLine = 'PC Num.  ';                                                 % We will insert this line later

  for k = 1:min(SmlUD.Image.nLayers,SmlUD.Image.nPC)                            % The outer loop goes down the rows
    NewLine = num2str(k);
    for j = 1:9-length(NewLine)
      NewLine = [NewLine,' '];
    end;
    for WhichPC = 1:min(SmlUD.Image.nLayers,SmlUD.Image.nPC)                    % This loop goes across each column
      if SmlUD.Image.P(k,WhichPC) >= 0
        NewLine = [NewLine,' ',num2str(SmlUD.Image.P(k,WhichPC),'%0.4f'),'  '];   % If it is positive, leave a leading space
      else
        NewLine = [NewLine,num2str(SmlUD.Image.P(k,WhichPC),'%0.4f'),'  '];
      end;
    end;
    ListboxString  = [ListboxString,cell({NewLine})];
    NewLoadingLine = [NewLoadingLine,' PC ',num2str(k)];
    for j = 1:9+k*9-length(NewLoadingLine)
      NewLoadingLine = [NewLoadingLine,' '];
    end
  end;
  ListboxString = [ListboxString(1:LinesSoFar) cell({NewLoadingLine}) ListboxString(LinesSoFar+1:end)];
  set(handles.ImageStatsList,'String',ListboxString,'Visible','on','Value',1);
return;


% --------------------------------------------------------------------
function ModelBuildingMode_Callback(varargin)
% The user wishes to build a model between the image space and the score space
% This will be done by switching between two plots or spaces: The image space (left side axes) and the score space (right side axes)
%
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Set the other radio buttons to have zero values and make invisible the GUI elements that are not required
% Step 2. Set up the left hand side image of the true RGB image or the the scores viewed as a composite RGB image
% Step 3. Set up the right hand side image containing the score space


% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  if get(handles.GUI_ExploreScores,'Value')                                     % We first need to exit Explore Mode if we are in Explore Mode
     set(handles.ModelButton,'State','off');
     errordlg({'Please end Explore mode first before continuing'},'End Explore mode','modal');
     return;
  end;

% STEP 1
  if gcbo==handles.ToolsModelBuilding                                           % The user has access this callback from the Menu bar
     set(handles.ModelButton,'State','on');
  end;
  set(handles.ToolsModelBuilding,'Checked','on');                               % Turn on the model-building check mark
  set(handles.ToolsLoadgAnalysis,'Checked','off');                              % and turn off the other check marks
  set(handles.ToolsScoreAnalysis,'Checked','off');

  if ~strcmp(get(handles.ModelButton,'State'),'on')                             % The user has pressed the model building button (we are currently in model building mode though)
    set(handles.ModelButton,'State','on');                                      % turn it back on and return
    return;                                                                     % It does not make sense to turn it off
  end;
  if strcmp(get(handles.ZoomStartButton,'State'),'on')                          % If the user came here with zoom mode on, turn it off, then continue
    set(handles.ZoomStartButton,'State','off');
    set(handles.ModelButton,'State','off');                                     % Trick the ZoomStartButton function to think we are not is ModelBuildingMode
    ZoomStartButton(GUI_Handle);                                                % so that it can clean up from the previous mode's zoom settings
    set(handles.ModelButton,'State','on');                                      % Then reset the ModelBuildingMode back on again
  end;
  set(handles.ScoresButton,       'State','off');
  set(handles.LoadingsButton,     'State','off');
  set(handles.ZoomStartButton,    'Enable','on');                               % Enable zooming in and out in this mode
  RemoveInterfaceWidgets;                                                       % Clean up the interface before we start: put this here: the ZoomStartButton command above still makes changes to the interface

  if ishandle(get(handles.ImageContextMenu,'Children'))                         % If the uicontextmenu has children
    delete(get(handles.ImageContextMenu,'Children'))                            % then delete them and replace them with new ones
  end;
  uimenu(handles.ImageContextMenu,'Label','Send this axis to a new figure window','Callback',@SendLHSImageToNewFigureWindow);
  uimenu(handles.ImageContextMenu,'Label','Send both image and score space to a new figure window','Callback',@SendBothToNewFigureWindow);

  if ishandle(get(handles.ScoreContextMenu,'Children'))                         % If the uicontextmenu has children
    delete(get(handles.ScoreContextMenu,'Children'))                            % then delete them and replace them with new ones
  end;
  uimenu(handles.ScoreContextMenu,'Label','Send this axis to a new figure window','Callback',@SendRHSImageToNewFigureWindow);
  uimenu(handles.ScoreContextMenu,'Label','Send both image and score space to a new figure window','Callback',@SendBothToNewFigureWindow);

% STEP 2
  set(handles.ImagePlotAxes,'Visible','off');                                   % Make the left hand side axes visible,
  if SmlUD.RGBmode                                                              % Show in true colour, if it is RGB image
     PopupString = {'R','G','B'};
  else                                                                          % else, show it as a composite RGB image
     PopupString = 'T 1';                                                       % and while we are here, put the list box values in
     for k = 2:min(SmlUD.Image.nPC,SmlUD.Image.nLayers)
        PopupString = strvcat(PopupString,['T ',num2str(k)]);
     end;
  end;
  set(handles.ImageRed,  'string',PopupString,'Value',SmlUD.RLayer,'Visible','on'); % set the list box values
  set(handles.ImageGreen,'string',PopupString,'Value',SmlUD.GLayer,'Visible','on'); % to their prior values
  set(handles.ImageBlue, 'string',PopupString,'Value',SmlUD.BLayer,'Visible','on');
  set(handles.ImageTextRed,       'Visible','on');
  set(handles.ImageTextGreen,     'Visible','on');                              % Also make their titles visible
  set(handles.ImageTextBlue,      'Visible','on');
  set(handles.LHSPlotLabel,       'Visible','on');                              % the title of the image space
  set(handles.ImageMaskColour,    'Visible','on','Value',1);
  set(handles.ImageMaskColourText,'Visible','on');
  set(handles.ScoreMaskColour,    'Visible','on','Value',6);
  set(handles.ScoreMaskColourText,'Visible','on');
  set(handles.ResetModel,         'Visible','on');
  set(handles.ResetImageClicks,   'Visible','on');

  UpdateLHS(GUI_Handle);

% STEP 3
  set(handles.ScorePlotAxes,'Visible','off');                                   % Make the left hand side axes invisible: no black edge
  PopupString = 'T 1';                                                          % Display the correct dropdown box content
  for k = 2:min(SmlUD.Image.nPC,SmlUD.Image.nLayers)
    PopupString = strvcat(PopupString,['T ',num2str(k)]);
  end;
  set(handles.TScoreHorizAxis,'Value' ,SmlUD.hAxis,'String',PopupString);       % Set the values to their defaults
  set(handles.TScoreVrtclAxis,'Value' ,SmlUD.vAxis,'String',PopupString);       % and set their popup strings
  set(handles.TScoreHorizAxis,'Visible','on');                                  % then make the required GUI elements visible
  set(handles.TScoreVrtclAxis,'Visible','on');
  set(handles.RHSPlotLabel,   'Visible','on');

  UpdateRHS(GUI_Handle);

% STEP 4 - setup information for mouse motion in the Model Building stage
  set(handles.GUI_ExploreScores,'Visible','on');
  set(handles.ZoomStartButton,'State','on');                                    % Set zoom mode on
  ZoomStartButton(GUI_Handle);                                                  % then call the function to kick-start the zoom mode

%  set(GUI_Handle,'WindowButtonDownFcn'  ,@ModelBuilding_MouseButton);  <-- delete line
return;



% --------------------------------------------------------------------
function UpdateLHS(varargin)
% This function is called when the LHS axes need to be updated.  The function determines that GUI's state
% such as Explore mode, showing T-scores, etc and updates the GUI accordingly.  This function is called by
% all three of Red, Green and Blue callbacks, because the code for every colour is the same.

% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1: Get the colour values from the popup boxes
% Step 2: Display the title for the image
% Step 3a: Redraw the LHS axis image: T-score mode
% Step 3a: Redraw the LHS axis image: Model building mode

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1:  Get the colour values from the popup boxes
  SmlUD.RLayer = get(handles.ImageRed,  'Value');                               % Get the current value for red
  SmlUD.GLayer = get(handles.ImageGreen,'Value');                               % Get the current value for green
  SmlUD.BLayer = get(handles.ImageBlue, 'Value');                               % Get the current value for blue
  setappdata(handles.GUI_Handle, 'Small_Data', SmlUD);                          % Save this new data in the small data structure
  L1 = SmlUD.RLayer;                                                            % Get the current settings
  L2 = SmlUD.GLayer;
  L3 = SmlUD.BLayer;

  handleAxis = handles.ImagePlotAxes;
  labelAxis  = handles.LHSPlotLabel;
  if nargin>1
      if ishandle(varargin{2})
          handleAxis = varargin{2};
      end;
  end;


% STEP 2: Display the correct title
  if SmlUD.RGBmode                                                              % If in RGB mode
      if (L1 == 1) && (L2 == 2) && (L3 == 3)                                    % The RGB combinations are in correct order
          set(labelAxis,'String','The RGB image from the data file');
      elseif (L1==L2) && (L2==L3) && (L3==L1)                                   % all layers are either R or G or B
          switch L1
          case 1
              set(labelAxis,'String','The red layer as a grayscale image');
          case 2
              set(labelAxis,'String','The green layer as a grayscale image');
          case 3
              set(labelAxis,'String','The blue layer as a grayscale image');
          end;
      else
          set(labelAxis,'String','The rearranged RGB image');
      end;
  else
      set(labelAxis,'String','The PC''s viewed as a composite RGB image','FontWeight','normal');
  end; % end: SmlUD.RGBmode


% STEP 3: Redraw the LHS axis image
% Step 3a: Score-space only
  hImage = findobj(handleAxis,'type','image');                                  % It is very important that there is only a single *image* in the
  delete(hImage);                                                               % axes (there can be other types of children, but only one child image)

  if strcmp(get(handles.ScoresButton,'State'),'on')
      if SmlUD.RGBmode                                                          % It is an RGB image - use the raw image data
          hImage = ShowImage(SmlUD.Image.X(:,:,[L1 L2 L3]),handleAxis,[],[],[],'direct');
      else                                                                      % otherwise show the scores as a composite RGB
          hImage = ShowImage(SmlUD.Image.T(:,:,[L1 L2 L3]),handleAxis,[],[],[],'direct');
      end;
      if nargin < 3, set(hImage,'UIContextMenu',handles.ImageContextMenu); end; % Activate the context menu for this item
      return;
  end;


% Step 3b: Loadings mode
  if strcmp(get(handles.LoadingsButton,'State'),'on')
      Eigenvalues      = diag(SmlUD.Image.Eigenvalues);                         % Get the eigenvalues
      SumEigenvalues   = sum(Eigenvalues);                                      % Compute the sum of the eigenvalues
      PercentExplained = Eigenvalues/SumEigenvalues .* 100;                     % Compute the percentage explained by each PC (eigenvalue)

      if SmlUD.Image.nLayers >= SmlUD.MaxNumPCBars                              % We have more bars than the maximum, so group
          oldPercentExplained  = PercentExplained;                              % the remaining PC's into the last bar
          PercentExplained     = oldPercentExplained(1:SmlUD.MaxNumPCBars-1);
          PercentExplained(SmlUD.MaxNumPCBars) = sum(oldPercentExplained(SmlUD.MaxNumPCBars:end));
      end

      set(get(handleAxis,'Parent'),'CurrentAxes',handleAxis);
      BarHandles = bar(PercentExplained,'k');  grid('on');                      % Plot the percentage explained
      if nargin < 3, set([BarHandles handles.ImagePlotAxes],'UIContextMenu',handles.ImageContextMenu); end;
      for i = 1:min(SmlUD.Image.nLayers,SmlUD.MaxNumPCBars)
          hTxt = text(i,PercentExplained(i)+8,num2str(PercentExplained(i),4));
          set(hTxt,'FontSize',8,'HorizontalAlignment','center','Color',[1 0 0]);
          if nargin < 3, set(hTxt,'UIContextMenu',handles.ImageContextMenu);
          end

          if SmlUD.Image.nLayers > SmlUD.MaxNumPCBars
              XStringLabel = '1';
              for k = 2:(SmlUD.MaxNumPCBars-1)
                  XStringLabel = strvcat(XStringLabel,num2str(k));
              end;
              XStringLabel = strvcat(XStringLabel,'Rest');
              set(gca,'XTickLabel',XStringLabel);
          else
              set(gca,'xtick',1:1:SmlUD.Image.nLayers);                         % change the axis limits to fixed values
          end;
          set(gca,'ytick',0:10:100);
          axis([1-0.5, min(SmlUD.MaxNumPCBars,SmlUD.Image.nLayers)+0.5, 0, 100+15]);    % expand plot slightly to accomodate the labels
          xlabel('Principal Component Number');
          ylabel('Percentage of Sum of Squares');
          set(labelAxis,'Visible','on','String','Percentage variation explained by each PC');
      end;
  end;

% Step 3c: Model Building Mode
  if strcmp(get(handles.ModelButton,'State'),'on')
    % This part updates the image plot according to the following order and rules
    % 1. Determine if there is a mask from the score space: YES - use it to create the new image with score space mask
    %                                                       NO  - create an empty mask and display image
    % 2. If there were previously selected pixels in the score space, then these are
    %    superimposed on top of the image and mask

      if SmlUD.expMode
          if SmlUD.RGBmode
              NewImage = SmlUD.Image.X(:,:,[L1 L2 L3]);
          else
              NewImage = SmlUD.Image.T(:,:,[L1 L2 L3]);
          end;
          set(handles.ZoomScoreInfo,'Visible','off');
      else % not in SmlUD.expMode
          if SmlUD.RGBmode                                                      % It is an RGB image - use the raw image data
              [NewImage,num] = libmia('mask2im', SmlUD.Image.T(:,:,SmlUD.hAxis), SmlUD.Image.T(:,:,SmlUD.vAxis), ...
                                                 double(SmlUD.Image.Mask), SmlUD.ImageMaskColour, SmlUD.Image.X(:,:,[L1 L2 L3]));
          else                                                                  % else use the scores
              [NewImage,num] = libmia('mask2im', SmlUD.Image.T(:,:,SmlUD.hAxis), SmlUD.Image.T(:,:,SmlUD.vAxis), ...
                                                 double(SmlUD.Image.Mask), SmlUD.ImageMaskColour, SmlUD.Image.T(:,:,[L1 L2 L3]));
          end;
          if sum(sum(SmlUD.Image.Mask))~=0
              set(handles.ZoomScoreInfo,'Visible','on','String',['Number of pixels under mask = ',num2str(num)]);
          end;

      end; %end: if-else-end: SmlUD.expMode
      hImage = ShowImage(NewImage,handleAxis,[1 SmlUD.Image.nCols],[1 SmlUD.Image.nRows],[],'direct',SmlUD.XLim,SmlUD.YLim);
      if nargin < 3, set(hImage,'UIContextMenu',handles.ImageContextMenu); end; % Activate the context menu for this item

      if ~isempty(SmlUD.Points.R)                                               % There are mouse clicks that must be added
          set(handleAxis,'NextPlot','add');
          temp = plot(SmlUD.Points.C,SmlUD.Points.R,'x','MarkerSize',8,'Color',SmlUD.ImageMaskColour,'Parent',handleAxis);
          hImage = [hImage;temp];                                               % Merge handle lists together
      end;
      if nargin < 3, set(hImage,'UIContextMenu',handles.ImageContextMenu); end; % Activate the context menu for these items
  end; % end: if strcmp(get(handles.ModelButton,'State'),'on')
return;






% --------------------------------------------------------------------
function UpdateRHS(varargin)
% This function updates the RHS axis when the drop down menus on that side are used.  This function may also be called
% when other elements change the GUI status and we need the right hand side image to be updated accordingly.
%
% The horizontal drop down  has twice as many options as the vertical one.   The first half are identical, while the second
% half are just residuals: eg: [T1 ; T2 ; Residual after T1 ; Residual after T2].  If the user selects an item in the first
% hald, then the vertical drop-down is activated and checked, else the vertical one is de-activated and hidden.
%
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Assess the GUI's state: was the item selected from the horizontal or vertical dropdown and if from the horizontal one,
%         was the item in the first or second half of the list.
% Step 2a. We are in T-score analysis mode: show the T-score or residual image
% Step 2b. We are in Loading analysis mode: show the loadings plot or bar plot
% Step 2c. The user wants to see a different score plot combination.

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1: Assess the GUI's status
  if SmlUD.adjMode                                                              % Do not do any updating of the RHS when in adjustment mode
      return;
  end;
  hSelect = 1;                                                                  % Assume the horizontal axis drop down was used to get here
  if varargin{1} == handles.TScoreVrtclAxis
      hSelect = 0;                                                              % if not, then change to vertical axis
  end;
  hValue = get(handles.TScoreHorizAxis,'Value');
  vValue = get(handles.TScoreVrtclAxis,'Value');
  nPC    = SmlUD.Image.nPC;
  handleAxis = handles.ScorePlotAxes;
  labelAxis  = handles.RHSPlotLabel;
  if nargin>1
      if ishandle(varargin{2})
          handleAxis = varargin{2};
      end;
  end;




% STEP 2a: A residual or T-score image must be drawn
  if strcmp(get(handles.ScoresButton,'State'),'on')
      if hValue > min(SmlUD.Image.nPC,SmlUD.Image.nLayers)                      % Draw residual
          whichResid  = hValue - nPC;                                           % For calculation details see Geladi and Grahn, p 183
          residMatrix = eye(SmlUD.Image.nLayers) - SmlUD.Image.P(:,1:whichResid) * SmlUD.Image.P(:,1:whichResid)';
          [residImage,nColours] = libmia('residual',SmlUD.Image.X,residMatrix);
          nColours = round(0.7*nColours);                                       % Only a low number of distinct levels of colour
          if nColours == 0, nColours = 12; end;
          hImage = ShowImage(residImage,handleAxis,[],[],[],'direct');

          if whichResid == 1
              set(labelAxis,'String', 'Residual intensity image after 1 PC extracted','Visible','on');
          else
              set(labelAxis,'String', ['Residual intensity image after ',num2str(whichResid),' PC''s extracted']);
          end;
      else                                                                      % else draw a T-score intensity image
          nColours = 256;
          hImage = ShowImage(SmlUD.Image.T(:,:,hValue),handleAxis,[],[],[],'direct');
          set(labelAxis,'String', ['Intensity image of principal component ',num2str(hValue)]);
      end; %end: if hValue > min(SmlUD.Image.nPC,SmlUD.Image.nLayers)

      set(GUI_Handle,'ColorMap',jet(nColours));                                 % to distinguish residual features more clearly
      image([0 1],[1 nColours],(1:nColours)','Parent',handles.ColourBarAxes);
      set(handles.ColourBarAxes,'YDir','normal','YAxisLocation','right','XTick',[]);% Draw the colourbar
      if nargin < 3, set(hImage,'UIContextMenu',handles.ScoreContextMenu); end; % Activate the context menu for this item
      return;
  end; %end: if strcmp(get(handles.ScoresButton,'State'),'on')

% STEP 2b: Loadings analysis mode
  if strcmp(get(handles.LoadingsButton,'State'),'on')
      Loadings = SmlUD.Image.P;                                                 % Get the loadings values
      set(get(handleAxis,'Parent'),'CurrentAxes',handleAxis);
      if hValue == vValue                                                       % plot a bar chart for that P_value
          BarHandles = bar(Loadings(:,hValue),'k');
          set(handleAxis,'YLim',[-1 1]);                                        % stretch the Y-axis to the upper and lower limits
          if nargin < 3, set([BarHandles handleAxis],'UIContextMenu',handles.ScoreContextMenu); end;
          set(labelAxis, 'String', ['Loading vector for principal component ',num2str(hValue)]);
      else                                                                      % Plot a scatter plot if the P's are different
          plot(Loadings(:,hValue),Loadings(:,vValue),'k.','MarkerSize',20);
          set(handleAxis,'YLim',[-1 1],'XLim',[-1 1]);                          % Add text labels still: see Geladi and Grahn, p 117 for example
          set(labelAxis, 'String', ['Loading vector of PC ',num2str(hValue),' against PC ',num2str(vValue)]);
      end;
      grid(handleAxis,'on');
      return;
  end;

% STEP 2c: Model building mode
  if strcmp(get(handles.ModelButton,'State'),'on')
      if hValue == vValue                                                       % Both the horizontal and vertical axes have same value
          if hSelect                                                            % If horizontal axis was changed, check that
              if vValue+1 > nPC                                                 %     if we bump the vertical axis one position on that we won't exceed nPC
                  set(handles.TScoreVrtclAxis,'Value',1);         vValue = 1;
              else                                                              %     else safely bump it up by one position
                  set(handles.TScoreVrtclAxis,'Value',vValue+1);  vValue = vValue+1;
              end;
          else
              if hValue+1 > nPC
                  set(handles.TScoreHorizAxis,'Value',1);         hValue = 1;
              else
                  set(handles.TScoreHorizAxis,'Value',hValue+1);  hValue = hValue+1;
              end;
          end
      end
      SmlUD.hAxis = hValue;
      SmlUD.vAxis = vValue;

      MyCLA(handleAxis);                                                        % Clean up axes first
      SmlUD.Image.ScoreImage = uint8(libmia('ScorePlot', SmlUD.Image.T(:,:,SmlUD.hAxis), SmlUD.Image.T(:,:,SmlUD.vAxis)  ));
      
      % original code
      %Index255 = find(SmlUD.Image.ScoreImage==255);                            % the above "uint8" conversion does the following: any number > 255 must be set to 255, any number < 0 is set to zero.
      %SmlUD.Image.ScoreImage(Index255) = 254;                                  % Set it one level down to reserve the 255 for Exploring mode
      % modified (Nov 2006)
      SmlUD.Image.ScoreImage(SmlUD.Image.ScoreImage==255) = 254;
      
      hImage = ShowImage(SmlUD.Image.ScoreImage,handleAxis,[1 256],[1 256],[],'direct');
      if nargin < 3, set(hImage,'UIContextMenu',handles.ScoreContextMenu); end; % Activate the context menu for this item
      set(GUI_Handle,'Colormap',SmlUD.ScoreSpaceColourMap);                     % Put on some labels
      set(labelAxis,'String','Score space');

      if ~isempty(SmlUD.Image.MaskXdata) && (SmlUD.Image.MaskXindex == hValue) && ~isempty(SmlUD.Image.MaskYdata) && (SmlUD.Image.MaskYindex == vValue) && ~SmlUD.expMode
          set(get(handleAxis,'Parent'),'CurrentAxes',handleAxis);               % quick way to make these axes current: my must use get(...,'Parent'), because different figure windows call this function
          xdata = SmlUD.Image.MaskXdata;
          ydata = SmlUD.Image.MaskYdata;
          set(handleAxis,'NextPlot','add');                                     % "hold on"
          patchColour = MaskColour_Callback(GUI_Handle,get(handles.ScoreMaskColour,'Value'));
          hPatch = patch(xdata,ydata,patchColour);     % draw the patch at the specified x- and y-values
          if nargin < 3, set(hPatch,'UIContextMenu',handles.ScoreContextMenu); end;
          MyCLA(handles.ZoomScoreAxes);
          ShowImage(SmlUD.Image.Mask,handles.ZoomScoreAxes,[],[],[],'scaled');
     end;

      if ~isempty(SmlUD.Points.R)                                               % There are mouse clicks that must be added
          set(handleAxis,'NextPlot','add');
          hPlot = plot(double(SmlUD.Points.List(:,SmlUD.hAxis))+0.5 , double(SmlUD.Points.List(:,SmlUD.vAxis))+0.5 , '.' ,'MarkerSize',8, 'Color',SmlUD.ScoreMaskColour,'Parent',handleAxis);
          if nargin < 3, set(hPlot,'UIContextMenu',handles.ScoreContextMenu); end;
      end;
      setappdata(GUI_Handle,'Small_Data',SmlUD);
  end; %end: if strcmp(get(handles.ModelButton,'State'),'on')
return;


% --------------------------------------------------------------------
function varargout = MaskColour_Callback(varargin)
% The user wants to change the colour used for masks, patches and dots in the score space
% Update: SmlUD.ScoreMaskColour
% The user wants to change the colour of the mask and dots in the score space to one of the following
% values (their [R G B] values are also shown:
% White     [  1   1   1  ]
% Red       [  1   0   0  ]
% Green     [  0   1   0  ]
% Blue      [  0   0   1  ]
% Yellow    [  1   1   0  ]
% Black     [  0   0   0  ]  <-- not available for "handles.ScoreMaskColour"
%
% Step 0: Get the handle of the figure window and required data for this function, including the ui-widget that called this function
% Step 1: Get the desired colour vector and store its data in SmlUD
% Step 2: Redraw the score space with the new dots and mask

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'ST_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  if varargin{1} == handles.ImageMaskColour
      imSpace = 1;
      hColour = handles.ImageMaskColour;
  else
      imSpace = 0;
      hColour = handles.ScoreMaskColour;
  end;

% STEP 1
  if isempty(varargin{2})
      NewColour = get(hColour,'Value');                                         % Get the desired colour
  else
      NewColour = varargin{2};
  end;

  switch NewColour
    case 1
      MaskColour = [  1   1   1  ];                                             % White (default)
    case 2
      MaskColour = [  1   0   0  ];                                             % Red
    case 3
      MaskColour = [  0   1   0  ];                                             % Green
    case 4
      MaskColour = [  0   0   1  ];                                             % Blue
    case 5
      MaskColour = [  1   1   0  ];                                             % Yellow
    case 6
      MaskColour = [  0   1   1  ];                                             % Cyan
    case 7
      MaskColour = [  0   0   0  ];                                             % Black
    otherwise
      MaskColour = [  1   1   1  ];                                             % White (default)
  end;

  if isempty(varargin{2})
      if imSpace
          SmlUD.ImageMaskColour = MaskColour;
      else
          SmlUD.ScoreMaskColour = MaskColour;                                   % Store the new mask colour
      end;
      setappdata(GUI_Handle,'Small_Data', SmlUD);                               % Save the data storage structures
      if ~SmlUD.expMode
          UpdateLHS(GUI_Handle);                                                % Now call these functions to update the LHS and RHS (except if we are in explore mode)
          UpdateRHS(GUI_Handle);
      end
  else
      varargout{1} = MaskColour;
  end;
return;

% --------------------------------------------------------------------
function MouseMotionSetup(GUI_Handle)
% This function starts the mouse motion capability that is used in various parts of the GUI.
% We store the various parameters of interest in a structure called MseUD in the GUI's storage area
% various parameters that handle zooming.

% Step 0: Get the background information
% Step 1: Used when the Zoom is being turned off
% Step 2: Used when the Zoom is being turned on
%      2a - Make a copy of the handles to the children in the current axes and store the necessary properties
%      2b - Save the data in the handles and update the GUI
% Step 3: Plot data in the smaller axes to show  the full image

% STEP 0
  global SmlUD
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 2:  Collect all the necessary data
  MseUD = [];
  MseUD.fraction = 0.5;                                                         % The starting zoom fraction (for zooming)
  MyCLA(handles.ZoomImageAxes);                                                 % Draw a copy of the image in the small axes as a reference to the user
  MseUD.hImageLHS = findobj(handles.ImagePlotAxes,'type','image');
  MseUD.hImageRHS = -10;
  if strcmp(get(handles.ScoresButton,'State'),'on')                             % Score mode
      set(handles.ImagePlotAxes,'XLim',SmlUD.XLim, 'YLim',SmlUD.YLim);
      set(handles.ScorePlotAxes,'XLim',SmlUD.XLim, 'YLim',SmlUD.YLim);
      MseUD.mode = 'Scores';
      MseUD.hImageRHS = findobj(handles.ScorePlotAxes,'type','image');
      MyCLA(handles.ZoomScoreAxes);
      ShowImage(get(MseUD.hImageLHS,'CData'),handles.ZoomImageAxes,[1 SmlUD.XLim],[1 SmlUD.YLim],[],'direct',SmlUD.XLim,SmlUD.YLim);
      ShowImage(get(MseUD.hImageRHS,'CData'),handles.ZoomScoreAxes,[1 SmlUD.XLim],[1 SmlUD.YLim],[],'direct',SmlUD.XLim,SmlUD.YLim);
  end;
  if strcmp(get(handles.ModelButton,'State'),'on')                              % Model building mode
      set(handles.ImagePlotAxes,'XLim',SmlUD.XLim, 'YLim',SmlUD.YLim);
      set(handles.ScorePlotAxes,'XLim',[0.5 256.5],'YLim',[0.5 256.5]);
      if SmlUD.CountMessage <= 6
          set(handles.ZoomSpaceMessage,'Visible','on','String','Press "Space" to freeze zoomed view. Press "h" for help.');
      end;
      set(handles.ZoomXYLocation,'Visible','on','String','');
      MseUD.mode = 'ModelBuilding';
      ShowImage(get(MseUD.hImageLHS,'CData'),handles.ZoomImageAxes,[1 SmlUD.XLim],[1 SmlUD.YLim],[],'direct',SmlUD.XLim,SmlUD.YLim);
  end;

  % The reference axes, hAxisCopyLHS, is a copy of the original axes but with no content (we delete any content).  These
  % axes are used just to provide the "CurrentPoint" relative to the mouse.
  MseUD.hAxisCopyLHS = copyobj(handles.ImagePlotAxes,GUI_Handle);               % Make a copy of the current axis (and children)
  set(findobj(MseUD.hAxisCopyLHS,'type','children'),'HandleVisibility','on');   % Set the handles of these copied children to be on, but
  set(MseUD.hAxisCopyLHS,'Visible','off','PlotBoxAspectRatioMode','manual');    % make the axis (and its children) invisible though (not necessary for images though)
  axes(MseUD.hAxisCopyLHS);                                                     % Make the axes current and then delete only the children with visible handles
  delete(findobj(get(MseUD.hAxisCopyLHS,'Children'),'flat','HandleVisibility','on','serializable','on'));
%  MseUD.hImageLHS = findobj(handles.ImagePlotAxes,'type','image');              % Find the handle to the ***image*** in the LHS axis

  MseUD.hAxisCopyRHS = copyobj(handles.ScorePlotAxes,GUI_Handle);                % Do the same in the score axes
  set(findobj(MseUD.hAxisCopyRHS,'type','children'),'HandleVisibility','on');
  set(MseUD.hAxisCopyRHS,'Visible','off','PlotBoxAspectRatioMode','manual');
  axes(MseUD.hAxisCopyRHS);
  delete(findobj(get(MseUD.hAxisCopyRHS,'Children'),'flat','HandleVisibility','on','serializable','on'));


  MseUD.XOrig  = SmlUD.XLim;
  MseUD.YOrig  = SmlUD.YLim;
  MseUD.XRange = diff(SmlUD.XLim);
  MseUD.YRange = diff(SmlUD.YLim);
  MseUD.XDist  = MseUD.fraction * MseUD.XRange;
  MseUD.YDist  = MseUD.fraction * MseUD.YRange;
  MseUD.WBMF   = get(GUI_Handle,'WindowButtonMotionFcn');                       % Get the current WindowButtonMotionFcn
  MseUD.Freeze = 0;
  MseUD.hLine1 = [];                                                            % Set these empty: they will be used later to create the zoom outline in the small windows
  MseUD.hLine2 = [];
  MseUD.hGrid1 = [];
  MseUD.hGrid2 = [];
  MseUD.LastCP = [0 0];                                                         % Set the last point to be empty


% Step 2b: Save the data and update the GUI
  setappdata(GUI_Handle,'Mouse_Data',MseUD);                                    % Store the mouse information in the GUI's storage area
  setappdata(GUI_Handle,'Small_Data',SmlUD);
  set(GUI_Handle,'WindowButtonMotionFcn',@MouseMotion_Callback,'Resize','off'); % Resize must be off: if not, the CurrentPoint properties get messed up
  set(GUI_Handle,'WindowButtonDownFcn',  @MouseButton_Callback);
  set(GUI_Handle,'KeyPressFcn',@GUIKeyPress_Callback);
  set([ handles.ZoomMoreButton   handles.ZoomLessButton ],'Enable','on');
return;


% --------------------------------------------------------------------
function varargout = ZoomStartButton(varargin)
% This function starts the zooming ability in the GUI.
% Step 0: Get the background information
% Step 1: Zoom is being turned off: clean up the small axes
% Step 2: Zoom is being turned on: set up the GUI
% Step 3: Plot data in the smaller axes that shows the full image

% STEP 0
  global SmlUD
  varargout = [];
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  MseUD = getappdata(GUI_Handle,'Mouse_Data');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1: Used when the Zoom is being turned off
  if strcmp(get(handles.ZoomStartButton,'State'),'off')                         % We were in zoom mode and now are turning it off
      % NOTE: This part of the function is called before Explore Mode commences (Zoom is turned off during exploring)
      % Also: use the "strcmp" form (not "if SmlUD.ZoomMode"), because the State is turned off to trick the Zoom mode to end itself
      SmlUD.ZoomMode = 0;
      MseUD.Freeze = 0;
      setappdata(GUI_Handle,'Mouse_Data',MseUD);
      set(handles.ZoomMoreButton,'Enable','off');
      set(handles.ZoomLessButton,'Enable','off');
      MyCLA(handles.ZoomImageAxes);
      MyCLA(handles.ZoomScoreAxes);
      set(handles.ZoomSpaceMessage,'Visible','off');
      set(handles.ZoomXYLocation,'Visible','off');
      UpdateLHS(GUI_Handle);
      UpdateRHS(GUI_Handle);
      return;
  end;

% STEP 2:  Used when the Zoom is being turned on
  SmlUD.XLim    = [0 SmlUD.Image.nCols] + 0.5;
  SmlUD.YLim    = [0 SmlUD.Image.nRows] + 0.5;
  SmlUD.ZoomMode = 1;
  MouseMotionSetup(GUI_Handle);                                                 % First get these MseUD parameters correctly set up
  MseUD = getappdata(GUI_Handle,'Mouse_Data');                                  % Fetch these storage structures again
  SmlUD = getappdata(GUI_Handle,'Small_Data');
  MseUD.Freeze = 0;

% STEP 3: Show the full image in the little axes
  setappdata(GUI_Handle,'Small_Data',SmlUD);
  setappdata(GUI_Handle,'Mouse_Data',MseUD);
return;

% --------------------------------------------------------------------
function varargout = ZoomChangeButton(varargin)
% This function is called if the user wants to zoom in or out.  It determines the new zoom factor and adjusts the images accordingly.

  global SmlUD
  varargout = [];
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  MseUD = getappdata(GUI_Handle,'Mouse_Data');
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

  if handles.ZoomMoreButton == varargin{1}
      MseUD.fraction = max(0.01,MseUD.fraction*0.81);                           % The user wants to zoom in closer
  elseif handles.ZoomLessButton == varargin{1}
      MseUD.fraction = min(0.95,MseUD.fraction*1.21);                              % The user wants to zoom out
  end;
  if MseUD.fraction >= 1
      set(handles.ZoomLessButton,'Enable','off');
  elseif MseUD.fraction <= 0.01
      set(handles.ZoomMoreButton,'Enable','off');
  else
      set(handles.ZoomLessButton,'Enable','on');
      set(handles.ZoomMoreButton,'Enable','on');
  end;

  MseUD.XDist = MseUD.fraction * MseUD.XRange;
  MseUD.YDist = MseUD.fraction * MseUD.YRange;

  x1 = (MseUD.XRange - MseUD.XDist)/2;
  x2 = (MseUD.XRange + MseUD.XDist)/2;
  y1 = (MseUD.YRange - MseUD.YDist)/2;
  y2 = (MseUD.YRange + MseUD.YDist)/2;

  if ishandle(MseUD.hLine1)                                                     % Give a visual indication by how much the zooming is happening
      set(MseUD.hLine1,'XData',[x1 x1 x2 x2 x1],'YData',[y1 y2 y2 y1 y1],'Visible','on');
  end;
  if ishandle(MseUD.hLine2)
      set(MseUD.hLine2,'XData',[x1 x1 x2 x2 x1],'YData',[y1 y2 y2 y1 y1],'Visible','on');
  end;
  set(handles.ImagePlotAxes,'XLim',[x1 x2],'YLim',[y1 y2]);
  if strcmp(MseUD.mode,'Scores')
      set(handles.ScorePlotAxes,'XLim',[x1 x2],'YLim',[y1 y2]);
  end
  SmlUD.XLim = [x1 x2];
  SmlUD.YLim = [y1 y2];
  setappdata(GUI_Handle,'Small_Data',SmlUD);

  if MseUD.Freeze == 1
      MseUD.Freeze = 0;
      if SmlUD.CountMessage < 6
          set(handles.ZoomSpaceMessage,'String','Use "Space bar" to freeze zoomed view. Press "h" for help.','Visible','on');
      end;
      SmlUD.CountMessage = SmlUD.CountMessage + 1;
  end;
  setappdata(GUI_Handle,'Mouse_Data',MseUD);
  setappdata(GUI_Handle,'Small_Data',SmlUD);
return;


% --------------------------------------------------------------------
function MouseMotion_Callback(varargin)
% This functions handles the mouse interaction with the user.  It handles the point clicking in the image space and
% mask creation in the score space (including mask modification).  Zooming is also included when in zoom mode
%
% Step 0: Collect the nessary data, including the mouse's current location
% Step 1: Determine if the mouse is in the LHS axes or the RHS axes
% Step 2: Check if the image if the little image needs updating
% Step 3: Handle the RHS and LHS cases

% STEP 0: Collect data
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  MseUD   = getappdata(GUI_Handle,'Mouse_Data');
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
%  set(GUI_Handle,'Interruptible','on','BusyAction','cancel');                   % Interrupt any mouse action and cancel it


% STEP 1: LHS or RHS
  cpLHS = get(MseUD.hAxisCopyLHS,'CurrentPoint');                               % Get the current point (CP) in the LHS axes
  cpRHS = get(MseUD.hAxisCopyRHS,'CurrentPoint');
  %LHSaxis = 0; RHSaxis = 0; cp = [0 0];                                         % Initialize (incase we are not in LHS or RHS)


% % KGD: added 18 November 2005 to fix how R14 addresses mouse pointer location outside the image axis
% % One could previously rely on the fact that cpRHS(1,3) or  cpLHS(1,3) were 1 when the mouse was inside the axis.
% % R14 does not follow this convention anymore.  The new code should work in both R12, R13 and R14 versions.
% %   if cpRHS(1,3) == 1
% %     cp = cpRHS;
% %     RHSaxis = 1;                                                                % Indicate that we got the point from the RHS axis
% %   end;
% %   if cpLHS(1,3) == 1
% %     cp = cpLHS;
% %     LHSaxis = 1;                                                                % Indicate that we got the point from the LHS axis
% %   end;

% % KGD: removed this on 1 March 2006 (a better way to get pixel information)
% % Noted due to Salvador Garcia: problem occurs when image space has less than 256 pixels in the vertical direction
%   if any(cpLHS(1,1:2)<0) || any([cpLHS(1,3) cpRHS(1,3)]~=[1 1])                   % to the left of the LHSaxis, above or below the axes
%       LHSaxis = 0; RHSaxis = 0; cp = [0 0];
%       disp('to the left, or above, or below all')
%   else
%       if all(cpLHS(1,1:2) < [SmlUD.Image.nCols SmlUD.Image.nRows])                % in the LHS axis
%           LHSaxis = 1;  RHSaxis = 0; cp = cpLHS(1,1:2);
%           disp('in the LHS axis')
%       else
%           LHSaxis = 0;
%           if any(cpRHS(1,1:2)<0)                                                  % to the left of the RHS axis, but not in the LHS axis
%               RHSaxis = 0; cp = [0 0];
%               disp('to the left of RHS, but not inside LHS')
%           else
%               if all(cpRHS(1,1:2) < [256 256])                                    % in the RHS axis
%                   RHSaxis = 1;  cp = cpRHS(1,1:2);
%                   disp('in the RHS axis')
%               else
%                   RHSaxis = 0;                                                    % to the right of the RHS axis
%                   cp = [0 0];
%                   disp('to the right of the RHS axis');
%               end
%           end
%       end
%   end

  if cpLHS(1,1)>=0 && (cpLHS(1,1) < SmlUD.Image.nCols) && cpLHS(1,2)>=0 && (cpLHS(1,2) < SmlUD.Image.nRows)
      LHSaxis = 1;
      RHSaxis = 0;
      cp = cpLHS(1,1:2);
      %disp('in LHS axis');
  elseif cpRHS(1,1)>=0 && (cpRHS(1,1) < 256) && cpRHS(1,2)>=0 && (cpRHS(1,2) < 256)
      LHSaxis = 0;
      RHSaxis = 1;
      cp = cpRHS(1,1:2);
      %disp('in RHS axis');
  else
      LHSaxis = 0;
      RHSaxis = 0;
      cp = [0 0];
      %disp('neither axis');
  end
  x = cp(1,1);
  y = cp(1,2);


% STEP 2: Check if the little images needs updating
  if ~ishandle(MseUD.hImageLHS) && SmlUD.ZoomMode                               % Only redraw little image if it has changed
      MseUD.hImageLHS = findobj(handles.ImagePlotAxes,'type','image');
      MyCLA(handles.ZoomImageAxes);
      ShowImage(get(MseUD.hImageLHS,'CData'),handles.ZoomImageAxes,[1 MseUD.XOrig],[1 MseUD.YOrig],[],'direct');
      setappdata(GUI_Handle,'Mouse_Data',MseUD);
  end
  if ~ishandle(MseUD.hImageRHS) && SmlUD.ZoomMode && strcmp(MseUD.mode,'Scores')% Only redraw little image if it has changed
      MseUD.hImageRHS = findobj(handles.ScorePlotAxes,'type','image');
      MyCLA(handles.ZoomScoreAxes);
      ShowImage(get(MseUD.hImageRHS,'CData'),handles.ZoomScoreAxes,[1 MseUD.XOrig],[1 MseUD.YOrig],[],'direct');
      setappdata(GUI_Handle,'Mouse_Data',MseUD);
  end

% Pre step 3:
  if SmlUD.adjModeButton                                                        % This allows the mouse to go outside the score space
      x = cpRHS(1,1);                                                           % when in Adjust mode
      y = cpRHS(1,2);
      RHSaxis = 1;
  end;

% STEP 3: If RHS:  show LHS at full extent, change the mouse to circle, update tiny images if necessary and save data
%         If LHS:  change mouse to cross ans show location.  If in zoom mode: zoom image, draw zoom limit lines, update tiny images, save data
%         Neither: change mouse to arrow, show LHS at full extent, hide X,Y text
  if RHSaxis && strcmp(MseUD.mode,'ModelBuilding') && ~xor(SmlUD.expMode,SmlUD.adjMode)  % we are in RHS, but not in SmlUD.expMode nor in SmlUD.adjMode
     %disp(['NORMAL: expMode = ',num2str(SmlUD.expMode),' and adjMode = ',num2str(SmlUD.adjMode),' and overall we have a ',num2str(RHSaxis & xor(~SmlUD.expMode,~SmlUD.adjMode))]);

      set(GUI_Handle,'Pointer','custom','PointerShapeCData',SmlUD.CirclePointerShape,'PointerShapeHotSpot',[8 8]);
      set(handles.ZoomXYLocation,'Visible','off');
      if ishandle(handles.LineH1)                                               % User is busy constructing a mask
          Xmask = SmlUD.Image.MaskXdata;
          Ymask = SmlUD.Image.MaskYdata;
          LengthX = length(Xmask);
          if LengthX >= 3
              Xmask = [Xmask(1:LengthX-1) cp(1,1) Xmask(LengthX)];
              Ymask = [Ymask(1:LengthX-1) cp(1,2) Ymask(LengthX)];
          else
              Xmask = [Xmask cp(1,1)];
              Ymask = [Ymask cp(1,2)];
          end;
          set([handles.LineH1 handles.LineH2],'XData',Xmask,'YData',Ymask,'Clipping','off');
      end
      return;
  end;

  if RHSaxis && strcmp(MseUD.mode,'ModelBuilding') && SmlUD.expMode               % in the RHS axes and ExploreMode
      set(GUI_Handle,'Pointer','custom','PointerShapeCData',SmlUD.expPointerShape,'PointerShapeHotSpot',[8 8]);
      N = SmlUD.ScoreMaskSize/2;
      cMin = max([ 1,   floor(x-N+1)            ]);                             % bottom left
      cMax = min([ 256,  ceil(x+N  ), cMin+N*2-1]);                             % bottom right
      rMin = max([ 1,   floor(y-N+1)            ]);                             % top left
      rMax = min([ 256,  ceil(y+N  ), rMin+N*2-1]);                             % top right


      if ishandle(handles.expLine)
          set(handles.expLine,'Parent',handles.ScorePlotAxes,'XData',[cMin+0.5 cMin+0.5 cMax cMax cMin+0.5], ...
                                                             'YData',[rMin+0.5 rMax rMax rMin+0.5 rMin+0.5], ...
                                                             'Color',SmlUD.ScoreMaskColour,'Visible','on');
      else
          handles.expLine = line('Parent',handles.ScorePlotAxes,'XData',[cMin+0.5 cMin+0.5 cMax cMax cMin+0.5], ...
                                 'YData',[rMin+0.5 rMax rMax rMin+0.5 rMin+0.5],'Visible','on','Clipping','on', ...
                                 'LineStyle','-','Color',SmlUD.ScoreMaskColour);
          setappdata(GUI_Handle,'Handles_Data',handles);
      end;

      set(handles.hZoomScoreImage,'CData',SmlUD.Image.ScoreImage(rMin:rMax,cMin:cMax));   % Score space subset currently under the mouse
      set(handles.ZoomScoreAxes,'XTickLabel',[cMin cMax],'YTickLabel',[rMin rMax],'Visible','on');
      BI = uint8(zeros(256));
      BI(rMin:rMax,cMin:cMax) = uint8(ones(rMax-rMin+1,cMax-cMin+1));

      % The code from this point on is reused just below here in SmlUD.adjMode: any fundamental changes must be duplicated
      L1 = SmlUD.RLayer;    L2 = SmlUD.GLayer;      L3 = SmlUD.BLayer;          % Get the image order in the image plot
      if SmlUD.RGBmode                                                          % use the raw image data
          [NewImage,num] = libmia('mask2im',SmlUD.Image.T(:,:,SmlUD.hAxis),SmlUD.Image.T(:,:,SmlUD.vAxis), ...
                                            double(BI), SmlUD.ImageMaskColour, SmlUD.Image.X(:,:,[L1 L2 L3]));
      else                                                                      % else use the scores (only difference is the last argument)
          [NewImage,num] = libmia('mask2im',SmlUD.Image.T(:,:,SmlUD.hAxis),SmlUD.Image.T(:,:,SmlUD.vAxis), ...
                                            double(BI), SmlUD.ImageMaskColour, SmlUD.Image.T(:,:,[L1 L2 L3]));
      end;

      h = findobj(handles.ImagePlotAxes,'type','image');
      set(h,'CData',NewImage(:,:,1:3),'EraseMode','background');
      h = findobj(handles.ScorePlotAxes,'type','image');
      set(h,'CData',SmlUD.Image.ScoreImage,'EraseMode','background');
      drawnow;

      set(handles.ZoomScoreInfo,'Visible','on','String',['Number of pixels under mask = ',num2str(num)]);
      return;
  end;

  if RHSaxis && strcmp(MseUD.mode,'ModelBuilding') && SmlUD.adjMode             % in the RHS axes and AdjustMode
     %disp(['ADJUST: ',num2str(toc)]);%expMode = ',num2str(SmlUD.expMode),' and adjMode = ',num2str(SmlUD.adjMode),' and overall we have a ',num2str(RHSaxis & xor(~SmlUD.expMode,~SmlUD.adjMode))]);
      pntRadius = 15;
      if ishandle(handles.adjLine) && ishandle(handles.adjPoint)
          xData = get(handles.adjLine,'XData');
          yData = get(handles.adjLine,'YData');
          distance = sqrt((xData-x).^2 + (yData-y).^2);
          [value,index] = min(distance);                                        % Find the point closest to the mouse

          if SmlUD.adjModeButton                                                % The user has the mouse button down and is dragging (click and drag)
              %set(GUI_Handle,'Pointer','fleur');
              set(GUI_Handle,'Pointer','custom','PointerShapeCData',SmlUD.HandClosedPointerShape,'PointerShapeHotSpot',[10 9]);
              nPoints = length(xData);
              if (index ~= 1) && (index ~= nPoints)
                  X = [ xData(1:index-1)  x  xData(index+1:end) ];
                  Y = [ yData(1:index-1)  y  yData(index+1:end) ];
              else
                  X = [x xData(2:end-1) x];
                  Y = [y yData(2:end-1) y];
              end;
              set(handles.adjLine, 'XData',X,'YData',Y,'Visible','on','Color',SmlUD.ScoreMaskColour);
              set(handles.adjPoint,'XData',x,'YData',y,'Visible','on','MarkerFaceColor',SmlUD.ScoreMaskColour,'Color',SmlUD.ScoreMaskColour);
              SmlUD.Image.Mask = ComputeScoreMask(X,Y,256,256);

              % The code from this point on is reused just above here when not in SmlUD.adjMode: any fundamental changes must be duplicated
              L1 = SmlUD.RLayer;      L2 = SmlUD.GLayer;      L3 = SmlUD.BLayer;
              if SmlUD.RGBmode                                                  % use the raw image data
                  [NewImage,num] = libmia('mask2im',SmlUD.Image.T(:,:,SmlUD.hAxis),SmlUD.Image.T(:,:,SmlUD.vAxis), ...
                                          SmlUD.Image.Mask, SmlUD.ImageMaskColour, SmlUD.Image.X(:,:,[L1 L2 L3]));
              else                                                              % else use the scores (only difference is the last argument)
                  [NewImage,num] = libmia('mask2im',SmlUD.Image.T(:,:,SmlUD.hAxis),SmlUD.Image.T(:,:,SmlUD.vAxis), ...
                                          SmlUD.Image.Mask, SmlUD.ImageMaskColour, SmlUD.Image.T(:,:,[L1 L2 L3]));
              end;
              MseUD.hImageLHS = findobj(handles.ImagePlotAxes,'type','image');
              set(MseUD.hImageLHS,'CData',NewImage(:,:,1:3),'EraseMode','background');
              h = findobj(handles.ZoomImageAxes,'type','image');
              set(h,'CData',NewImage(:,:,1:3),'EraseMode','background');
              setappdata(GUI_Handle,'Mouse_Data',MseUD);

              set(handles.ZoomScoreInfo,'Visible','on','String',['Number of pixels under mask = ',num2str(num)]);
              drawnow;                                                          % This line is important: otherwise the animation doesn't work
          else                                                                  % The user is just moving the mouse without the mouse buttons being pressed
              if value < pntRadius
                  set(GUI_Handle,'Pointer','custom','PointerShapeCData',SmlUD.HandOpenPointerShape,'PointerShapeHotSpot',[10 9]);
                  set(handles.adjPoint,'XData',xData(index),'YData',yData(index),'Visible','on','MarkerFaceColor',SmlUD.ScoreMaskColour,'Color',SmlUD.ScoreMaskColour);
              else
                  set(GUI_Handle,'Pointer','crosshair');
                  set(handles.adjPoint,'Visible','off');
              end;
          end;
          setappdata(GUI_Handle,'Small_Data',SmlUD);
      else
          return;
      end; % end: if ishandle(handles.adjLine) && ishandle(handles.adjPoint)

      return;
  end; %end: if RHSaxis && SmlUD.adjMode

  % in the LHS axes
  if LHSaxis && strcmp(MseUD.mode,'ModelBuilding') && ~(SmlUD.expMode)
      set(GUI_Handle,'Pointer','cross');
      axisCP = get(handles.ImagePlotAxes,'CurrentPoint');
      set(handles.ZoomXYLocation,'String',['X = ',num2str(round(axisCP(1,1))),';  Y = ',num2str(round(axisCP(1,2)))],'Visible','on');
      if MseUD.Freeze                                                           % Handle this case before the remaining cases below: don't do anything if in freeze mode
         return;
      end;

      if  SmlUD.ZoomMode  %strcmp(get(handles.ZoomStartButton,'State'),'on')
          x1 = min( max(0,x-0.5*MseUD.XDist) , MseUD.XRange-MseUD.XDist ) + 0.5;
          y1 = min( max(0,y-0.5*MseUD.YDist) , MseUD.YRange-MseUD.YDist ) + 0.5;
          x2 = x1 + MseUD.XDist;
          y2 = y1 + MseUD.YDist;
          set(handles.ImagePlotAxes,'XLim',[x1 x2],'YLim',[y1 y2]);
          SmlUD.XLim = [x1 x2];
          SmlUD.YLim = [y1 y2];
          if ishandle(MseUD.hLine1),                                            % Change the line if it exists as a handle
              set(MseUD.hLine1,'XData',[x1 x1 x2 x2 x1],'YData',[y1 y2 y2 y1 y1],'Visible','on');
          else
              set(GUI_Handle,'CurrentAxes',handles.ZoomImageAxes);                                      % else draw it for the first time
              MseUD.hLine1 = line([x1 x1 x2 x2 x1],[y1 y2 y2 y1 y1]);
              setappdata(GUI_Handle,'Mouse_Data',MseUD);
          end
          setappdata(GUI_Handle,'Small_Data',SmlUD);
      end
      return;
  end

  % in the LHS axes and ExploreMode
  if LHSaxis && strcmp(MseUD.mode,'ModelBuilding') && SmlUD.expMode
      set(GUI_Handle,'Pointer','custom','PointerShapeCData',SmlUD.expPointerShape,'PointerShapeHotSpot',[8 8]);
      set(handles.ZoomXYLocation,'Visible','off');
      N = SmlUD.ImageMaskSize/2;
      cMin = max([ 1,            floor(x-N+1)            ]);                    % bottom left
      cMax = min([ MseUD.XRange,  ceil(x+N  ), cMin+N*2-1]);                    % bottom right
      rMin = max([ 1,            floor(y-N+1)            ]);                    % top left
      rMax = min([ MseUD.YRange,  ceil(y+N  ), rMin+N*2-1]);                    % top right

      if ishandle(handles.expLine)
          set(handles.expLine,'Parent',handles.ImagePlotAxes,'XData',[cMin cMin cMax cMax cMin], ...
                                                             'YData',[rMin rMax rMax rMin rMin], ...
                                                             'Color',SmlUD.ImageMaskColour,'Visible','on');
      else
          handles.expLine = line('Parent',handles.ImagePlotAxes,'XData',[cMin cMin cMax cMax cMin], ...
                                 'YData',[rMin rMax rMax rMin rMin],'Visible','on','Clipping','on', ...
                                 'LineStyle','-','Color',SmlUD.ImageMaskColour);
          setappdata(GUI_Handle,'Handles_Data',handles);
      end;
      set([handles.ZoomScoreAxes handles.ZoomScoreInfo],'Visible','off');
      set(handles.hZoomScoreImage,'CData',[]);

      t_horiz = SmlUD.Image.T(rMin:rMax,cMin:cMax,SmlUD.hAxis);
      t_vrtcl = SmlUD.Image.T(rMin:rMax,cMin:cMax,SmlUD.vAxis);
      ScoreImage = libmia('ScorePlot',t_horiz,t_vrtcl);
      % This works well, especially for images with many layers - we don't compute results that won't be used.   But is only a valid approach if the
      % T-scores have been computed from projection onto and scaling with the correct loadings and scaling matrices.  In this case we are exploring
      % the image used to build the model, so T-scores are correctly projected and scaled already.
      nonZeros = find(ScoreImage ~= 0);
      ScoreImage = SmlUD.Image.ScoreImage;
      ScoreImage(nonZeros) = 255;                                               % The non-zero values are mapped to 255, which is shown as cyan on the figure's colormap.
      h = findobj(handles.ScorePlotAxes,'type','image');
      set(h,'CData',uint8(ScoreImage),'EraseMode','background');
      drawnow;
      return;
  end

  if (LHSaxis||RHSaxis) && strcmp(MseUD.mode,'Scores') && SmlUD.ZoomMode %strcmp(get(handles.ZoomStartButton,'State'),'on')
      set(GUI_Handle,'Pointer','cross');
      axisCP = get(handles.ImagePlotAxes,'CurrentPoint');
      set(handles.ZoomXYLocation,'String',['X = ',num2str(round(axisCP(1,1))),';  Y = ',num2str(round(axisCP(1,2)))],'Visible','on');
      if MseUD.Freeze                                                           % Handle this case before the remaining cases below: don't do anything if in freeze mode
         return;
      end;

      x1 = min( max(0,x-0.5*MseUD.XDist) , MseUD.XRange-MseUD.XDist ) + 0.5;
      y1 = min( max(0,y-0.5*MseUD.YDist) , MseUD.YRange-MseUD.YDist ) + 0.5;
      x2 = x1 + MseUD.XDist;
      y2 = y1 + MseUD.YDist;
      set(handles.ImagePlotAxes,'XLim',[x1 x2],'YLim',[y1 y2]);
      set(handles.ScorePlotAxes,'XLim',[x1 x2],'YLim',[y1 y2]);
      SmlUD.XLim = [x1 x2];
      SmlUD.YLim = [y1 y2];
      if ishandle(MseUD.hLine1),                                                % Change the line if it exists as a handle
          set(MseUD.hLine1,'XData',[x1 x1 x2 x2 x1],'YData',[y1 y2 y2 y1 y1],'Visible','on');
      else
          axes(handles.ZoomImageAxes);                                          % else draw it for the first time
          MseUD.hLine1 = line([x1 x1 x2 x2 x1],[y1 y2 y2 y1 y1]);
          setappdata(GUI_Handle,'Mouse_Data',MseUD);
      end
      if ishandle(MseUD.hLine2),                                                % Change the line if it exists as a handle
          set(MseUD.hLine2,'XData',[x1 x1 x2 x2 x1],'YData',[y1 y2 y2 y1 y1],'Visible','on');
      else
          axes(handles.ZoomScoreAxes);                                          % else draw it for the first time
          MseUD.hLine2 = line([x1 x1 x2 x2 x1],[y1 y2 y2 y1 y1]);
          setappdata(GUI_Handle,'Mouse_Data',MseUD);
      end
      setappdata(GUI_Handle,'Small_Data',SmlUD);
      return;
  end;



  % neither the LHS or RHS axes
  if (~RHSaxis && ~LHSaxis)
      set(GUI_Handle,'Pointer','arrow');
      set([handles.ZoomXYLocation handles.ZoomScoreInfo],'Visible','off');
      if MseUD.Freeze                                                           % This avoids resetting the limits when in freeze mode
          return;
      elseif strcmp(MseUD.mode,'Scores')
          set(handles.ImagePlotAxes,'XLim',MseUD.XOrig,'ylim',MseUD.YOrig);     % If we are not inside the image space then reset limits
          set(handles.ScorePlotAxes,'XLim',MseUD.XOrig,'ylim',MseUD.YOrig);
          SmlUD.XLim = MseUD.XOrig;
          SmlUD.YLim = MseUD.YOrig;
          setappdata(GUI_Handle,'Small_Data',SmlUD);
      else
          set(handles.ImagePlotAxes,'XLim',MseUD.XOrig,'ylim',MseUD.YOrig);     % If we are not inside the image space then reset limits
          SmlUD.XLim = MseUD.XOrig;
          SmlUD.YLim = MseUD.YOrig;
          setappdata(GUI_Handle,'Small_Data',SmlUD);
      end

      if ishandle(MseUD.hLine1)
          set(MseUD.hLine1,'Visible','off');
      end
      if ishandle(MseUD.hLine2)
          set(MseUD.hLine2,'Visible','off');
      end

      if SmlUD.expMode
          set(handles.ZoomScoreAxes,'Visible','off');
          set(handles.hZoomScoreImage,'CData',[]);
      end
      return;
  end
%  set(GUI_Handle,'Interruptible','off','BusyAction','queue');
return;




% --------------------------------------------------------------------
function MouseButton_Callback(varargin)
% This function is called when the left mouse button is pressed/clicked

% Step 0: Get the required data for this function; the point which was click and the type of click (left, right, double, etc)
% Step 1: Determine where the user clicked the mouse: LHS axes or RHS
% Step 2: If clicked in score space:
% Step 3: If clicked in image space: get the clicked point, do a local zoom, plot the point with "X", map back to score space

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  ClickType = get(GUI_Handle,'SelectionType');
  if ~strcmp(ClickType,'normal')                                                % If it wasn't a left click, then return
      if ~strcmp(ClickType,'alt')                                               % This avoids going into the function when user right clicks on mask images
          return;
      end;
    return;
  end;

% STEP 1: LHS or RHS
  cpLHS = get(handles.ImagePlotAxes,'CurrentPoint');
  cpRHS = get(handles.ScorePlotAxes,'CurrentPoint');
  %LHSaxis = 0; RHSaxis = 0; 
  cp = [0 0];                                         % Initialize (incase we are not in LHS or RHS)

  %% KGD: commented out this code on 18 November 2005.  See further comments below.

%   if cpRHS(1,3) == 1
%     cp = cpRHS;
%     RHSaxis = 1;                                                                % Indicate that we got the point from the RHS axis
%   end;
%   if cpLHS(1,3) == 1
%     cp = cpLHS;
%     LHSaxis = 1;                                                                % Indicate that we got the point from the LHS axis
%   end;

% % KGD: added 18 November 2005 to fix how R14 addresses mouse pointer location outside the image axis
% % One could previously rely on the fact that cpRHS(1,3) or  cpLHS(1,3) were 1 when the mouse was inside the axis.
% % R14 does not follow this convention anymore.  The new code should work in both R12, R13 and R14 versions.
%   if any(cpLHS(1,1:2)<0) || all([cpLHS(1,3) cpRHS(1,3)]~=[1 1])                   % to the left of the LHSaxis, above or below the axes
%       LHSaxis = 0; RHSaxis = 0; cp = [0 0];
%       %disp('to the left, or above, or below all')
%   else
%       if all(cpLHS(1,1:2) < [SmlUD.Image.nCols SmlUD.Image.nRows])                % in the LHS axis
%           LHSaxis = 1;  RHSaxis = 0; cp = cpLHS(1,1:2);
%           %disp('in the LHS axis')
%       else
%           LHSaxis = 0;
%           if any(cpRHS(1,1:2)<0)                                                  % to the left of the RHS axis, but not in the LHS axis
%               RHSaxis = 0; cp = [0 0];
%               %disp('to the left or RHS, but not inside LHS')
%           else
%               if all(cpRHS(1,1:2) < [256 256])                                    % in the RHS axis
%                   RHSaxis = 1;  cp = cpRHS(1,1:2);
%                   %disp('in the RHS axis')
%               else
%                   RHSaxis = 0;                                                    % to the right of the RHS axis
%                   %disp('to the right of the RHS axis');
%               end
%           end
%       end
%   end
%   x = cp(1,1);
%   y = cp(1,2);

  
  
% % KGD: removed the above code on 1 March 2006 (a better way to get pixel information)
% % Noted due to Salvador Garcia: problem occurs when image space has less than 256 pixels in the vertical direction
  if cpLHS(1,1)>=0 && (cpLHS(1,1) < SmlUD.Image.nCols) && cpLHS(1,2)>=0 && (cpLHS(1,2) < SmlUD.Image.nRows)
      LHSaxis = 1;
      RHSaxis = 0;
      cp = cpLHS(1,1:2);
      %disp('in LHS axis');
  elseif cpRHS(1,1)>=0 && (cpRHS(1,1) < 256) && cpRHS(1,2)>=0 && (cpRHS(1,2) < 256)
      LHSaxis = 0;
      RHSaxis = 1;
      cp = cpRHS(1,1:2);
      %disp('in RHS axis');
  else
      LHSaxis = 0;
      RHSaxis = 0;
      cp = [0 0];
      %disp('neither axis');
  end
  x = cp(1,1);
  y = cp(1,2);  
  
% STEP 2: Click in the score space
  if RHSaxis && ~(SmlUD.adjMode) && strcmp(get(handles.ModelButton,'State'),'on')
      if ~isempty(SmlUD.Image.MaskXdata)                                        % First delete the current score-space patch
          hPatch = findobj(handles.ScorePlotAxes,'type','patch');
          if ishandle(hPatch)
              delete(hPatch)
          end;
      end;

      SmlUD.ScoreMaskState = uisuspend(GUI_Handle);
      set(GUI_Handle,...
          'WindowButtonDownFcn',  @ScoreMaskNextButtonDown, ...
          'WindowButtonMotionFcn',@MouseMotion_Callback, ...
          'KeyPressFcn',          @GUIKeyPress_Callback, ...
          'Pointer',              'custom', ...
          'PointerShapeCData',    SmlUD.CirclePointerShape, ...
          'PointerShapeHotSpot',  [8 8]);
      SmlUD.Image.MaskXdata  = [x x];
      SmlUD.Image.MaskYdata  = [y y];
      SmlUD.Image.MaskXindex = get(handles.TScoreHorizAxis,'Value');
      SmlUD.Image.MaskYindex = get(handles.TScoreVrtclAxis,'Value');
      handles.LineH1        = line('Parent',handles.ScorePlotAxes,'XData',[],'YData',[],'Visible','off', ...
                                   'Clipping','off','Color','k','LineStyle','-');
      handles.LineH2        = line('Parent',handles.ScorePlotAxes,'XData',[],'YData',[],'Visible','off', ...
                                   'Clipping','off','Color','w','LineStyle',':');
      setappdata(GUI_Handle,'Small_Data',SmlUD);                                % Save the points and state in data storage structure
      setappdata(GUI_Handle,'Handles_Data',handles);
      set(handles.LineH1,'XData',SmlUD.Image.MaskXdata,'YData',SmlUD.Image.MaskYdata,'Visible','on');
      set(handles.LineH2,'XData',SmlUD.Image.MaskXdata,'YData',SmlUD.Image.MaskYdata,'Visible','on');
      set(handles.TScoreHorizAxis,'Enable','off');
      set(handles.TScoreVrtclAxis,'Enable','off');
      set(handles.ScoreMaskColour,'Enable','off');
      set(handles.ImageMaskColour,'Enable','off');                              % All of these above are able to delete the children in score space if executed

      ErrorCatch = 0;
      try
          waitfor(handles.LineH1,'UserData','Completed');
      catch
          ErrorCatch = 1;
      end;

      if ErrorCatch==1
          ErrorStatus = 'trapped';
      elseif (~ishandle(handles.LineH1) || ~strcmp(get(handles.LineH1,'UserData'),'Completed'))
          ErrorStatus = 'unknown';
      elseif (SmlUD.Image.MaskXdata(1) == SmlUD.Image.MaskXdata(2)) || (SmlUD.Image.MaskYdata(1) == SmlUD.Image.MaskYdata(2))
          if length(SmlUD.Image.MaskXdata) == 2
              ErrorStatus = 'samevalue';
          else
              ErrorStatus = 'OK';
          end;
      else
          ErrorStatus = 'OK';
      end;

      if ishandle(handles.LineH1), delete(handles.LineH1), end;                 % Delete line H1
      if ishandle(handles.LineH2), delete(handles.LineH2), end;                 % Delete line H2
      if ishandle(GUI_Handle),                                                  % Restore the figure window settings.  I changed this to include the 'nochildren' option
          uirestore(SmlUD.ScoreMaskState);% ,'nochildren'),                     % because it was interferring with other ui-controls.  But I've discovered that it leads
      end;                                                                      % to the mask images loosing their 'ButtonDownFcn' capability. So 'nochildren' should NOT be used.
      SmlUD.ScoreMaskState = [];

      switch ErrorStatus
          case 'OK'                                                             % successful mask creation
              SmlUD = getappdata(GUI_Handle,'Small_Data');

              xData = SmlUD.Image.MaskXdata;                                    % The actual mask coordinates
              yData = SmlUD.Image.MaskYdata;
              if xData(1) ~= xData(end)
                  xData(end+1) = xData(1);
              end;
              if yData(1) ~= yData(end)                                         % The first and last points must be the same
                  yData(end+1) = yData(1);
              end;

              SmlUD.Image.MaskXindex = get(handles.TScoreHorizAxis,'Value');
              SmlUD.Image.MaskYindex = get(handles.TScoreVrtclAxis,'Value');
              SmlUD.Image.MaskXdata = xData;
              SmlUD.Image.MaskYdata = yData;
              SmlUD.Image.Mask = uint8(ComputeScoreMask(SmlUD.Image.MaskXdata,SmlUD.Image.MaskYdata,256,256));   % compute the mask matrix (256 x 256)

              set(handles.TScoreHorizAxis,'Enable','on');
              set(handles.TScoreVrtclAxis,'Enable','on');
              set(handles.ScoreMaskColour,'Enable','on');
              set(handles.ImageMaskColour,'Enable','on');
              set(handles.StartAdjustMask,'Visible','on');

              UpdateRHS(GUI_Handle);                                            % Call another function to compute and display mask
              UpdateLHS(GUI_Handle);
          case 'trapped'
              %disp('TRAPPED: An error occured during mask creation in the score space.  Please redraw mask.');
              MyCLA(handles.ZoomScoreAxes);
          case 'unknown'
              %disp('UNKNOWN: An error occured during mask creation in the score space.  Please redraw mask.');
              MyCLA(handles.ZoomScoreAxes);
          case 'samevalue'
              uiwait(errordlg({'The first and second point of the mask are the same.  Mask was not successfully completed'},'Mask creation error','modal'));
              MyCLA(handles.ZoomScoreAxes);
      end;    % end: switch
      return; % important: do not delete
  end; %end - if RHSaxis && ~(SmlUD.adjMode)

  if RHSaxis && SmlUD.adjMode
      % just a precaution: the mouse button mode is implemented in : adjMouseButton(varargin)
  end; % end: if RHSaxis && SmlUD.adjMode

  if LHSaxis && strcmp(get(handles.ModelButton,'State'),'on')
      SmlUD.Points.C(end+1) = round(cp(1,1));                                   % Now get the pixel intensities of the T_scores (for all scores) as if the user
      SmlUD.Points.R(end+1) = round(cp(1,2));                                   % had clicked the mouse buttons on T_score images rather than RGB images
      SmlUD.Points.List     = [SmlUD.Points.List; squeeze(SmlUD.Image.T(SmlUD.Points.R(end),SmlUD.Points.C(end),:))'];    % update the list
      setappdata(handles.GUI_Handle,'Small_Data',SmlUD);                        % Save

      set(handles.ScorePlotAxes,'NextPlot','add');                              % Mark the corresponding score space point
      hPlot = plot(double(SmlUD.Points.List(end,SmlUD.hAxis))+0.5 , double(SmlUD.Points.List(end,SmlUD.vAxis))+0.5 , '.' ,'MarkerSize',8, 'Color',SmlUD.ScoreMaskColour,'Parent',handles.ScorePlotAxes);
      set(hPlot,'UIContextMenu',handles.ScoreContextMenu);                      % Activate the context menu for this item

      set(handles.ImagePlotAxes,'NextPlot','add');                              % Plot an "X" where the user clicked in image space
      hPlot = plot(SmlUD.Points.C(end),SmlUD.Points.R(end),'x' ,'MarkerSize',8, 'Color',SmlUD.ImageMaskColour,'Parent',handles.ImagePlotAxes);
      set(hPlot,'UIContextMenu',handles.ImageContextMenu);
  end; %end - if not(RHSaxis)

return;                                                                         % End - of function

% --------------------------------------------------------------------
function ScoreMaskNextButtonDown(varargin)
% The user may click mouse buttons to create the mask in the score space - this functions handles the second and subsequent clicks
% Step 0: Load the required data
% Step 1: If a single click is made, add this point to the list, else
% Step 2: If a double click is made, then indicate that the mask creation is complete

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  ClickType = get(GUI_Handle,'SelectionType');                                  % What sort of mouse click occurred ?

  if ~strcmp(ClickType,'open')                                                  % if not a double click, then add new point
    ClickedPoint = get(handles.ScorePlotAxes,'CurrentPoint');
    SmlUD.Image.MaskXdata = [SmlUD.Image.MaskXdata(1:end-1) ClickedPoint(1,1) SmlUD.Image.MaskXdata(end)];
    SmlUD.Image.MaskYdata = [SmlUD.Image.MaskYdata(1:end-1) ClickedPoint(1,2) SmlUD.Image.MaskYdata(end)];
    setappdata(GUI_Handle,'Small_Data',SmlUD);                                  % Save the updated points in data storage
    set([handles.LineH1 handles.LineH2],'XData',SmlUD.Image.MaskXdata,'YData',SmlUD.Image.MaskYdata);
  end;

  if ~strcmp(ClickType,'normal')                                                % if not a normal single click, then complete mask
    set(handles.LineH1,'UserData','Completed');
  end;
return;

% --------------------------------------------------------------------
function GUIKeyPress_Callback(varargin)
% The user may press keys during the GUI's execution.  This functions handles these keyboard events
% Step 0: Load the required data
% Step 1: Perform the action depending on the key that was pressed
% Step 1: If delete or backspace pressed, then remove previous point
% Step 2: If enter or return, then complete the line selection process

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  KeyPressed = get(GUI_Handle,'CurrentCharacter');                              % What key has the user pressed ?

  switch KeyPressed
  %---------- Ctrl-E (to start Explore mode)
  case {char(5)}
      if get(handles.GUI_ExploreScores,'Value')
          return;
      elseif strcmp(get(handles.ModelButton,'State'),'on')
          set(handles.GUI_ExploreScores,'Value',1);
          ExploreScores_Callback(GUI_Handle);
      end;

  return;
  %---------- Ctrl-Z (to start/stop Zoom mode)
  case {char(26)}
      if strcmp(get(handles.ZoomStartButton,'State'),'off')
          set(handles.ZoomStartButton,'State','on');
      else
          set(handles.ZoomStartButton,'State','off');
      end
      ZoomStartButton(GUI_Handle);
  return;

  %---------- Delete or Backspace
  case {char(8), char(127)}
      if ishandle(handles.LineH1)                                               % The user may have accidentally pushed it
          switch length(SmlUD.Image.MaskXdata)
          case 0
              % do nothing - no data selected yet
          case 1
              % do nothing - no data selected yet
          case 2
              set(handles.LineH1,'UserData','Completed');                       % The user wants to go too far back, so start over
          otherwise
              SmlUD.Image.MaskXdata(end-1) = [];                                % Remove the last selected point
              SmlUD.Image.MaskYdata(end-1) = [];
          end;
          setappdata(GUI_Handle,'Small_Data',SmlUD);                            % Save the updated points in data storage
          set([handles.LineH1 handles.LineH2],'XData',SmlUD.Image.MaskXdata,'YData',SmlUD.Image.MaskYdata);
      end;

  %---------- Enter or return
  case {char(13), char(3)}
      if ishandle(handles.LineH1)
          set(handles.LineH1,'UserData','Completed');                           % Indicate that we are finished selecting mask
      end;

  %---------- Space bar
  case char(32)
      if SmlUD.ZoomMode                                                         % Only appropriate in Zoom mode
          MseUD = getappdata(GUI_Handle,'Mouse_Data');
          if (MseUD.Freeze == 1)
              MseUD.Freeze = 0;
              if SmlUD.CountMessage <= 6
                  set(handles.ZoomSpaceMessage,'String','Use "Space bar" to freeze zoomed view. Press "h" for help.');
                  SmlUD.CountMessage = SmlUD.CountMessage + 1;
              else
                  set(handles.ZoomSpaceMessage,'String','');
              end;
          else
              MseUD.Freeze = 1;
              if SmlUD.CountMessage <= 6
                  set(handles.ZoomSpaceMessage,'String','Push "Space bar" again to unfreeze current view.');
                  SmlUD.CountMessage = SmlUD.CountMessage + 1;
              else
                  set(handles.ZoomSpaceMessage,'String','Image is frozen (push "space bar" to unfreeze)','Visible','on');
              end;
          end;
      end;
      setappdata(GUI_Handle,'Mouse_Data',MseUD);                                % Save all the changes
      setappdata(GUI_Handle,'Small_Data',SmlUD);

  %---------- Up arrow
  case char(30)
      if SmlUD.ZoomMode
          ZoomChangeButton(handles.ZoomMoreButton);                             % Trick the function into thinking it was called by the pushbuttons
      end

  %---------- Down arrow
  case char(31)
      if SmlUD.ZoomMode
          ZoomChangeButton(handles.ZoomLessButton);
      end

  %---------- Small letter "h" or capital "H" (for help)
  case {char(72), char(104)}
      if SmlUD.ZoomMode                                                         % Only appropriate in Zoom mode
          msgbox({'The zoom mode helps one select pixels in the image space more', ...
                  'accurately.  However as one moves the mouse the limits of the', ...
                  'zoomed area are constantly adjusted.  Press the space bar to', ...
                  'keep the limits fixed temporarily.  This will help you select pixels', ...
                  'in the image space.  ', ' '...
                  'After you have finished selecting the desired pixels you should', ...
                  'press the space bar again to unfreeze the image space.'}, ...
                'Help in zoom mode','help','replace');
      end;
  end;% end: switch

  %no further instructions allowed here: the functions above do many different things in different contexts
return;


% --------------------------------------------------------------------
function varargout = ResetModel_Callback(varargin)
% The user wants to reset the model: removes clicked points and additionally prompts the user if s/he would like to delete all saved masks.
% Step 0: Get background info
% Step 1: Delete clicked points
% Step 2: Delete any saved masks
% Step 3: Save and update the GUI

% STEP 0
  global SmlUD
  varargout = [];
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1:
  SmlUD.hAxis           = 1;          set(handles.TScoreHorizAxis,'Value',1);   % and adjust the callback values accordingly
  SmlUD.vAxis           = 2;          set(handles.TScoreVrtclAxis,'Value',2);
  SmlUD.Points.List     = [];
  SmlUD.Points.R        = [];
  SmlUD.Points.C        = [];
  SmlUD.ImageMaskColour = [1 1 1];    set(handles.ImageMaskColour,'Value',1);
  SmlUD.ScoreMaskColour = [0 1 1];    set(handles.ScoreMaskColour,'Value',6);

  uiresume(GUI_Handle);                                                         % These two steps are in-case we've gotten trapped in a mask adjustment mess up
  drawnow;
  SmlUD.adjMode = 0;
  set([handles.StartAdjustMask handles.EndAdjustMask],'Visible','off')

  MyCLA(handles.ZoomScoreAxes);
  MyCLA(handles.ZoomImageAxes);
  set(handles.ZoomScoreInfo,'Visible','off');
  SmlUD.ScoreMaskState = [];
  SmlUD.Image.Mask = uint8(zeros(256));
  SmlUD.Image.MaskXdata = [];
  SmlUD.Image.MaskYdata = [];
  SmlUD.Image.MaskXindex = [];
  SmlUD.Image.MaskYindex = [];

% STEP 3:
  setappdata(GUI_Handle,'Small_Data',SmlUD);
  UpdateLHS(GUI_Handle);
  UpdateRHS(GUI_Handle);
return;


% --------------------------------------------------------------------
function ResetImageClicks_Callback(varargin)
% The user wants to reset the clicked points in the image space
%
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Change the fields in SmlUD related to mouse clicks to empty
% Step 2. Save the SmlUD structure and redraw the image space and score space images

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1
  SmlUD.Points.List     = [];
  SmlUD.Points.R        = [];
  SmlUD.Points.C        = [];

% STEP 2:
  setappdata(GUI_Handle,'Small_Data', SmlUD);
  UpdateLHS(GUI_Handle);
  UpdateRHS(GUI_Handle);
return;


% --------------------------------------------------------------------
function HistogramAnalysis_Callback(varargin)
% [This function was HistogramAnalysis.m, called to display histogram statistics of the images.]
% This function draws the figure window, adds the uicontrols and axes in the correct
% places, stores required information in the user data and returns control to the workspace.

% Steps to follow
% 1. If window exists, delete it, and then create a new one.
% 2. Determine image properties and store them
% 3. Setup the controls in the window
% 4. Make the window appear and put default data there to start off with

% STEP 1: setting up the window
  global SmlUD;
  FindHandle = findobj(allchild(0), 'tag','HistogramAnalysis');                 % Get the histogram handle (if it exists)
  GUI_Handle = findobj(allchild(0), 'tag','MACC_MIA_GUI');                      % Get the MACCMIA GUI_Handle (the fool-proof way)
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if ~isempty(FindHandle)                                                       % The window exists,
    delete(FindHandle(1));                                                      % so delete it
  end;
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

  ScreenDepth = get(0,'ScreenDepth');
  if ScreenDepth>8,   grayres=256;  else   grayres=128; end;

  HistogramAnalysis = figure('Name','MACC MIA - Histogram Analysis', ...        % The name appearing in the window
                             'NumberTitle','off', ...                           % Prevents display of Figure No. x in title
                             'tag','HistogramAnalysis',...                      % Figure tag
                             'HandleVisibility','on', ...                       % The handles are visible to the workspace
                             'Visible','off', ...                               % Make it invisible while we set it up
                             'Resize','off', ...                                % Do not allow the user to resize the window
                             'BusyAction','queue', ...                          % Put other conflicting commands in a queue
                             'Interruptible','off', ...                         % Don't interrupt while setting it up
                             'Color',[0.8 0.8 0.8], ...                         % Background colour
                             'Pointer','watch', ...                             % Make the pointer look like an hour-glass
                             'DoubleBuffer','on', ...                           % Allows fast redrawing of images
                             'IntegerHandle','off', ...                         % Prevent handles from being integers
                             'Colormap', gray(grayres), ...                     % Set the default colourmap
                             'Menubar','none' ...                               % Do not allow the menu bar to appear
                             );
  FigurePosition = get(HistogramAnalysis, 'position');                          % Get the position of the current figure
  FigurePosition(3:4) = [450 600];                                              % Now adjust its size according to our needs
  ScreenSize = get(0,'ScreenSize');

  if ( (FigurePosition(3)+10 > ScreenSize(3) ) || ( FigurePosition(4)+45 > ScreenSize(4)) )
      delete(HistogramAnalysis);                                                % Screen size is too small
      error('Screen resolution is too low (or text fonts are too big) to run this analysis');
  end
  dx = ScreenSize(3) - FigurePosition(1) - FigurePosition(3) - 10;
  dy = ScreenSize(4) - FigurePosition(2) - FigurePosition(4) - 45;
  if (dx < 0)
      FigurePosition(1) = max(5,FigurePosition(1) + dx);
  end
  if (dy < 0)
      FigurePosition(2) = max(5,FigurePosition(2) + dy);
  end
  set(HistogramAnalysis, 'position', FigurePosition);
  Cols = FigurePosition(3);


  % STEP 2: Make a composite image from all channels - simple averaging across all channels
  SmlUD.Image.Composite = zeros(SmlUD.Image.nRows,SmlUD.Image.nCols);
  SmlUD.Image.Composite = sum(double(SmlUD.Image.X),3) / SmlUD.Image.nLayers;
  SmlUD.Image.Composite = uint8(SmlUD.Image.Composite);                         % This step chops negative values from double images, to be zero values
  setappdata(GUI_Handle,'Small_Data', SmlUD);

  % STEP 3: Setup the controls
  Space             = 25;                                                       % Spacing
  ImageSize         = 175;
  ImageSpace        = (Cols - 2*ImageSize)/3;                                   % Inter-image spacing
  BigSpace          = 50;                                                       % Space between top and bottom
  PopupHeight       = 37.5;                                                     % Height of popup button
  PopupWidth        = 150;                                                      % Width of popup button
  WinColour         = [0.8 0.8 0.8];                                            % Window colour

  % Defaults for axes
  Axs.Interruptible  = 'off';
  Axs.BusyAction     = 'queue';
  Axs.Units          = 'Pixels';
  Axs.Parent         = HistogramAnalysis;
  Axs.ydir           = 'reverse';
  Axs.XLim           = [0.5 128.5];
  Axs.YLim           = [0.5 128.5];
  Axs.CLim           = [0 1];
  Axs.XTick          = [];
  Axs.YTick          = [];

  % Defaults for controls
  Ctl.Interruptible = 'off';
  Ctl.BusyAction    = 'queue';
  Ctl.Units         = 'Pixels';
  Ctl.Parent        = HistogramAnalysis;

  % Defaults for the text
  Txt.Interruptible = 'off';
  Txt.BusyAction    = 'queue';
  Txt.Parent        = HistogramAnalysis;
  Txt.Style         = 'text';
  Txt.Units         = 'pixels';
  Txt.Horiz         = 'left';
  Txt.Background    = WinColour;
  Txt.Foreground    = 'black';
  Txt.FontSize      = 10;

  % Set up place for the image
  handles.hImageAxes  = axes(Axs,'Position',[ImageSpace BigSpace+3*Space+2*PopupHeight+175 ImageSize ImageSize],'Tag','Axes to display the current channel');
  handles.hImageStretch = axes(Axs,'Position',[ImageSpace+ImageSize+ImageSpace BigSpace+3*Space+2*PopupHeight+175 ImageSize ImageSize],'Tag','Axes to display the current channel');

  % Set up place for the histogram
  Strip = 0.075;
  handles.hHistAxes = axes(Axs,'Position', [Space BigSpace+1*Space+1*PopupHeight+Strip*175 400 (1-Strip)*175],'Tag','The histogram axes');

  % Setup place for the stripe at the bottom of the histogram
  handles.hStripAxes = axes(Axs,'Position',[Space BigSpace+1*Space+1*PopupHeight           400 Strip*175]);
  set(handles.hStripAxes,'XLim',[0 1],'YLim',[0 1],'XTick',[],'YTick',[]);


  % Channel selection popup menu
  if SmlUD.RGBmode
      PopupString = 'All channels   - Averaged';
      PopupString = strvcat(PopupString,'Channel 1 of 3 - Red','Channel 2 of 3 - Green','Channel 3 of 3 - Blue');
  else
      PopupString = 'All channels - Averaged';
      for k = 1:SmlUD.Image.nLayers
          PopupString = strvcat(PopupString,['Channel ',num2str(k),' of ',num2str(SmlUD.Image.nLayers)]);
      end;
  end;
  handles.CurrentChannelPopop = uicontrol(Ctl, ...
                                     'Style', 'Popupmenu',...
                                     'Position',[Cols-Space-PopupWidth BigSpace PopupWidth PopupHeight], ...
                                     'Enable','on', ...
                                     'String',PopupString, ...
                                     'Tag','OpPop',...
                                     'Callback',@UpdateHistogram);

  % Controls that just display static text
  handles.hMin   = uicontrol(Txt,'Position',[Space     BigSpace-10 60 20],'String','Min:   ');
  handles.hMax   = uicontrol(Txt,'Position',[Space+130 BigSpace-10 60 20],'String','Max:   ');
  handles.hMean  = uicontrol(Txt,'Position',[Space     BigSpace-40 60 20],'String','Mean:  ');
  handles.hStdDv = uicontrol(Txt,'Position',[Space+130 BigSpace-40 60 20],'String','Std. Dev:');
  handles.hNumPx = uicontrol(Txt,'Position',[Space+260 BigSpace-40 90 20],'String','Num Pixels:');

  % Controls that just display dynamic text, but even if Style='edit', the user won't be allowed to change them
  Txt.Style         = 'edit';                                                   % Rather change selected fields than redefine
  Txt.Horiz         = 'right';
  Txt.Enable        = 'inactive';
  handles.VMin   = uicontrol(Txt,'Position',[Space+60  BigSpace-10 60 20],'String',' ');
  handles.VMax   = uicontrol(Txt,'Position',[Space+190 BigSpace-10 60 20],'String',' ');
  handles.VMean  = uicontrol(Txt,'Position',[Space+60  BigSpace-40 60 20],'String',' ');
  handles.VStdDv = uicontrol(Txt,'Position',[Space+190 BigSpace-40 60 20],'String',' ');
  handles.VNumPx = uicontrol(Txt,'Position',[Space+350 BigSpace-40 70 20],'String',' ');

  % -----------------------------------
  % STEP 4 - Make the window appear and put default data there
  % -----------------------------------

  % After all this setting up we can save the data and make the figure window visible
  setappdata(handles.GUI_Handle,'Handles_Data', handles);

  % Display the figure and histogram
  if SmlUD.RGBmode
      ShowImage(SmlUD.Image.X,handles.hImageAxes,[],[],[],'direct');
      set(get(handles.hImageAxes,'title'),'string','True RGB image','Visible','on');
      ShowImage(SmlUD.Image.Composite, handles.hImageStretch,[],[],[],'direct');
      set(get(handles.hImageStretch,'title'),'string','Composite (average) image','Visible','on');
  else
      ShowImage(SmlUD.Image.Composite,handles.hImageAxes,[],[],'ScaledCLim','scaled');
      set(get(handles.hImageAxes,'title'),'string','Arithmetic mean over all channels','Visible','on');
      limits = Compute_ImageLimits(SmlUD.Image.Composite,[0.01 0.99]);
      adjust = Compute_Stretch(SmlUD.Image.Composite,limits,0);
      ShowImage(adjust,handles.hImageStretch,[],[],'ScaledCLim','scaled');
      set(get(handles.hImageAxes,'title'),'string','Saturation stretch','Visible','on');
  end;
  MakeHistogram(HistogramAnalysis,'Average',0);                                 % Show the composite image to start off with

  drawnow;
  set(HistogramAnalysis,'Pointer','arrow', ...                                  % Set the mouse to an arrow head
                        'HandleVisibility','callback', ...                      % only callbacks can see the handles
                        'Visible','on');
return;

% --------------------------------------------------------------------
function y = Compute_Histogram(x,bins)
% This function avoids using imhist from the image processing toolbox.  It accepts uint8 and double without any problems.
  bins = bins(:)';
  bins = bins + max(eps,eps*abs(bins));
  y = histc(x(:),[-inf bins],1);                                                % Use the Matlab MEX histogram function to do the work
  y(2,:) = y(2,:)+y(1,:);
  y(end-1,:) = y(end-1,:)+y(end,:);
  y = y(1:end-1,:);
return

% --------------------------------------------------------------------
function limits = Compute_ImageLimits(x,limits,varargin)
% This function avoids using "stretchlim" from the image processing toolbox
  if nargin==2
      nbins = 0:255;
  else
      nbins = varargin{1};
  end;
  %tol_low = limits(1);
  %tol_high = limits(2);
  %[m,n,p] = size(x);
  p = size(x,3);

  for k = 1:p                                                                   % Find limits for each layer by
      N = Compute_Histogram(x(:,:,k),nbins);
      CumHist = cumsum(N)/sum(N);                                               % Computing the cumulative histogram
      low = min(find(CumHist > limits(1)));                                     % and taking the appropriate limits
      high = min(find(CumHist >= limits(2)));
      lowhigh(:,k) = [low;high];
  end;
  nbins = 256;
  limits = (lowhigh - 1)/(nbins-1);
return

  % --------------------------------------------------------------------
function y = Compute_Stretch(x,limits,rgb)
% This function avoids using "imadjust" from the image processing toolbox
  [m,n,p] = size(x);
  y = double(zeros(m,n,p));
  if rgb
      for k = 1:p
          layer = double(x(:,:,k)) ./ 255;
          layer = max( limits(1,k), min(limits(2,k),layer) );
          y(:,:,k) = ( layer - limits(1,k) ) ./ (limits(2,k) - limits(1,k));
      end;
      y = uint8(y.*255);
  else
      minx = double(min(x(:)));
      maxx = double(max(x(:)));
      rangex = maxx - minx;
      limits = minx + limits .* rangex;
      x    = double(max(limits(1),min(limits(2),x)));
      y    = (x-limits(1)) ./ ( limits(2)-limits(1) );
  end;
return;


% --------------------------------------------------------------------
function MakeHistogram(FigureHandle,ChannelType,ChannelNumber)
% This function displays the histogram information for the required channel.
% Inputs : FigureHandle - the handle to the figure window containing the histogram axis (HistogramAnalysis)
%          ChannelType  - either "Average" or "Red" or "Green" or "Blue" - last three are only for true RGB images
%          ChannelNumber- the n^th channel of the image data, ie: SmlUD.Image.X(:,:,n)
%                       - if ChannelNumber==0, the user wants the averaged, or composite image: SmlUD.Image.Composite

  global SmlUD
  FigureHandle = getGUIHandle(FigureHandle,'HistogramAnalysis');                % Get the histogram handle (if it exists)
  GUI_Handle = findobj(allchild(0), 'tag','MACC_MIA_GUI');                      % Get the MACCMIA GUI_Handle (the fool-proof way)
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  setptr(FigureHandle, 'watch');

  if ChannelNumber > 0
      ImageData = SmlUD.Image.X(:,:,ChannelNumber);
      minX = SmlUD.Image.minX(ChannelNumber);
      maxX = SmlUD.Image.maxX(ChannelNumber);
      avgX = SmlUD.Image.avgX(ChannelNumber);
      stdX = sqrt(SmlUD.Image.varX(ChannelNumber));
  else
      ImageData = SmlUD.Image.Composite;
      [sumX,avgX,minX,maxX,varX] = libmia('stats',ImageData);
      stdX = sqrt(varX);
  end
  nPixels = SmlUD.Image.nRows * SmlUD.Image.nCols;                              % Number of pixels in image channel


  StripImage = repmat((1:256)/256,[1 1 3]);
  switch ChannelType
  case 'Average'
    % Nothing to do, StripImage is the same as above
  case 'Red'
      StripImage(:,:,2:3)   = repmat((1:256)*0,[1 1 2]);
  case 'Green'
      StripImage(:,:,1:2:3) = repmat((1:256)*0,[1 1 2]);
  case 'Blue'
      StripImage(:,:,1:2)   = repmat((1:256)*0,[1 1 2]);
  otherwise
      setptr(FigureHandle, 'arrow');                                            % This case should not occur - error - but exit gracefully
      return;
  end;


  % Put the histogram stripe first and clean up the axis limits
  MyCLA(handles.hStripAxes);                                                     % Make the histogram stripe active
  set(FigureHandle,'CurrentAxes',handles.hStripAxes);
  StripLimits = [get(handles.hStripAxes,'XLim') get(handles.hStripAxes,'YLim')]; % Set the limits for the stripe
  HistLimits  = zeros(1,4);
  StripX      = [(1/256)/2 1-(1/256)/2];
  switch SmlUD.Image.Class
  case 'uint8'
     Bins = 0:255;
     Number = Compute_Histogram(ImageData(:),Bins);
     HistLimits(1)  = -4;                                                       % Set the lower and upper values
     HistLimits(2)  = 255+4;
     HistLimits(4)  = max(Number)*1.05; % min(4*Variance,max(Number)*1.1);      % Set the maximum y-scale
     StripLimits(1) = -4;                                                       % Do this to show extreme edges properly
     StripLimits(2) = 255+4;
     StripX = 255*StripX;
     j=line([0 255 255 0 0],[0 0 1 1 0]);   set(j,'linestyle','-')              % Puts a border around the colour stripe
     set(j,'Color',get(handles.hStripAxes,'XColor'))
  case 'uint16'
     %Bins = 0:128:65535;
     %Number = Compute_Histogram(ImageData(:),Bins);
     setptr(FigureHandle, 'arrow');
     delete(FigureHandle);
     error('Case not handled yet - come back to this still');
     %StripX = 65535*StripX;
     %Limits(1) = -4;
     %Limits(2) = 65535;
  case 'double'
     rangeX = maxX - minX;
     HistLimits(1)  = minX - (rangeX)*0.05;                                     % 5% of image range below the lowest pixel intensity
     HistLimits(2)  = maxX + (maxX-minX)*0.05;
     Bins = linspace(HistLimits(1),HistLimits(2),256);
     Number = Compute_Histogram(ImageData(:),Bins);
     HistLimits(4)  = max(Number)*1.05; % min(4*Variance,max(Number)*1.1);
     StripLimits(1) = HistLimits(1);
     StripLimits(2) = HistLimits(2);
     StripX = (rangeX)*StripX+minX;
     j=line([minX maxX maxX minX minX],[minX minX maxX maxX minX]);
     set(j,'linestyle','-');                                                    % Puts a border around the colour stripe
     set(j,'Color',get(handles.hStripAxes,'XColor'))
  otherwise
     setptr(FigureHandle, 'arrow');
     delete(FigureHandle);
     error('Histogram error: The input image must be of class uint8, uint16, or double.');
  end
  image(StripX,[0 1],StripImage);
  set(handles.hStripAxes,'YTickLabelMode','manual','YTickLabel','','Ytick',[]);
  axis(StripLimits);                                                            % Update the histogram stripe

  % Now draw the histogram itself
  MyCLA(handles.hHistAxes);                                                     % Select the axes to plot the histogram
  set(FigureHandle,'CurrentAxes',handles.hHistAxes);
  stem(Bins,Number);                                                            % Use a stem plot
  h = get(handles.hHistAxes,'children');                                        % Get the handles of the axes's children
  hStems = findobj(h,'flat','Marker','o');                                      % Remove 'o's from stem plot
  set(hStems,'Marker','none');

  set(handles.hHistAxes,'XTickLabelMode','manual','XTickLabel','');             % Remove axis labels
  set(handles.hHistAxes,'Xtick', [], 'Ytick', []);
  axis off;
  axis(HistLimits);                                                             % Update the axis limits
  set(get(handles.hHistAxes,'title'),'string','Histogram','Visible','on');      % and title

  %Update the displayed information
  set(handles.VMin,  'String',num2str(minX));
  set(handles.VMax,  'String',num2str(maxX));
  set(handles.VMean, 'String',num2str(avgX));
  set(handles.VStdDv,'String',num2str(stdX));
  set(handles.VNumPx,'String',num2str(nPixels));

  setptr(FigureHandle, 'arrow');                                                % Change the mouse pointer back to an arrow
return;


% --------------------------------------------------------------------
function UpdateHistogram(varargin)
% ------------------------------------------------------------------
% The user has selected to view the histogram of other channels or other images in the data file.
% Change the histogram axes to display this new information.
%
% Steps to follow: 1 - Get the desired channel and the current image number
%                  2 - Update the histogram accordingly

% Step 1
  global SmlUD
  FigureHandle = getGUIHandle(varargin{1},'HistogramAnalysis');                 % Get the histogram handle (if it exists)
  GUI_Handle = findobj(allchild(0), 'tag','MACC_MIA_GUI');                      % Get the MACCMIA GUI_Handle (the fool-proof way)
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  ChannelSelected = get(handles.CurrentChannelPopop,'value');                   % Get the channel the user selected

  % Step 2
  MyCLA(handles.hImageStretch);
  set(FigureHandle,'CurrentAxes',handles.hImageStretch);

  if SmlUD.RGBmode

      if ChannelSelected == 1
          ShowImage(SmlUD.Image.Composite,handles.hImageStretch,[],[],[],'direct');
          set(get(handles.hImageStretch,'title'),'string','Composite (average) image','Visible','on');
      else
          ShowImage(SmlUD.Image.X(:,:,ChannelSelected-1),handles.hImageStretch,[],[],[],'direct');
      end


      switch ChannelSelected
          case 1
              MakeHistogram(FigureHandle,'Average',0);
          case 2
              set(get(handles.hImageStretch,'title'),'string','Red channel','Visible','on')
              MakeHistogram(FigureHandle,'Red',1);
          case 3
              set(get(handles.hImageStretch,'title'),'string','Green channel','Visible','on')
              MakeHistogram(FigureHandle,'Green',2);
          case 4
              set(get(handles.hImageStretch,'title'),'string','Blue channel','Visible','on')
              MakeHistogram(FigureHandle,'Blue',3);
      end
  else
      if ChannelSelected == 1
          [sumX,meanX,minX,maxX,varX] = libmia('stats',SmlUD.Image.Composite);
      else
          minX = double(SmlUD.Image.minX(ChannelSelected-1));
          maxX = double(SmlUD.Image.maxX(ChannelSelected-1));
      end
      rangeX = maxX - minX;
      low  = minX - rangeX*0.05;
      high = maxX + (maxX-minX)*0.05;
      bins = linspace(low,high,256);
      if ChannelSelected == 1
          ShowImage(SmlUD.Image.Composite,handles.hImageAxes,[],[],'ScaledCLim','scaled');
          set(get(handles.hImageAxes,'title'),'string','Arithmetic mean over all channels','Visible','on');
          limits = Compute_ImageLimits(SmlUD.Image.Composite,[0.01 0.99],bins);
          if sum(limits)==2
              limits = Compute_ImageLimits(SmlUD.Image.Composite,[0.0 1.0]);    % Redo, if both limits equal one
          end;
          adjust = Compute_Stretch(SmlUD.Image.Composite,limits,0);
      else
          ShowImage(SmlUD.Image.X(:,:,ChannelSelected-1),handles.hImageAxes,[],[],'ScaledCLim','scaled');
          title(['Channel ',num2str(ChannelSelected-1)]);
          limits = Compute_ImageLimits(SmlUD.Image.X(:,:,ChannelSelected-1),[0.01 0.99],bins);
          if sum(limits)==2
              limits = Compute_ImageLimits(SmlUD.Image.X(:,:,ChannelSelected-1),[0.0 1.0]);     % Redo, if both limits equal one
          end;
          adjust = Compute_Stretch(SmlUD.Image.X(:,:,ChannelSelected-1),limits,0);
      end
      ShowImage(adjust,handles.hImageStretch,[],[],[],'scaled');
      set(get(handles.hImageStretch,'title'),'string',['Saturation stretch on channel ',num2str(ChannelSelected-1)],'Visible','on');
      MakeHistogram(FigureHandle,'Average',ChannelSelected-1);
  end;
return;


% --------------------------------------------------------------------
function ExploreScores_Callback(varargin)
% Enter "Explore" mode - where the user can explore aspects of the score space and its relation to the image space.
% Step 0: Load any required data for this function
% Step 1: Turn Explore Mode Off by undoing any GUI changes made
% Step 2: Set the GUI settings that will allow the Explore mode to work

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  MseUD   = getappdata(GUI_Handle,'Mouse_Data');
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;
  ExploreOnOff = get(handles.GUI_ExploreScores,'Value');                        % See if Explore mode is on or off

% STEP 1 - Turn Explore OFF
  if not(ExploreOnOff)                                                          % The user wants to end exploring: the code will resume in this function at the "uiwait" point
    uiresume(GUI_Handle);
    return;
  end;

% STEP 2 - Turn Explore ON
  if SmlUD.adjMode                                                              % No exploring during adjustment mode
      set(handles.GUI_ExploreScores,'Value',0);
      return;
  end;
  if SmlUD.ZoomMode                                                             % First turn zooming off
      set(handles.ZoomStartButton,'State','off');
      ZoomStartButton(GUI_Handle);                                              % This will set SmlUD.ZoomMode to zero
      ReturnZoomOn = 1;
  else
      ReturnZoomOn = 0;
  end;

  prevHorizString = get(handles.TScoreHorizAxis,'String');
  set(handles.TScoreHorizAxis,'String',get(handles.TScoreVrtclAxis,'String'),'Value',1);    % The update function will correct any inconsistencies with this value

  SmlUD.XLim = [0 SmlUD.Image.nCols] + 0.5;                                     % Reset the zoom extents to display the image space
  SmlUD.YLim = [0 SmlUD.Image.nRows] + 0.5;
  SmlUD.expMode = 1;
  setappdata(GUI_Handle,'Small_Data', SmlUD);                                   % Save the changes so far

  UpdateLHS(GUI_Handle);                                                        % But these functions must be modified first to notice that explore mode is on
  UpdateRHS(GUI_Handle);
  MyCLA(handles.ZoomImageAxes);                                                 % Remove the zoom in of the image space

  % In future: just use the "RemoveInterfaceWidgets" function, and then add back the necessary widgets required

  EnableOff = [handles.File handles.FileOpenButton handles.FilePrintButton handles.ZoomStartButton ...
               handles.ZoomMoreButton handles.ZoomLessButton handles.StartAdjustMask];
  EnableOn = EnableOff;
  for k = 1:length(EnableOff)                                                   % The explore mode could be entered when some of these UI-widgets
      if strcmp(get(EnableOff(k),'Enable'),'off')                               % were not enabled, eg: ApplyAll.  To avoid turning them on
          EnableOn(k) = Inf;                                                    % afterwards, make a list, "EnableOn" of the widgets that need
      else                                                                      % to be turned on.
          set(EnableOff(k),'Enable','off');
      end
  end
  if ~isempty(MseUD)
      MseUD.Freeze = 0;
      set(handles.ZoomSpaceMessage,'String','Press "Space" to freeze zoomed view. Press "h" for help.','Visible','off');
      setappdata(GUI_Handle,'Mouse_Data',MseUD);
  end;

  set([handles.ResetModel handles.ResetImageClicks],'Visible','off');
  set(handles.ScoreMaskColourText,'String',{'Mouse','Colour'});

  set(handles.expImageSideMaskSizeText,'Visible','on');                         % Activate callback to change the mask size
  set(handles.expScoreSideMaskSizeText,'Visible','on');
  set(handles.expScoreMaskSize,'Visible','on','String',num2str(SmlUD.ScoreMaskSize));
  set(handles.expImageMaskSize,'Visible','on','String',num2str(SmlUD.ImageMaskSize));

% Step 2b - Setup settings for the Explore Mode regarding the mouse
  SmlUD.expPointerShape  = ones(16).*NaN;                                       % Create a new pointer shape: transparent with a little square dot in the center
  SmlUD.expPointerShape(8:9,8:9) = ones(2);
  N = SmlUD.ScoreMaskSize/2;
  X = [128-N+1 128-N+1 128+N 128+N   128-N+1];
  Y = [128-N+1 128+N   128+N 128-N+1 128-N+1];
  handles.expLine = line('Parent',handles.ScorePlotAxes,'XData',X,'YData',Y,'Visible','on','Clipping','on','LineStyle','-','Color',SmlUD.ScoreMaskColour);
  handles.hZoomScoreImage = image([128-N+1 128+N],[128-N+1 128+N], SmlUD.Image.ScoreImage((128-N+1):(128+N),(128-N+1):(128+N)), ...
                                  'Parent',handles.ZoomScoreAxes,'CDataMapping','direct');
  set(handles.ZoomScoreAxes,      'XTick',[128-N+1 128+N],'YTick',[128-N+1 128+N],'DataAspectRatio',[1 1 1],'TickLength',[0 0]);

  prevKeyPress = get(GUI_Handle,'KeyPressFcn');
  prevWBDF = get(GUI_Handle,'WindowButtonDownFcn');
  set(GUI_Handle,'KeyPressFcn','','WindowButtonDownFcn','','ColorMap',SmlUD.ScoreSpaceColourMap);

% Step 2c - Save the settings and redraw the score and image space plots
  setappdata(GUI_Handle,'Small_Data',SmlUD);                                    % Store the large data storage structure
  setappdata(GUI_Handle,'Handles_Data',handles);                                % Store the handles structure

% Step 2d - Wait for the user to finish exploring and then clean up

  %ErrorCatch = 0;  : not handled yet
  try
      uiwait(GUI_Handle);
  catch
      %ErrorCatch = 1;      : not handled yet
  end;

  if ishandle(handles.expLine)
      delete(handles.expLine);
  end;
  if ishandle(GUI_Handle),
      set(GUI_Handle,'KeyPressFcn',prevKeyPress);                               % Restore the figure data
      set(GUI_Handle,'WindowButtonDownFcn',prevWBDF);
  end

  set(handles.ScoreMaskColourText,'String',{'Mask','Colour'});                  % UNDO some settings that were made for ExploreMode
  for k = 1:length(EnableOn)
      if ishandle(EnableOn(k))
          set(EnableOn(k),'Enable','on');                                       % Re-enable UI-widgets that were turned off
      end
  end
  set(handles.TScoreHorizAxis,'String',prevHorizString,'Value',1);

  set(handles.expImageSideMaskSizeText,'Visible','off');                        % Return off the callback to change the mask size
  set(handles.expScoreSideMaskSizeText,'Visible','off');
  set([handles.expScoreMaskSize handles.expImageMaskSize handles.ZoomScoreInfo],'Visible','off');
  set([handles.ResetModel handles.ResetImageClicks],'Visible','on');
  MyCLA(handles.ZoomScoreAxes);                                                 % Remove the zoom in on the score space

  if ReturnZoomOn
      set(handles.ZoomStartButton,'State','on');
      ZoomStartButton(GUI_Handle);                                              % This will set SmlUD.ZoomMode to zero
  end;

  SmlUD.expMode = 0;
  setappdata(GUI_Handle,'Small_Data',SmlUD);                                    % Save data, then update the images
  UpdateLHS(GUI_Handle);
  UpdateRHS(GUI_Handle);
return;


% --------------------------------------------------------------------
function expMaskSizeAdjust_Callback(varargin)
% The user wishes to change the size of the square mask used by the mouse for exploring either the score space or the image space
% Approach:
% 1. Get the size, convert to an even integer between 0 and 256/2 = 128
% 2. Save the new size data in SmlUD
% 3. Update the information on the interface
% STEP 0

  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end
  newSizeImage = ceil(real(str2double(get(handles.expImageMaskSize,'String'))));
  newSizeScore = ceil(real(str2double(get(handles.expScoreMaskSize,'String'))));

  if isnan(newSizeScore)                                                        % Check that the user entered numbers, if garbage
      newSizeScore = SmlUD.expScoreMaskSize;                                    % then set it to the default value
  end
  if isnan(newSizeImage)
      newSizeScore = SmlUD.expImageMaskSize;
  end

  if mod(newSizeScore,2)                                                        % If an odd number, make it even, go one up
      newSizeScore = newSizeScore+1;
  end
  if mod(newSizeImage,2)
      newSizeImage = newSizeImage+1;
  end

  if (newSizeScore < 5)                                                         % If less than 5, set it to four
      newSizeScore = 4;
  end
  if (newSizeImage < 5)
      newSizeImage = 4;
  end

  if (newSizeScore>128)                                                         % If greater than 128 set to 128
      newSizeScore = 128;
  end
  maxSize = 0.5 * max(SmlUD.Image.nRows,SmlUD.Image.nCols);
  if (newSizeImage>maxSize)
      newSizeImage = maxSize;
  end

  SmlUD.ScoreMaskSize = newSizeScore;
  SmlUD.ImageMaskSize = newSizeImage;
  set(handles.expScoreMaskSize,'String',num2str(newSizeScore));
  set(handles.expImageMaskSize,'String',num2str(newSizeImage));
  setappdata(GUI_Handle,'Small_Data',SmlUD);                                    % Store the new mask size
return;


% --------------------------------------------------------------------
function AdjustMask_Callback(varargin)
% Enter "Adjust " mode - where the user can adjust the mask to capture the desired image space features
% Step 0: Load any required data for this function
% Step 2: Set the GUI settings that will allow the Explore mode to work

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1: Find the current mask and make it appear "editable"
  hPatch = findobj(handles.ScorePlotAxes,'type','patch');
  if ishandle(hPatch)
      delete(hPatch);
  end;
  xData = SmlUD.Image.MaskXdata;
  yData = SmlUD.Image.MaskYdata;
  handles.adjLine  = line('Parent',handles.ScorePlotAxes,'XData',xData,'YData',yData,'Clipping','off', ....
                          'LineStyle','-','Color',SmlUD.ScoreMaskColour,'Marker','o');
  handles.adjPoint = line('Parent',handles.ScorePlotAxes,'XData',0,'YData',0,'Clipping','off','LineStyle','none', ...
                          'Color',SmlUD.ScoreMaskColour,'Marker','o','MarkerFaceColor',SmlUD.ScoreMaskColour,'Visible','off');
  SmlUD.adjMode = 1;                                                            % We are in adjustment mode
  SmlUD.adjModeButton = 0;                                                      % The mouse button is not pushed down
  set([handles.TScoreHorizAxis handles.TScoreVrtclAxis],'Enable','off');        % Disable certain buttons
  set([handles.FileOpenButton handles.FileSaveButton handles.FilePrintButton],'Enable','off');
  set([handles.ScoresButton handles.LoadingsButton handles.ModelButton handles.GUI_ExploreScores handles.ResetImageClicks],'Enable','off')

  set(handles.EndAdjustMask,'Visible','on','Enable','on');
  setappdata(GUI_Handle,'Small_Data',SmlUD);                                    % Store the data
  setappdata(GUI_Handle,'Handles_Data',handles);                                % Store the handles structure

  prevWBDF = get(GUI_Handle,'WindowButtonDownFcn');
  prevWBUF = get(GUI_Handle,'WindowButtonUpFcn');
  set(GUI_Handle,'WindowButtonDownFcn','global SmlUD,SmlUD.adjModeButton = 1;set(gcbf,''Pointer'',''custom'',''PointerShapeCData'',SmlUD.HandOpenPointerShape,''PointerShapeHotSpot'',[10 9]);', ...
      'WindowButtonUpFcn','global SmlUD,SmlUD.adjModeButton = 0;set(gcbf,''Pointer'',''custom'',''PointerShapeCData'',SmlUD.HandOpenPointerShape,''PointerShapeHotSpot'',[10 9]);');
  
  %ErrorCatch = 0;  : not handled yet
  try
      uiwait(GUI_Handle);                                                       % <-- need a corresponding uiresume here: we have a pushbutton for this
  catch
      %ErrorCatch = 1;   : not handled yet
  end;

  if ishandle(handles.adjLine) && ishandle(handles.adjPoint)
      delete(handles.adjPoint);
      xData = get(handles.adjLine,'XData');
      yData = get(handles.adjLine,'YData');
      delete(handles.adjLine);

      SmlUD.Image.MaskXindex = get(handles.TScoreHorizAxis,'Value');
      SmlUD.Image.MaskYindex = get(handles.TScoreVrtclAxis,'Value');
      SmlUD.Image.MaskXdata = xData;
      SmlUD.Image.MaskYdata = yData;
      SmlUD.Image.Mask = uint8(ComputeScoreMask(SmlUD.Image.MaskXdata,SmlUD.Image.MaskYdata,256,256));   % compute the mask matrix (256 x 256)

      MyCLA(handles.ZoomScoreAxes);
      ShowImage(SmlUD.Image.Mask,handles.ZoomScoreAxes,[],[],[],'scaled');
  end;

  SmlUD.adjMode = 0;
  UpdateRHS(GUI_Handle);                                                        % Call another function to compute and display mask: this must come after adjMode has been turned off
  UpdateLHS(GUI_Handle);
  set([handles.TScoreHorizAxis handles.TScoreVrtclAxis],'Enable','on');
  set([handles.FileOpenButton handles.FilePrintButton],'Enable','on');          %"handles.FileSaveButton" is not enabled in this version
  set([handles.ScoresButton handles.LoadingsButton handles.ModelButton handles.GUI_ExploreScores handles.ResetImageClicks],'Enable','on')
  set(handles.EndAdjustMask,'Visible','off','Enable','off');

  setappdata(GUI_Handle,'Small_Data',SmlUD);

  if ishandle(GUI_Handle),
      set(GUI_Handle,'WindowButtonDownFcn',prevWBDF,'WindowButtonUpFcn',prevWBUF);
  end
return;


% --------------------------------------------------------------------
function SendLHSImageToNewFigureWindow(varargin)
% The user has right-clicked on a LHS image and wants to use that figure in the new figure window
% for whatever reason: printing, editing, copy to clipboard, etc.
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Copy the current image to the new figure and set the position of the image

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1
  NewFigureWindow   = figure('Visible','on');                                   % Create an invisible figure first
  handleAxis = axes;
  UpdateLHS(GUI_Handle,handleAxis,'nothing');                                   % It must have three inputs
  labelText = get(handles.LHSPlotLabel,'String');
  set(get(handleAxis,'title'),'String',labelText,'Visible','on');
  Txt = uicontrol('Parent',NewFigureWindow,'Style','Text','Units','Pixels','HorizontalAlign','Left','FontSize',10,'ForegroundColor',[0 0 0], ...
                  'Background',get(NewFigureWindow,'Color'),'Position',[1 1 200 23],'String','Exported from MACCMIA');
  extents = get(Txt,'Extent');
  set(Txt,'Position',[1 1 200 extents(4)]);

  if ~(SmlUD.RGBmode)                                                           % Only print associations if it is a non-RGB image
      if strcmp(get(handles.ScoresButton,'State'),'on') || strcmp(get(handles.ModelButton,'State'),'on')
          R = SmlUD.RLayer;    G = SmlUD.GLayer;    B = SmlUD.BLayer;
          NewXLabel = get(handleAxis,'XLabel');
          set(NewXLabel,'String',strcat(['Associations: Red = T_',num2str(R),'   Green = T_',num2str(G),'   Blue = T_',num2str(B)]),'Visible','on');
      end;
  end;
  set(NewFigureWindow,'Visible','on','Tag','Exported from MACCMIA','Toolbar','figure');            % Reveal the final result
return;


% --------------------------------------------------------------------
function SendRHSImageToNewFigureWindow(varargin)
% The user has right-clicked on a RHS image and wants to use that figure in the new figure window
% for whatever reason: printing, editing, copy to clipboard, etc.
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Copy the current image to the new figure and set the position of the image

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1
  NewFigureWindow   = figure('Visible','off');                                  % Create an invisible figure first
  handleAxis = axes;
  UpdateRHS(GUI_Handle,handleAxis,'nothing');                                   % It must have three inputs
  labelText = get(handles.RHSPlotLabel,'String');
  set(get(handleAxis,'title'),'String',labelText,'Visible','on');
  Txt = uicontrol('Parent',NewFigureWindow,'Style','Text','Units','Pixels','HorizontalAlign','Left','FontSize',10,'ForegroundColor',[0 0 0], ...
                  'Background',get(NewFigureWindow,'Color'),'Position',[1 1 200 23],'String','Exported from MACCMIA');
  extents = get(Txt,'Extent');
  set(Txt,'Position',[1 1 200 extents(4)]);

  if strcmp(get(handles.ScoresButton,'State'),'on')  % get(handles.GUI_ScoresAnalysis,'Value')    % If we are in the scores analysis mode - add colour bar
      set(NewFigureWindow,'Colormap',get(GUI_Handle,'Colormap'));               % We must set the figure's colour map before calling
      colorbar('vert','peer',handleAxis);                                       % the colorbar.m function to draw the colour bar
      set(NewFigureWindow,'Visible','on','Tag','Exported from MACCMIA');
  elseif strcmp(get(handles.LoadingsButton,'State'),'on')                       % If we are in the loading analysis mode - no colour bar
      RHSXLabel = get(handleAxis,'XLabel');
      RHSYLabel = get(handleAxis,'YLabel');
      P_XAxis   = get(handles.TScoreHorizAxis,'Value');                         % Get the X-axis setting: P1, P2, etc
      P_YAxis   = get(handles.TScoreVrtclAxis,'Value');                         % and the same for the Y-axis
      if ~(P_XAxis == P_YAxis)                                                  % If they are different then we need X and Y labels
          set(RHSYLabel,'String',['P_',num2str(get(handles.TScoreVrtclAxis,'Value'))]);
          RHSYLabelPosn = get(RHSYLabel,'Position');
          set(RHSYLabel,'Position',[RHSYLabelPosn(1)+0 RHSYLabelPosn(2:3)],'Visible','on');
      end;
      set(RHSXLabel,'String',['P_',num2str(get(handles.TScoreHorizAxis,'Value'))],'Visible','on');
      set(NewFigureWindow,'Visible','on','Tag','Exported from MACCMIA');
  elseif strcmp(get(handles.ModelButton,'State'),'on')                          % If we are in the model building mode - no colour bar
      RHSXLabel = get(handleAxis,'XLabel');
      RHSYLabel = get(handleAxis,'YLabel');
      set(RHSXLabel,'String',['T_',num2str(SmlUD.hAxis)],'Visible','on');
      set(RHSYLabel,'String',['T_',num2str(SmlUD.vAxis)]);
      RHSYLabelPosn = get(RHSYLabel,'Position');
      set(RHSYLabel,'Position',[0 RHSYLabelPosn(2:3)],'Visible','on');
      set(NewFigureWindow,'Visible','on','Colormap',get(GUI_Handle,'Colormap'),'Tag','Exported from MACCMIA');
  end;
return;


% --------------------------------------------------------------------
function SendBothToNewFigureWindow(varargin)
% The user has right-clicked on one of the Model building images and wants to send them both to a new figure window
% for whatever reason: printing, editing, copy to clipboard, etc.
% Step 0. Get the handle of the figure window and load any required data for this function
% Step 1. Copy the current image to the new figure and set the position of the image

% STEP 0
  global SmlUD
  GUI_Handle = getGUIHandle(varargin{1},'MACC_MIA_GUI');
  handles = getappdata(GUI_Handle,'Handles_Data');                              % Load the structure containing the handles
  if isempty(SmlUD)
      SmlUD = getappdata(GUI_Handle,'Small_Data');                              % Get the data storage structure
  end;

% STEP 1
  NewFigureWindow = figure('Visible','off');                                    % Create an invisible figure first
  hLHS = subplot(1,2,1);
  hRHS = subplot(1,2,2);
  UpdateLHS(GUI_Handle,hLHS,'nothing');
  UpdateRHS(GUI_Handle,hRHS,'nothing');

  labelText = get(handles.LHSPlotLabel,'String');
  set(get(hLHS,'title'),'String',labelText,'Visible','on');
  labelText = get(handles.RHSPlotLabel,'String');
  set(get(hRHS,'title'),'String',labelText,'Visible','on');
  Txt = uicontrol('Parent',NewFigureWindow,'Style','Text','Units','Pixels','HorizontalAlign','Left','FontSize',10,'ForegroundColor',[0 0 0], ...
                  'Background',get(NewFigureWindow,'Color'),'Position',[1 1 200 23],'String','Exported from MACCMIA');
  extents = get(Txt,'Extent');
  set(Txt,'Position',[1 1 200 extents(4)]);

  LHSXLabel         = get(hLHS,'XLabel');
  RHSXLabel         = get(hRHS,'XLabel');
  RHSYLabel         = get(hRHS,'YLabel');

  if ~(SmlUD.RGBmode) && strcmp(get(handles.ModelButton,'State'),'on')          % Only print associations if it is a non-RGB image
      R = SmlUD.RLayer;    G = SmlUD.GLayer;    B = SmlUD.BLayer;
      set(LHSXLabel,'String',strcat(['Associations: Red = T_',num2str(R),'   Green = T_',num2str(G),'   Blue = T_',num2str(B)]),'Visible','on');
  end;
  set(RHSXLabel,'String',['T_',num2str(SmlUD.hAxis)],'Visible','on');
  set(RHSYLabel,'String',['T_',num2str(SmlUD.vAxis)]);
  RHSYLabelPosn = get(RHSYLabel,'Position');
  set(RHSYLabel,'Position',[0 RHSYLabelPosn(2:3)],'Visible','on');
  set(NewFigureWindow,'Visible','on','Colormap',get(GUI_Handle,'Colormap'),'Tag','Exported from MACCMIA');
return;